/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_image.h"
#include "xplab_tprint.h"

/* local prototypes ********************************************************/
static cv_def_i full_correlation2(pic_def ref, sector_def rbs, pic_def pic, 
				  run_def *crun);
static void contour_marking(c_def center, sector_def clip, pic_def pic, 
			    PICT mark, int n4, CONTOUR_TYPE contour_type);
static h_contour_lst contour_registering(sector_def sr, pic_def pic);
static int distance_compatibility(int dist, double dt);
static void distance_update(int_lst dlst, double dt);
static double expect_mean_value(double_lst dst, int bin);
static double expect_variance(double_lst dst, double mv);
static double_lst distribution_func(double_lst dst);
static int change_sector(area_def ar, sector_def *sec, int sx, int sy, 
			 int ex, int ey);
static void cut_bmps(pic_def src, pic_def dst, sector_def srcs, 
		     int dst_offset_x, int dst_offset_y);
static void four2(double data[], ulong nn[], int isign);
static void four1(double data[], ulong nn, int isign);
static void power_spectral_density_2D(fft_def fd1, fft_def fd2);
static void power_density_auto(fft_def fd1, int pre_fd1_fft);
static void power_density_cross(fft_def fd1, int pre_fd1_fft, 
				fft_def fd2, int pre_fd2_fft);
static void renorm_2D(fft_def fd);
static void separate_complex_array(double *data, int length);

/* Functions **************************************************************/

/* Overlapping degree of two images.

Input : - image A and B
        - Offset B from A (lag = b - a)
Output: - degree (number of common pixels which are not background)
*/
int
overlapping_degree(pic_def pica, pic_def picb, int lx, int ly) 
{
  int tx, ty, sx, sy, syy, tyy, t1x, t1y, t2x, t2y, sxmin, symin, ovl = 0;

  t1x = XPMAX(0, lx);
  t1y = XPMAX(0, ly);
  t2x = XPMIN(pica.ar.width-1, lx+picb.ar.width-1);
  t2y = XPMIN(pica.ar.height-1, ly+picb.ar.height-1);
  sxmin = XPMAX(0, -lx); 
  symin = XPMAX(0, -ly);

  for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {
    syy = sy*picb.ar.width;
    tyy = ty*pica.ar.width;

    for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
      if (pica.mem[tx + tyy] != BACKGROUND && 
	  picb.mem[sx + syy] != BACKGROUND)
	ovl++;
    }
  }

  return ovl;
}



/* Wrapper for correlation of two images A and B.

Revision: 26.09.2003 E.Sack

Input : - binary image A
        - binary image B
        - pointer to control structure
Output: - matching features 
Precondition:  - it is KW(A,B) = KW(B,A), independent of image order. 
Postcondition: - CV of 1.0 only if both images are identic in 
                 size AND content.
*/
cv_def_i
full_correlation1(pic_def ref, pic_def pic, run_def *crun) 
{
  sector_def pic_rbs;

  assert(crun);

  AREA_TO_SECTOR(pic.ar, pic_rbs);

  return full_correlation2(ref, pic_rbs, pic, crun);        /* wraps input */
}



/* 2-D correlation of a reference within a clipping of an image.

Revision: 27.09.2003 E.Sack

Input : - binary reference
        - sector of binary image
        - binary image
        - pointer to control structure
Output: - matching features 
Postcondition: - CV of 1.0 only if reference and clipping are identic in 
                 size AND content.
*/
static cv_def_i
full_correlation2(pic_def ref, sector_def rbs, pic_def pic, run_def *crun) 
{
  int      good, xlag = 0, ylag = 0, kmax = 0;                    /* output */
  int      lymax, lxmax, lx, ly;                                     /* lag */
  int      t1x, t2x, t1y, t2y, tx, ty, tyy, gxx, gyy;                /* ROI */
  int      sx, sy, syy, sxmin, symin;                          /* reference */
  int      rx_1 = ref.ar.width - 1, ry_1 = ref.ar.height - 1;  /* reference */
  int      a1x = rbs.column.start, a2x = rbs.column.end;           /* image */
  int      a1y = rbs.line.start, a2y = rbs.line.end;               /* image */
  cv_def_i kwert;                                                 /* output */

  assert(crun);
  assert(a1x >= 0 && a1x < pic.ar.width);
  assert(a2x >= 0 && a2x < pic.ar.width);
  assert(a1y >= 0 && a1y < pic.ar.height);
  assert(a2y >= 0 && a2y < pic.ar.height);

  lymax = a2y - a1y;
  lxmax = a2x - a1x;

  gyy = ry_1 + a1y; 
  gxx = rx_1 + a1x; 

  for (ly = -ry_1; ly <= lymax && *crun->pactive==TRUE; ly++) {     /* lags */
    t1y = XPMAX(0,ly) + a1y;
    t2y = XPMIN(a2y, ly+gyy);
    symin = XPMAX(0, -ly);

    for (lx = -rx_1; lx <= lxmax && *crun->pactive==TRUE; lx++) {       
      xplab_progress_set_value(crun, -1.0, 0.0);                /* user IRQ */

      t1x = XPMAX(0,lx) + a1x;
      t2x = XPMIN(a2x,lx+gxx);
      sxmin = XPMAX(0,-lx);
      good = 0;

      for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {    /* correlation */
	syy = sy*ref.ar.width;
	tyy = ty*pic.ar.width;

	for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
          good += (pic.mem[tx + tyy] == ref.mem[sx + syy]) ? 1:0;
	}
      }

      if (kmax < good) {
	kmax = good;
	xlag = lx;
	ylag = ly;
      }
    }
  }

  kwert.cv = (double)kmax/(double)(MAX((lxmax+1)*(lymax+1), 
				       (rx_1+1)*(ry_1+1)));
  kwert.x = xlag;
  kwert.y = ylag;

  return kwert;
}



/* Calculates complete 2-D correlation. If the minimal correlation level 
   is definitely not reachable the calculation is interrupted. The 
   calculation is not interrupted, if the correlation level is met
   (asymmetric interrupt condition), to get the exact correlation value.

Input : - reference image
        - reference mask
        - sector of image
        - binary image
        - mass of mask
        - mask border
        - maximal allowed lag
        - minimal correlation level
Output: - matching features 
Usage:  - the given sector defines the sliding range (lags) and the
          range of correlation.
*/
cv_def_i 
masked_correlation_abbrev(pic_def ref, pic_def mask, sector_def rbs,
                          pic_def pic, double msm, int mask_border, 
		          int noise_border, double corrlevel)
{
  int     good, bad, lvl, xlag = 0, ylag = 0, kmax = 0;           /* output */
  int     sx, sy, syy, sxmin, symin;                                /* mask */
  int     lymin, lymax, lxmin, lxmax, lx, ly;                        /* lag */
  int     t1x, t2x, t1y, t2y, tx, ty, tyy, gxx, gyy;                 /* ROI */
  int     rx_1 = ref.ar.width-1, ry_1 = ref.ar.height-1;   /* masked refer. */
  int     a1x = rbs.column.start, a2x = rbs.column.end;            /* image */
  int     a1y = rbs.line.start, a2y = rbs.line.end;                /* image */
  cv_def_i kwert;                                                 /* output */

  assert(mask.ar.width==ref.ar.width);
  assert(mask.ar.height==ref.ar.height);

  lymin = XPMAX(-ry_1+mask_border, -noise_border-mask_border);
  lxmin = XPMAX(-rx_1+mask_border, -noise_border-mask_border);
  lymax = XPMIN(-mask_border,noise_border-ry_1+mask_border) + a2y - a1y;
  lxmax = XPMIN(-mask_border,noise_border-rx_1+mask_border) + a2x - a1x;

  gyy = ry_1 + a1y;
  gxx = rx_1 + a1x;

  lvl = (int)(msm * (1.0 - corrlevel));

  for (ly = lymin; ly <= lymax; ly++) {                             /* lags */
    t1y = XPMAX(0,ly) + a1y;
    t2y = XPMIN(a2y,ly+gyy);
    symin = XPMAX(0,-ly);

    for (lx = lxmin; lx <= lxmax; lx++) {
      t1x = XPMAX(0,lx) + a1x;
      t2x = XPMIN(a2x,lx+gxx);
      sxmin = XPMAX(0,-lx);
      good = bad = 0;

      for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {    /* correlation */
        syy = sy*ref.ar.width;
        tyy = ty*pic.ar.width;

	for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
	  if (!mask.mem[sx + syy]) continue;

          if (pic.mem[tx + tyy] == ref.mem[sx + syy]) {
	    good++;                                             /* matching */
	  }
	  else {
	    bad++;                                          /* not matching */
	  }
	}

	if (bad > lvl) {                                   /* premature end */
	  break;
	}
      }

      if (kmax < good) {
	kmax = good;                                /* reference lags image */
	xlag = lx;
	ylag = ly;
      }
    }
  }

  kwert.cv = (double) kmax;
  kwert.x = xlag;
  kwert.y = ylag;

  return kwert;
}



/* Calculates correlation as in masked_correlation_abbrev(), however the
   ranges of lags and correlation are separated.

Input : - reference image
        - reference mask
        - sector of image
        - extended sector of image
        - binary image
        - mass of mask
        - mask border
        - maximal allowed lag
        - minimal correlation level
Output: - matching features 
Usage:  - the sector defines the sliding range (lags) while the 
          extended sector defines the range of correlation.
*/
cv_def_i 
masked_correlation_ext(pic_def ref, pic_def mask, sector_def rbs, 
		       sector_def rbse, pic_def pic, double msm, 
		       int mask_border, int noise_border, double corrlevel)
{
  int      wi, good, bad, lvl, xlag = 0, ylag = 0, kmax = 0;      /* output */
  int      sx, sy, syy, sxmin, symin;                               /* mask */
  int      lymin, lymax, lxmin, lxmax, lx, ly;                       /* lag */
  int      t1x, t2x, t1y, t2y, tx, ty, tyy, gxx, gyy;                /* ROI */
  int      rx_1 = ref.ar.width-1, ry_1 = ref.ar.height-1;  /* masked refer. */
  int      a1x = rbs.column.start, a2x = rbs.column.end;           /* image */
  int      a1y = rbs.line.start, a2y = rbs.line.end;               /* image */
  int      e1x = rbse.column.start, e2x = rbse.column.end;     /* extension */
  int      e1y = rbse.line.start, e2y = rbse.line.end;         /* extension */
  cv_def_i kwert;                                                 /* output */

  assert(mask.ar.width==ref.ar.width);
  assert(mask.ar.height==ref.ar.height);

  lymin = XPMAX(-ry_1+mask_border, -noise_border-mask_border);
  lxmin = XPMAX(-rx_1+mask_border, -noise_border-mask_border);
  lymax = XPMIN(-mask_border,noise_border-ry_1+mask_border) + a2y - a1y;
  lxmax = XPMIN(-mask_border,noise_border-rx_1+mask_border) + a2x - a1x;

  gyy = ry_1 + a1y;
  gxx = rx_1 + a1x;

  lvl = (int)(msm * (1.0 - corrlevel));

  for (ly = lymin; ly <= lymax; ly++) {                             /* lags */
    t1y = XPMAX(e1y,ly+a1y);
    t2y = XPMIN(e2y,ly+gyy);
    symin = XPMAX(0,e1y-a1y-ly);

    for (lx = lxmin; lx <= lxmax; lx++) {
      t1x = XPMAX(e1x,lx+a1x);
      t2x = XPMIN(e2x,lx+gxx);
      sxmin = XPMAX(0,e1x-a1x-lx);
      good = bad = 0;

      for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {    /* correlation */
        syy = sy*ref.ar.width;
        tyy = ty*pic.ar.width;

	for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
	  if (!mask.mem[wi = sx + syy]) {
	    continue;
	  }

          if (pic.mem[tx + tyy] == ref.mem[wi]) {
	    good++;                                             /* matching */
	  }
	  else {
	    bad++;                                          /* not matching */
	  }
	}

	if (bad > lvl) {                                   /* premature end */
	  break;
	}
      }

      if (kmax < good) {
	kmax = good;                                /* reference lags image */
	xlag = lx;
	ylag = ly;
      }
    }
  }

  kwert.cv = (double) kmax;
  kwert.x = xlag;
  kwert.y = ylag;

  return kwert;
}



/* Calculates sum of absolut difference of image and reference on a
   fixed lag.

Revision: 02.02.2003 E.Sack

Usage: Images are weighted.

Input : - masked gray-value reference
        - reference mask
        - clipping
        - gray-value image
        - horizontal and vertical lag
Output: - matching pixel number
*/
cv_def
wtd_correlation(pic_def ref, pic_def mask, sector_def rbs, pic_def pic,
                int lx, int ly)
{
  int    wp, wi, wr, knl = 0, bad = 0, sum = 0;
  int    sx, sy, syy, sxmin = XPMAX(0,-lx), symin = XPMAX(0,-ly);   /* mask */
  int    t1x, t2x, t1y, t2y, tx, ty, tyy;                            /* ROI */
  cv_def feat;

  assert(mask.ar.width==ref.ar.width);
  assert(mask.ar.height==ref.ar.height);

  t1y = XPMAX(0,ly)+rbs.line.start;
  t1x = XPMAX(0,lx)+rbs.column.start;
  t2y = XPMIN(rbs.line.end,ly+ref.ar.height-1+rbs.line.start);
  t2x = XPMIN(rbs.column.end,lx+ref.ar.width-1+rbs.column.start);

  for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {
    syy = sy*ref.ar.width;
    tyy = ty*pic.ar.width;

    for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
      if (!mask.mem[wi = sx + syy]) {
	continue;
      }
      
      bad += abs((wr = pic.mem[tx + tyy]) - (wp = ref.mem[wi]));  /* pixels */
      sum += XPMAX(1,wp);                               /* with mask border */
      knl += (wp ? wr : 0);                           /* hits on NB pattern */
    }
  }

  feat.good = sum-bad;                            /* overlay degree by mask */
  feat.kernel = knl;                            /* overlay degree by kernel */

  return feat;      
}



/* Scans a clipping for a foreground pixel with a reference on a
   fixed lag.

Revision: 30.05.2004 E.Sack

Input : - reference mask
        - clipping
        - image
        - horizontal and vertical lag
Output: - start coordinate in clipping (negative values = nothing found)
*/
c_def
scan_bmp_masked(pic_def mask, sector_def rbs, pic_def pic, int lx, int ly)
{
  int   sx, sy, syy, sxmin = XPMAX(0,-lx), symin = XPMAX(0,-ly);   /* mask */
  int   t1x, t2x, t1y, t2y, tx, ty, tyy;                            /* ROI */
  c_def coo;

  ASSERT_SECTOR(rbs, pic.ar);

  init_c_def(&coo);

  t1y = XPMAX(0,ly)+rbs.line.start;
  t1x = XPMAX(0,lx)+rbs.column.start;
  t2y = XPMIN(rbs.line.end, ly+mask.ar.height-1+rbs.line.start);
  t2x = XPMIN(rbs.column.end, lx+mask.ar.width-1+rbs.column.start);

  for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {
    syy = sy*mask.ar.width;
    tyy = ty*pic.ar.width;

    for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
      if (!mask.mem[sx + syy]) {
	continue;
      }
      
      if (pic.mem[tx + tyy] == FOREGROUND) {
	coo.x = tx;
	coo.y = ty;

	return coo;
      }
    }
  }

  return coo;      
}



/* Calculates correlation as in masked_correlation_ext(), however the
   lag is fixed. 

Input : - reference image
        - reference mask
        - sector of image
        - extended sector of image
        - binary image
        - 2-D lag
Output: - matching pixel number
Usage:  - The function is normally used to refine the result of a previous
          sliding correlation.
*/
int
wtd_correlation_ext(pic_def ref, pic_def mask, sector_def rbs, 
		    sector_def rbse, pic_def pic, int lx, int ly)
{
  int wp, wi, bad = 0, sum = 0;                                   /* output */
  int sx, sy, syy, sxmin, symin;                                    /* mask */
  int t1x, t2x, t1y, t2y, tx, ty, tyy;                               /* ROI */

  assert(mask.ar.width==ref.ar.width);
  assert(mask.ar.height==ref.ar.height);

  t1y = XPMAX(rbse.line.start,ly+rbs.line.start);
  t1x = XPMAX(rbse.column.start,lx+rbs.column.start);
  t2y = XPMIN(rbse.line.end,ly+ref.ar.height-1+rbs.line.start);
  t2x = XPMIN(rbse.column.end,lx+ref.ar.width-1+rbs.column.start);
  sxmin = XPMAX(0,rbse.column.start-rbs.column.start-lx); 
  symin = XPMAX(0,rbse.line.start-rbs.line.start-ly);

  for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {
    syy = sy*ref.ar.width;
    tyy = ty*pic.ar.width;

    for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
      if (!mask.mem[wi = sx + syy]) {
	continue;
      }
      
      bad += abs(pic.mem[tx + tyy] - (wp = ref.mem[wi]));
      sum += XPMAX(1,wp);
    }
  }

  return sum-bad;              /* nur die Guten zhlen wg. berdeckungsgrad */
}



/* Ascertains a contour regarding a mark in a clipping.

Revision: 31.05.2004 E.Sack

Input : - image
        - clipping
        - mark
        - scan direction
        - neighborhood
Output: - contour list
Postcondition: - only one contour of possibly several ones got.
*/
h_contour_lst
xgen_h_contour(pic_def pic, sector_def fr, PICT search, int bot_up, int n4)
{
  int           d, oflag = 1;
  c_def         start;
  h_contour_lst ctr;

  ASSERT_SECTOR(fr, pic.ar);

  init_h_contour_lst(&ctr);

  d = xscan_bmp(fr, &start, pic, search, &oflag, bot_up);

  if (!d) {
    return ctr;
  }

  ctr = xgen_h_contour_from_start(pic, fr, start, n4);

  return ctr;
}



/* Scans a bitmap for a certain mark.

Revision: 01.07.2004 E.Sack

Input : - clipping
        - pointer to running coordinate
        - image
        - mark
        - pointer to init flag (1: reset running coordinate)
        - scan direction
Output: - running coordinate updated as flagged
        - init flag updated
        - flag (0: no mark found, 1: mark found)
*/
int
xscan_bmp(sector_def clip, c_def *c, pic_def pic, PICT mark, int *init_scan,
          int bot_up) 
{
  int yy, sinit = *init_scan;

  assert(clip.column.start >= 0); 
  assert(clip.line.start  >= 0); 
  assert(clip.column.end < pic.ar.width); 
  assert(clip.line.end  < pic.ar.height); 

  *init_scan = 0;

  if (bot_up) {
    c->y = sinit ? clip.line.end : c->y;

    for (; c->y >= clip.line.start; c->y--) {
      yy = c->y*pic.ar.width;

      for (c->x = clip.column.end; c->x >= clip.column.start; c->x--) {
	if (pic.mem[c->x + yy] == mark) {
	  return 1;
	}
      } 
    }
  }
  else {
    c->y = sinit ? clip.line.start : c->y;

    for (; c->y <= clip.line.end; c->y++) {
      yy = c->y*pic.ar.width;

      for (c->x = clip.column.start; c->x <= clip.column.end; c->x++) {
	if (pic.mem[c->x + yy] == mark) {
	  return 1;
	}
      } 
    }
  }

  return 0;                            /* clipping scanned without success */
}



/* Ascertains a contour regarding a mark from a start coordinate in a
   clipping.

Revision: 30.05.2004 E.Sack

Input : - image
        - clipping
        - start coordinate
        - neighborhood (0: 8-N, 1: 4-N)
Output: - contour list
Postcondition: - only one contour of possibly several ones got.
*/
h_contour_lst
xgen_h_contour_from_start(pic_def pic, sector_def fr, c_def start, int n4)
{
  PICT          search = pic.mem[start.x + start.y*pic.ar.width];
  h_contour_lst ctr;

  ASSERT_SECTOR(fr, pic.ar);

  contour_marking(start, fr, pic, search, n4, CONTOUR_OUTER_INC);
  ctr = contour_registering(fr, pic);                     /* reset contour */

  if (ctr.type != CONTOUR_OUTER_INC) {             /* avoid inner contours */
    contour_marking(start, fr, pic, search, n4, CONTOUR_INNER_EXC);

    free_h_contour_lst(&ctr);
    ctr = contour_registering(fr, pic);                   /* reset contour */
  }                 /* attempt to get an outer contour may be unsuccessful */

  return ctr;
}



/* Flags a contour of marked pixels in a clipping with a given start
   coordinate using edge detection on a 3x3-frame around a contour pixel.
   4/8 neighborhood is used for the testcases on the frame. Searching
   stops if a closed contour is detected.

   4-N testcases of center X:   8-N testcases of center X:
          +-----+                      +-----+
          | |B| |                      |A|B|C|
          |H|X|D|                      |H|X|D|
          | |F| |                      |G|F|E|
          +-----+                      +-----+

Revision: 31.05.2004 E.Sack

Input : - start coordinate
        - clipping
        - image
        - contour mark
        - neighborhood (0: 8-N, 1: 4-N)
        - contour type
Output: - image updated
Precondition:  - start coordinate contains a marked pixel not necessarily
                 being a contourpixel.
               - caller is responsible for multiple bits set of the pixel.
Postcondition: - there can be more than one contour in the clipping, 
                 however only one is flagged (holes are not recognized). 
                 It is up to the caller to manage multiple contours.
               - inner contour means a contour of a hole resulting in
                 swapped start/end points of the whole contour. It is up
                 to the caller to resolve that fact.
               - only bit inquiry used
               - contour marks set in image

Usage: Use 'contour_registering' to get a contour list instead of a marked image.
*/
static void
contour_marking(c_def center, sector_def clip, pic_def pic, PICT mark, int n4,
	        CONTOUR_TYPE contour_type)
{
  /* 4-N and 8-N coordinates                                               */
#define PCENTER pic.mem[center.x + center.y*pic.ar.width]
#define COO_H(ctmp) do {ctmp = center; ctmp.x--;} while(0)
#define COO_B(ctmp) do {ctmp = center; ctmp.y--;} while(0)
#define COO_D(ctmp) do {ctmp = center; ctmp.x++;} while(0)
#define COO_F(ctmp) do {ctmp = center; ctmp.y++;} while(0)

  /* 4-N coordinates                                                       */
#define COO_A(ctmp) do {ctmp = center; ctmp.x--; ctmp.y--;} while(0)
#define COO_C(ctmp) do {ctmp = center; ctmp.x++; ctmp.y--;} while(0)
#define COO_E(ctmp) do {ctmp = center; ctmp.x++; ctmp.y++;} while(0)
#define COO_G(ctmp) do {ctmp = center; ctmp.x--; ctmp.y++;} while(0)

  /* attention must be made not to leave allocated memory                  */
#define PVALUE(coo) pic.mem[coo.x + coo.y*pic.ar.width]

/* is value within interval?                                               */
#define PIVAL(i,ival) ((i >= ival.start) && (i <= ival.end))

/* is coordinate within clipping?                                          */
#define PCLIP(coo,sec) (PIVAL(coo.x, sec.column) && PIVAL(coo.y, sec.line))

/* contains coordinate a marked pixel?                                     */
#define PMARK(coo,mrk) (PVALUE(coo) & mrk)

/* contains coordinate a marked pixel and is it within clipping?           */
#define IS_FG(coo) (PCLIP(coo, clip) && PMARK(coo, mark))

  int        watchdog = 0, contour_found = 0;
  char       tcase;
  c_def      ctmp;
  sector_def isec;

  ASSERT_SECTOR(clip, pic.ar);
  AREA_TO_SECTOR(pic.ar, isec);

  assert(mark&PCENTER);
  assert(!(mark&CONTOURSTART));
  assert(!(mark&CONTOUREND));

  switch (contour_type) {
  case CONTOUR_OUTER_INC:
    while (IS_FG(center)) {
      center.x--;                /* start on left border for outer contour */
    }

    center.x++;                                            /* left contour */
    tcase = 'H';                                      /* starting testcase */
    break;

  case CONTOUR_INNER_EXC:
    while (IS_FG(center)) {
      center.x++;                /* start on right border for outer contour */
    }

    center.x--;                                            /* right contour */
    tcase = 'D';                                       /* starting testcase */
    break;

  default: assert(0); break;
  }

  while (!contour_found) {
    switch (tcase) {                                          /* clockwise */
    case 'H':                                      /* testcase for 4N & 8N */
      COO_H(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position H */
	  tcase = n4?'F':'G';                        /* call next testcase */
	  break;
	}
      }

      if (PVALUE(center) & CONTOURSTART) {               /* contour closed */
	contour_found = 1;
	break;
      }

      PCENTER |= CONTOURSTART;                       /* contour line start */
      watchdog++;

      tcase = n4?'B':'A';                            /* call next testcase */
      break;

    case 'D':                                      /* testcase for 4N & 8N */
      COO_D(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position D */
	  tcase = n4?'B':'C';                        /* call next testcase */
	  break;
	}
      }

      if (PCENTER & CONTOUREND) {                        /* contour closed */
	contour_found = 1;
	break;
      }

      PCENTER |= CONTOUREND;                           /* contour line end */
      watchdog--;
      tcase = n4?'F':'E';                            /* call next testcase */
      break;

    case 'B':                                      /* testcase for 4N & 8N */
      COO_B(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position B */
	  tcase = 'H';                               /* call next testcase */
	  break;
	}
      }

      tcase = n4?'D':'C';                            /* call next testcase */
      break;

    case 'F':                                      /* testcase for 4N & 8N */
      COO_F(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position F */
	  tcase = 'D';                               /* call next testcase */
	  break;
	}
      }

      tcase = n4?'H':'G';                            /* call next testcase */
      break;

    case 'A':                                           /* testcase for 8N */
      COO_A(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position A */
	  tcase = 'G';                               /* call next testcase */
	  break;
	}
      }

      tcase = 'B';                                   /* call next testcase */
      break;

    case 'C':                                           /* testcase for 8N */
      COO_C(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position C */
	  tcase = 'H';                               /* call next testcase */
	  break;
	}
      }

      tcase = 'D';                                   /* call next testcase */
      break;

    case 'E':                                           /* testcase for 8N */
      COO_E(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position E */
	  tcase = 'C';                               /* call next testcase */
	  break;
	}
      }

      tcase = 'F';                                   /* call next testcase */
      break;

    case 'G':                                           /* testcase for 8N */
      COO_G(ctmp);

      if (PCLIP(ctmp, isec)) {                             /* for security */
	if (IS_FG(ctmp)) {
	  center = ctmp;                      /* move center to position G */
	  tcase = 'D';                               /* call next testcase */
	  break;
	}
      }

      tcase = 'H';                                   /* call next testcase */
      break;

    default: assert(0); break;
    }
  }

  assert(!watchdog);                                 /* must be consistent */

#undef COO_H
#undef COO_B
#undef COO_D
#undef COO_F
#undef PCENTER
#undef PVALUE
#undef PIVAL
#undef PCLIP
#undef PMARK
#undef IS_FG
}



/* Enrolls a contour marked in an image to a contour list.
   Contour marked in horizontal lines, start/end mark is in any order:
   start<end: outer contour, start>end: inner contour. The start-end-pair
   in a line must be complete. Although there can be more than one contour
   in an image line, the list contains only one contour pair per entry. A
   contour can never change its fashion: it is an inner contour or an outer
   contour. This algorithm of contour registering does not allow overlayed
   contours.

Usage: Use 'contour_marking' to mark a contour in an image.

Revision: 30.05.2004 E.Sack

Input : - clipping
        - image
Output: - contour
        - image updated (see postcondition)
Precondition:  - see description above
Postcondition: - contour marking bits reset
               - The type of contour (inner <-> outer) is reported in the
                 list.
*/
static h_contour_lst
contour_registering(sector_def sr, pic_def pic) 
{
  int           idx = 0, y, yy, x, width = pic.ar.width;
  int           watchdog = 0, is_start, is_end, incomplete_pair = 0;
  PICT          *pixel = 0;
  h_contour_lst ctr;

  ASSERT_SECTOR(sr, pic.ar);
  assert(SECTOR_SIZE(sr) > 0);

  init_h_contour_lst(&ctr);

  for (y = sr.line.start; y <= sr.line.end; y++) {
    yy = y*width;

    for (x = sr.column.start; x <= sr.column.end; x++) { /* hor. line scan */
      pixel = &pic.mem[x + yy];

      is_start = *pixel & CONTOURSTART;
      is_end = *pixel & CONTOUREND;           

      if (!is_start && !is_end) {
	continue;                                   /* not a contour pixel */
      }

      if (!incomplete_pair) {
	idx = ctr.cnt++;                               /* new contour line */

	if (idx) {
	  REALLOC(h_contour, ctr.cnt, ctr.hcl);
	}
	else {
	  CALLOC(h_contour, ctr.cnt, ctr.hcl);
	}

	init_h_contour(&ctr.hcl[idx]);
      }

      if (is_start) {
	ctr.hcl[idx].xil.start = x;
	*pixel &= ~CONTOURSTART;                  /* erase contour marking */
	watchdog++;
	incomplete_pair = incomplete_pair?0:1;
      }

      if (is_end) {                            /* contour may be one pixel */
	ctr.hcl[idx].xil.end = x;
	ctr.hcl[idx].yco     = y;             /* entered in list only once */
	*pixel &= ~CONTOUREND;                    /* erase contour marking */
	watchdog--;
	incomplete_pair = incomplete_pair?0:1;
      }

      if (!incomplete_pair) {
	CONTOUR_TYPE type = (ctr.hcl[idx].xil.start > ctr.hcl[idx].xil.end) ?
	  CONTOUR_INNER_EXC : CONTOUR_OUTER_INC;

	if (ctr.type == CONTOUR_UNDEF) {
	  ctr.type = type;                           /* enter contour type */
	}
	else {
	  assert(ctr.type == type);     /* contour type must be consistent */
	}
      }
    }
  }

  assert(!watchdog);                                 /* must be consistent */

  return ctr;
}



/* Enlarges or lessens a sector regarding an area by given values.

Input : - area
        - pointer to sector
        - deltas
Output: - sector updated
        - Flag (0 sector lies (partially) within area,
                1 sector lies outer area
                2 sector is zero or negative)
Postcondition: some values of sector may be negative
*/
static int
change_sector(area_def ar, sector_def *sec, int sx, int sy, int ex, int ey)
{
 sec->column.start += sx;
 sec->column.end   += ex;
 sec->line.start  += sy;
 sec->line.end    += ey;

 if (sec->column.end < 0 || sec->line.end < 0 ||
       sec->column.start >= ar.width || sec->line.start >= ar.height) {
       return 1;                                       /* sector fall out */
}

 if (sec->column.end < sec->column.start || 
     sec->line.end < sec->line.start) {
       return 2;                      /* sector clips to zero or negative */
}

 return 0;
}



/* Copies a clipping of a source image to the destination image. The 
   clipping is not constrained to the image sizes, however it have to
   cover the source image partially.

Input : - source image
        - destination image
        - clipping of the source
        - offset to the destination
Output: - destination image updated
*/
static void 
cut_bmps(pic_def src, pic_def dst, sector_def srcs, int dst_offset_x, 
	 int dst_offset_y)
{
  int      xo, xs, xe, yo, ys, ye;
  area_def cut;

  assert(dst_offset_x >= 0); assert(dst_offset_y >= 0);
  assert(src.ar.width > 0);  assert(src.ar.height > 0);
  assert(dst.ar.width > 0);  assert(dst.ar.height > 0);
  assert(src.mem); assert(dst.mem);
  assert(srcs.column.end >= 0 &&                   /* sector don't fall out */
         srcs.line.end  >= 0 &&
         srcs.column.start < src.ar.width && 
         srcs.line.start < src.ar.height);
  assert(srcs.column.end >= srcs.column.start && 
         srcs.line.end >= srcs.line.start);     /* sector clips to positive */

  xo = XPMIN(0, srcs.column.start);              /* offset from cutted area */
  xs = XPMAX(0, srcs.column.start);              /* real source area start  */
  xe = XPMIN(src.ar.width-1, srcs.column.end);   /* real source area end    */
  yo = XPMIN(0, srcs.line.start);
  ys = XPMAX(0, srcs.line.start);
  ye = XPMIN(src.ar.height-1, srcs.line.end);

  cut.width = xe-xs+1;                       /* real source area            */
  cut.height = ye-ys+1; 

  dst_offset_x -= xo;                        /* real destination area start */
  dst_offset_y -= yo;

  /* real compound area (we are working with offsets from the origin)       */
  cut.width -= XPMAX(0, dst_offset_x + cut.width - dst.ar.width);
  cut.height -= XPMAX(0, dst_offset_y + cut.height - dst.ar.height);

  /* real area placements and clippings                                     */
  copy_bmps(src, dst, cut, xs, ys, dst_offset_x, dst_offset_y);
}



/* Counts and deletes dots.

Revision: 18.01.2003 E.Sack

Input : - image
        - neighborhood
        - flag (0: delete dots, 1: don't delete dots)
Output: - image updated as flagged
        - number of dots
Postcondition: - image size unchanged
               - image may be empty
*/
int
pixel_erase(pic_def pic, NBH nbh, int support_dots)
{
#define B(X,Y)  pic.bvl[(X)   +  (Y)   *w]
#define O(X,Y)  pic.mem[(X)   +  (Y)   *w]
#define N(X,Y)  pic.mem[(X)   + ((Y)-1)*w]
#define E(X,Y)  pic.mem[(X)+1 +  (Y)   *w]
#define S(X,Y)  pic.mem[(X)   + ((Y)+1)*w]
#define W(X,Y)  pic.mem[(X)-1 +  (Y)   *w]
#define NE(X,Y) pic.mem[(X)+1 + ((Y)-1)*w]
#define SE(X,Y) pic.mem[(X)+1 + ((Y)+1)*w]
#define SW(X,Y) pic.mem[(X)-1 + ((Y)+1)*w]
#define NW(X,Y) pic.mem[(X)-1 + ((Y)-1)*w]

#define BVL(X,Y) if (pic.bvl && !support_dots) {              \
                   B(X,Y) = (B(X,Y) && !O(X,Y)) ? 0 : B(X,Y); \
                 }

#define ERASE_DOTS(P,D) if (!support_dots) {   \
                          P = D;               \
                        }

  int s, z, count = 0, w = pic.ar.width, h = pic.ar.height;
  int pixel;                                           /* pixel under test */

  assert(pic.mem);

  for (z = 1; z < h-1; z++) {                       /* image without frame */
    for (s = 1; s < w-1; s++) {
      pixel = O(s,z);
      if (((nbh == NBH4X) ||
	   ((W(s,z)  != pixel) &&
	    (E(s,z)  != pixel) &&
	    (N(s,z)  != pixel) &&
	    (S(s,z)  != pixel))) &&
          ((nbh == NBH4C) ||
           ((NW(s,z) != pixel) &&
            (SW(s,z) != pixel) &&
            (NE(s,z) != pixel) &&
            (SE(s,z) != pixel)))) {
	ERASE_DOTS(O(s,z), W(s,z));
	BVL(s,z);
	count++;
      }
    }
  }

  for (z = 1; z < h-1; z++) {                 /* left side without corners */
    pixel = O(0,z);
    if (((nbh == NBH4X) ||
	 ((E(0,z)  != pixel) &&
	  (N(0,z)  != pixel) &&
	  (S(0,z)  != pixel))) &&
	((nbh == NBH4C) ||
	 ((NE(0,z) != pixel) &&
	  (SE(0,z) != pixel)))) {
      ERASE_DOTS(O(0,z), E(0,z));
      BVL(0,z);
      count++;
    }
  }

  for (z = 1; z < h-1; z++) {                /* right side without corners */
    pixel = O(w-1,z);
    if (((nbh == NBH4X) ||
	 ((W(w-1,z)  != pixel) &&
	  (N(w-1,z)  != pixel) &&
	  (S(w-1,z)  != pixel))) &&
	((nbh == NBH4C) ||
	 ((NW(w-1,z) != pixel) &&
	  (SW(w-1,z) != pixel)))) {
      ERASE_DOTS(O(w-1,z), W(w-1,z));
      BVL(w-1,z);
      count++;
    }
  }

  for (s = 1; s < w-1; s++) {                /* upper side without corners */
    pixel = O(s,0);
    if (((nbh == NBH4X) ||
	 ((S(s,0)  != pixel) &&
	  (W(s,0)  != pixel) &&
	  (E(s,0)  != pixel))) &&
	((nbh == NBH4C) ||
	 ((SW(s,0) != pixel) &&
	  (SE(s,0) != pixel)))) {
      ERASE_DOTS(O(s,0), S(s,0));
      BVL(s,0);
      count++;
    }
  }

  for (s = 1; s < w-1; s++) {               /* bottom side without corners */
    pixel = O(s,h-1);
    if (((nbh == NBH4X) ||
	 ((N(s,h-1)  != pixel) &&
	  (W(s,h-1)  != pixel) &&
	  (E(s,h-1)  != pixel))) &&
	((nbh == NBH4C) ||
	 ((NW(s,h-1) != pixel) &&
	  (NE(s,h-1) != pixel)))) {
      ERASE_DOTS(O(s,h-1), N(s,h-1));
      BVL(s,h-1);
      count++;
    }
  }

  pixel = O(0,0);                                     /* upper left corner */
  if (((nbh == NBH4X) ||
       ((E(0,0) != pixel) &&
	(S(0,0) != pixel))) &&
      ((nbh == NBH4C) ||
       (SE(0,0) != pixel))) {
    ERASE_DOTS(O(0,0), E(0,0));
    BVL(0,0);
    count++;
  }

  pixel = O(0,h-1);                                  /* bottom left corner */
  if (((nbh == NBH4X) ||
       ((E(0,h-1) != pixel) &&
	(N(0,h-1) != pixel))) &&
      ((nbh == NBH4C) ||
       (NE(0,h-1) != pixel))) {
    ERASE_DOTS(O(0,h-1), E(0,h-1));
    BVL(0,h-1);
    count++;
  }

  pixel = O(w-1,0);                                  /* upper right corner */
  if (((nbh == NBH4X) ||
       ((W(w-1,0) != pixel) &&
	(S(w-1,0) != pixel))) &&
      ((nbh == NBH4C) ||
       (SW(w-1,0) != pixel))) {
    ERASE_DOTS(O(w-1,0), W(w-1,0));
    BVL(w-1,0);
    count++;
  }

  pixel = O(w-1,h-1);                               /* bottom right corner */
  if (((nbh == NBH4X) ||
       ((W(w-1,h-1) != pixel) &&
	(N(w-1,h-1) != pixel))) &&
      ((nbh == NBH4C) ||
       (NW(w-1,h-1) != pixel))) {
    ERASE_DOTS(O(w-1,h-1), W(w-1,h-1));
    BVL(w-1,h-1);
    count++;
  }

  return count;

#undef B
#undef O
#undef N
#undef NE
#undef E
#undef SE
#undef S
#undef SW
#undef W
#undef NW
#undef BVL
#undef ERASE_DOTS
}



/* Copies a masked image clipping.
   Terminology: R1 - mask border of NB, degree of dilatation = 1
                R2 - mask border of NB, degree of dilatation = 2

Revision: 27.08.2005 E.Sack

Input : - masked binary reference
        - reference mask
        - RBS of image
        - working image
        - original image
        - lags
        - kernel
        - clustering threshold
Output: - updated image 
Precondition: - clipping within image
Postcondition: - the image copied can grow within its mask.
*/
pic_def
xcopy_masked_area(pic_def ref, pic_def mask, sector_def rbs, pic_def pic,
                  pic_def org, int lx, int ly, kern_def ckern, 
		  int cluster_thold)
{
#define R1 (PICT)8    /* must be different to bitmap colors black and white */
#define R2 (PICT)16   /* must be different to bitmap colors black and white */
#define DSP 2                           /* additional border, multiple of 2 */
  PICT       *blood = 0;
  int        d, sx, sy, syy, poi;
  int        cx, cy, t1x, t2x, t1y, t2y, tx, ty, tyy, cyy, sxmin, symin;
  int        bx, by;
  int        rx = ref.ar.width, ry = ref.ar.height;            /* with mask */
  area_def   cut;
  pic_def    copy, m1, m2, rc, bp, bo;
  sector_def nbs, roi;

  init_pic_def(&copy);
  init_pic_def(&bp);
  init_pic_def(&bo);
  init_pic_def(&m1);
  init_pic_def(&m2);
  init_pic_def(&rc);

  roi.column.start = XPMAX(0, rbs.column.start+lx);   /* implicit with mask */
  roi.line.start = XPMAX(0, rbs.line.start+ly);
  roi.column.end = XPMIN(org.ar.width-1, rbs.column.start+lx+ref.ar.width-1);
  roi.line.end = XPMIN(org.ar.height-1, rbs.line.start+ly+ref.ar.height-1);

  copy.org.x = rbs.column.start+lx;       /* absolut origin according input */
  copy.org.y = rbs.line.start+ly;

  lx = rbs.column.start+lx-roi.column.start;                /* correct lags */
  ly = rbs.line.start+ly-roi.line.start;

  bx = INTERVAL_SIZE(roi.column);
  by = INTERVAL_SIZE(roi.line);

  /* copy RB-clipping enlarged by 2 pixels in temporary buffer because of
     R2 (can be greater than RB image).                                     */
  bp.ar.width = bx + DSP + DSP;
  bp.ar.height = by + DSP + DSP;
  CALLOC(PICT, AREA_SIZE(bp.ar), bp.mem);
  bo.ar = bp.ar;                                 /* dto. for original image */
  CALLOC(PICT, AREA_SIZE(bo.ar), bo.mem);

  d = change_sector(pic.ar, &roi, -DSP, -DSP, DSP, DSP);     /* make border */
  assert(!d);                                  /* sector shouldn't fall out */

  cut_bmps(pic, bp, roi, 0, 0);               /* copy centered and bordered */
  cut_bmps(org, bo, roi, 0, 0);

  m1.ar.width = rx + DSP;                    /* place for dilation stage 2  */
  m1.ar.height = ry + DSP;                   /* already placed dil. stage 1 */
  CALLOC(PICT, AREA_SIZE(m1.ar), m1.mem);
  m2.ar = m1.ar;
  CALLOC(PICT, AREA_SIZE(m2.ar), m2.mem);
  rc.ar = m1.ar;
  CALLOC(PICT, AREA_SIZE(rc.ar), rc.mem);                 /* redundant      */
  CALLOC(PICT, AREA_SIZE(rc.ar), rc.bvl);                 /* takes b-values */

  AREA_TO_SECTOR(ref.ar, nbs);

  cut_bmps(ref, rc, nbs, DSP>>1, DSP>>1);                 /* b-value taken */
  cut_bmps(mask, m1, nbs, DSP>>1, DSP>>1);/* mask contains dilation stage 1 */
  cut_bmps(mask, m2, nbs, DSP>>1, DSP>>1);

  cut.width = rx;
  cut.height = ry;
  dilate_in_box(m2);              /* dilation stage 2 from dilation stage 1 */
  mark_diff(mask, m2, cut, 0, 0, DSP>>1, DSP>>1, FOREGROUND);

  d = set_all_pixel(m2, FOREGROUND, R2);                    /* mark bound 2 */
  assert(d);                                            /* pixel must exist */

  mark_diff(ref, m1, cut, 0, 0, DSP>>1, DSP>>1, FOREGROUND);
  d = set_all_pixel(m1, FOREGROUND, R1);                    /* mark bound 1 */
  assert(d);                                            /* pixel must exist */

  add_grey(m2, m1, m1.ar, 0, 0, 0, 0);        /* merging for simple inquiry */
  add_grey(ref, m1, cut, 0, 0, DSP>>1, DSP>>1);
  sxmin = XPMAX(0,-lx)+(DSP>>1);
  symin = XPMAX(0,-ly)+(DSP>>1);
  t1x = XPMAX(0,lx)+DSP;
  t1y = XPMAX(0,ly)+DSP;
  t2x = XPMIN(bx-1, lx+rx-1)+DSP;
  t2y = XPMIN(by-1, ly+ry-1)+DSP;

  copy.ar.width = t2x-t1x+1;
  copy.ar.height = t2y-t1y+1;

  CALLOC(PICT, AREA_SIZE(copy.ar), copy.mem);
  CALLOC(PICT, AREA_SIZE(copy.ar), blood);

  for (ty = t1y, cy = 0, sy = symin; ty <= t2y; ty++, sy++, cy++) {
    cyy = cy*copy.ar.width;
    syy = sy*m1.ar.width;
    tyy = ty*bp.ar.width;

    for (tx = t1x, cx = 0, sx = sxmin; tx <= t2x; tx++, sx++, cx++) {
      poi = bp.mem[tx + tyy];                          /* point of interest */

      switch (m1.mem[sx + syy]) {                   /* mask marking inquiry */
      case FOREGROUND:                        /* NB marking: always catched */
        copy.mem[cx + cyy] = poi;           /* just a copy within reference */
        break;

      case R1: {                  /* NB dilation grade 1, pixel of interest */
        #define BIX(X,Y)  rc.bvl[sx+(X) + (sy+(Y))*rc.ar.width]
        #define BVL(X,Y) (BIX(X,Y)*ckern[((X)+1 + ((Y)+1)*CKERNEL_W)] != 0)
        #define PIX(X,Y)  bo.mem[tx+(X) + (ty+(Y))*bo.ar.width]
        #define R2M(X,Y) (m1.mem[sx+(X) + (sy+(Y))*m1.ar.width] == R2)
        #define R2K(X,Y) (R2M(X,Y) ? (PIX(X,Y)*ckern[((X)+1 + ((Y)+1)*CKERNEL_W)]):0)
	int sum = 0, bvl = 0;

        if (poi == BACKGROUND) break;                     /* of no interest */

	/* look 8-neighborhood for b-value which disallows extension        */
        bvl += BVL(-1,-1); bvl += BVL(0,-1); bvl += BVL(1,-1); 
        bvl += BVL(-1,0);  /**** POI ****/   bvl += BVL(1,0);
        bvl += BVL(-1,1);  bvl += BVL(0,1);  bvl += BVL(1,1); 
	bvl = XPMIN(bvl, 1);                    /* only existence asked for */

	/* look 8-neighborhood for superextension which disallows extension */
        sum += R2K(-1,-1);  sum += R2K(0,-1);  sum += R2K(1,-1); 
        sum += R2K(-1,0);   /* POI ignored */  sum += R2K(1,0);
        sum += R2K(-1,1);   sum += R2K(0,1);   sum += R2K(1,1); 

        if (!bvl) {                   /* no stop condition (b-value) nearby */
          copy.mem[cx + cyy] = poi;  /* add peripheral pixel on mask border */

 	  if (sum >= cluster_thold) {
 	    blood[cx + cyy] = 1;             /* stop-condition bloody pixel */
	  }
	}
      }
      break;

      case R2: break;        /* NB dilation grade 2, don't grow out of mask */
      case BACKGROUND: break;                         /* not to be remarked */
      default: assert(0); break;                      /* unresolved marking */
      }
    }
  }

  copy.bvl = blood;                                   /* take over B-values */

  free_pic_def(&m1);
  free_pic_def(&m2);
  free_pic_def(&rc);
  free_pic_def(&bp);
  free_pic_def(&bo);

  return copy;                       /* copy contains masked image clipping */

  #undef DSP
  #undef BIX
  #undef PIX
  #undef BVL
  #undef R2M
  #undef R2K
  #undef R1
  #undef R2
}



/* Marks the difference of two image clippings.

Input : - source
        - destination
        - clipping
        - offset
Output: - destination updated
        - number of different pixels
Postcondition: - equal pixels are erased
*/
int
mark_diff(pic_def src, pic_def dst, area_def cut, int sx, int sy, int dx, 
	  int dy, int mark)
{
  int s, z, found = 0;

  assert(src.mem);
  assert(dst.mem);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (z = 0; z < cut.height; z++) {
    for (s = 0; s < cut.width; s++) {
      if (dst.mem[s+dx + (z+dy)*dst.ar.width] != 
          src.mem[s+sx + (z+sy)*src.ar.width]) {
        dst.mem[s+dx + (z+dy)*dst.ar.width] = mark;     /* mark difference */
	found++;
      }
      else {
        dst.mem[s+dx + (z+dy)*dst.ar.width] = BACKGROUND;
      }
    }
  }

  return found;
}



/* Erases a masked image clipping.

   The intrinsic task is: how many pixels have to be erased to avoid the
   detection of the NB more than once, and how many pixels of the NB are
   allowed to be used from other references?

   Currently all NB pixels are deleted and all pixels of the mask which
   have not a 4-neighbourhood out of the mask.

Revision: 03.03.2002 E.Sack

Input : - masked binary reference
        - reference mask
        - RBS of image
        - image
        - original image
        - lag
Output: - updated image
*/
void 
erase_masked_area(pic_def ref, pic_def mask, sector_def rbs, pic_def pic,
		  pic_def org, int lx, int ly)
{
  int        sx, sy, sxmin, symin;                                /* mask  */
  int        t1x, t2x, t1y, t2y, tx, ty, tyy;                     /* ROI   */
  int        a1x, a2x, a1y, a2y;                                  /* image */
  pic_def    m, r;                  /* buffer to establish place on border */
  sector_def roi;

  #define NBH8_SPC 1
  #define MSK(X,Y) (m.mem[sx+(X) + (sy+(Y))*m.ar.width] == FOREGROUND)
  #define NB(X,Y) r.mem[sx+(X) + (sy+(Y))*r.ar.width]
  #define NBQ(X,Y) (NB(X,Y) == FOREGROUND)
  #define RP(X,Y) (tx+(X)<0 || ty+(Y)<0)? BACKGROUND:     \
                  org.mem[tx+(X) + (ty+(Y))*org.ar.width]
  #define RPQ(X,Y) (RP(X,Y) == FOREGROUND)

  init_pic_def(&m);
  init_pic_def(&r);

  roi.column.start = XPMAX(0, rbs.column.start+lx);   /* implicit with mask */
  roi.line.start = XPMAX(0, rbs.line.start+ly);
  roi.column.end = XPMIN(org.ar.width-1, rbs.column.start+lx+ref.ar.width-1);
  roi.line.end = XPMIN(org.ar.height-1, rbs.line.start+ly+ref.ar.height-1);

  a1x = roi.column.start; 
  a2x = roi.column.end;
  a1y = roi.line.start;
  a2y = roi.line.end;

  lx = rbs.column.start+lx-a1x;                             /* correct lags */
  ly = rbs.line.start+ly-a1y;

  m.ar.width = ref.ar.width + (NBH8_SPC<<1);              /* 8-neighborhood */
  m.ar.height = ref.ar.height + (NBH8_SPC<<1);            /* 8-neighborhood */
  CALLOC(PICT, m.ar.width*m.ar.height, m.mem);
  r.ar.width = ref.ar.width + (NBH8_SPC<<1);              /* 8-neighborhood */
  r.ar.height = ref.ar.height + (NBH8_SPC<<1);            /* 8-neighborhood */
  CALLOC(PICT, r.ar.width*r.ar.height, r.mem);

  copy_bmps(mask, m, ref.ar, 0, 0, NBH8_SPC, NBH8_SPC);   /* 8nbh at border */
  copy_bmps(ref, r, ref.ar, 0, 0, NBH8_SPC, NBH8_SPC);    /* 8nbh at border */

  t1x = XPMAX(0,lx) + a1x;
  t2x = XPMIN(a2x,lx+ref.ar.width-1 + a1x);
  t1y = XPMAX(0,ly) + a1y;
  t2y = XPMIN(a2y,ly+ref.ar.height-1 + a1y);

  sxmin = XPMAX(0,-lx) + NBH8_SPC;
  symin = XPMAX(0,-ly) + NBH8_SPC;

  for (ty = t1y, sy = symin; ty <= t2y; ty++, sy++) {
    tyy = ty*pic.ar.width;

    for (tx = t1x, sx = sxmin; tx <= t2x; tx++, sx++) {
      if (NBQ(0,0)) {
        pic.mem[tx + tyy] = BACKGROUND;                              /* poi */
	continue;
      }
      if (MSK(0,0) && !NBQ(0,0) &&
          (MSK(-NBH8_SPC,0) || NBQ(-NBH8_SPC,0) || !RPQ(-NBH8_SPC,0)) &&
          (MSK(+NBH8_SPC,0) || NBQ(+NBH8_SPC,0) || !RPQ(+NBH8_SPC,0)) &&
	  (MSK(0,-NBH8_SPC) || NBQ(0,-NBH8_SPC) || !RPQ(0,-NBH8_SPC)) &&
          (MSK(0,+NBH8_SPC) || NBQ(0,+NBH8_SPC) || !RPQ(0,+NBH8_SPC))) {
        pic.mem[tx + tyy] = BACKGROUND;                              /* poi */
	continue;
      }
    }
  }

  free_pic_def(&m);
  free_pic_def(&r);

  #undef NBH8_SPC
  #undef MSK
  #undef NB
  #undef NBQ
  #undef RP
  #undef RPQ
}



/* Merkmals-Mittelwert berechnen.

 Input: - Zeiger auf Vektor
        - Vektordimension
        - Nummer fuer unbercksichtigtes Element
Output: - Mittelwert
Postcondition: falls Nummer <0 oder >=Dimension: alle Elemente verwendet.
*/
cv_def_f
cv_def_f_mean_value(cv_def_f *in, int cnt, int skip)
{
  int     i, norm;
  cv_def_f mv = {0.0, 0.0, 0.0};

  assert(in);
  assert(cnt>0);

  for (i = 0; i < cnt; i++) {
    if (i == skip) continue;             /* Element bleibt unbercksichtigt */

    mv.cv += in[i].cv;
    mv.x += in[i].x;
    mv.y += in[i].y;
  }

  norm = skip < 0 || skip >= cnt ? cnt : cnt-1; 

  if (norm) {
    mv.cv /= (double)norm;
    mv.x /= (double)norm;
    mv.y /= (double)norm;
  }
  else {
    mv.cv = 0.0;
    mv.x = 0.0;
    mv.y = 0.0;
  }

  return mv;
}



/* Merkmals-Mittelwert berechnen.

 Input: - Zeiger auf Vektor
        - Vektordimension
        - Nummer fuer unbercksichtigtes Element
Output: - Mittelwert
Postcondition: falls Nummer <0 oder >=Dimension: alle Elemente verwendet.
*/
cv_def_f
cv_def_i_mean_value(cv_def_i *in, int cnt, int skip)
{
  int     i, norm;
  cv_def_f mv = {0.0, 0.0, 0.0};

  assert(in);
  assert(cnt>0);

  for (i = 0; i < cnt; i++) {
    if (i == skip) continue;             /* Element bleibt unbercksichtigt */

    mv.cv += in[i].cv;
    mv.x += (double)in[i].x;
    mv.y += (double)in[i].y;
  }

  norm = skip < 0 || skip >= cnt ? cnt : cnt-1; 

  if (norm) {
    mv.cv /= (double)norm;
    mv.x /= (double)norm;
    mv.y /= (double)norm;
  }
  else {
    mv.cv = 0.0;
    mv.x = 0.0;
    mv.y = 0.0;
  }

  return mv;
}



/* Erwartungstreuer Mittelwert einer Stichprobe bzw. Mittelwert einer
   Grundgesamtheit berechnen.

 Input: - Zeiger auf Array
        - Arraygre
        - unterdrckter Wert
        - Flag (0 - keinen Wert unterdrcken, 1 - Wert unterdrcken)
Output: - Mittelwert
*/
double
mean_value(int *in, int cnt, int suppress, int sval)
{
  int    i, rcnt;
  double mv;

  assert(in);
  assert(cnt > 0);

  for (i = 0, mv = 0.0, rcnt = 0; i < cnt; i++) {
    if (suppress == 1 && in[i] == sval) {
      continue;
    }

    mv += (double)in[i];
    rcnt++;
  }

  if (rcnt) {
    return mv/(double)rcnt;
  }
  
  return (double)sval;
}



/* Erwartungstreuer Mittelwert einer Stichprobe bzw. Mittelwert einer
   Grundgesamtheit berechnen.

 Input: - pointer to array
        - array size
Output: - mean value
*/
double
mean_value_double(double *in, int cnt)
{
  int    i;
  double mv;

  assert(in);
  assert(cnt > 0);

  for (i = 0, mv = 0.0; i < cnt; i++) {
    mv += in[i];                                              /* summation */
  }

  return mv/(double)cnt;
}



/* Merkmals-Standardabweichung berechnen.

 Input: - Zeiger auf Vektor
        - Vektordimension
        - Mittelwert
        - Nummer fuer unbercksichtigtes Element
Output: - Standardabweichung
Postcondition: falls Nummer <0 oder >=Dimension: alle Elemente verwendet.
*/
cv_def_f
cv_def_i_variance(cv_def_i *in, int cnt, cv_def_f mv, int skip)
{
  int     i, norm;
  cv_def_f var = {0.0, 0.0, 0.0};

  assert(in);
  assert(cnt>0);

  for (i = 0; i < cnt; i++) {
    if (i == skip) continue;             /* Element bleibt unbercksichtigt */

    var.cv += (in[i].cv-mv.cv) * (in[i].cv-mv.cv);
    var.x += ((double)in[i].x-mv.x) * ((double)in[i].x-mv.x);
    var.y += ((double)in[i].y-mv.y) * ((double)in[i].y-mv.y);
  }

  norm = skip < 0 || skip >= cnt ? cnt-1 : cnt-2; 

  if (norm) {
    var.cv = sqrt(var.cv/(double)(norm));
    var.x = sqrt(var.x/(double)(norm));
    var.y = sqrt(var.y/(double)(norm));
  }
  else {
    var.cv = 0.0;
    var.x = 0.0;
    var.y = 0.0;
  }

  return var;
}



/* Merkmals-Standardabweichung berechnen.

 Input: - Zeiger auf Vektor
        - Vektordimension
        - Mittelwert
        - Nummer fuer unbercksichtigtes Element
Output: - Standardabweichung
Postcondition: falls Nummer <0 oder >=Dimension: alle Elemente verwendet.
*/
cv_def_f
cv_def_f_variance(cv_def_f *in, int cnt, cv_def_f mv, int skip)
{
  int      i, norm;
  cv_def_f var = {0.0, 0.0, 0.0};

  assert(in);
  assert(cnt>0);

  for (i = 0; i < cnt; i++) {
    if (i == skip) continue;             /* Element bleibt unbercksichtigt */

    var.cv += (in[i].cv-mv.cv) * (in[i].cv-mv.cv);
    var.x += (in[i].x-mv.x) * (in[i].x-mv.x);
    var.y += (in[i].y-mv.y) * (in[i].y-mv.y);
  }

  norm = skip < 0 || skip >= cnt ? cnt-1 : cnt-2; 

  if (norm) {
    var.cv = sqrt(var.cv/(double)(norm));
    var.x = sqrt(var.x/(double)(norm));
    var.y = sqrt(var.y/(double)(norm));
  }
  else {
    var.cv = 0.0;
    var.x = 0.0;
    var.y = 0.0;
  }

  return var;
}



/* Erwartungstreue Standardabweichung einer Stichprobe aus einer
   Grundgesamtheit berechnen (Mittelwert geschtzt und erwartungstreu).

 Input: - Zeiger auf Array
        - Arraygre
        - Mittelwert
        - unterdrckter Wert
        - Flag (0 - keinen Wert unterdrcken, 1 - Wert unterdrcken)
Output: - Standardabweichung 
*/
double
standard_deviation(int *in, int cnt, double mv, int suppress, int sval)
{
  int    i, rcnt;
  double var;

  assert(in);
  assert(cnt);

  for (i = 0, var = 0.0, rcnt = 0; i < cnt; i++) {
    if (suppress == 1 && in[i] == sval) {
      continue;
    }

    var += ((double)in[i]-mv) * ((double)in[i]-mv);
    rcnt++;
  }

  if (rcnt <= 1) {
    return 0.0;                /* Ersatz: mittlere quadratische Abweichung */
  }

  return sqrt(var/(double)(rcnt-1));
}



/* L2-Norm (Euklidische 2D-Distanz) berechnen.

 Input: - Vektor
Output: - Norm
*/
double
euclid(double x, double y)
{
  return sqrt(x*x + y*y);
}



/* Mittelwert einer Verteilung berechnen.

Input:  - Verteilung
        - Schwelle
Output: - Mittelwert
*/
static double
expect_mean_value(double_lst dst, int bin)
{
  int    i;
  double sum;

  assert(bin <= dst.cnt);

  for (i = 0, sum = 0.0; i < bin; i++) {
    sum += (double)i * dst.dbl[i];
  }

  return sum;                               /* E(d) = sum[d(i)*H(d(i))]/N */
}



/* Verteilungsdichte einer Grundgesamtheit berechnen.

Input:  - Histogramm der Grundgesamtheit
Output: - Verteilungsdichte der Grundgesamtheit
Precondition: Die Histogramm-Struktur enthlt alle notwendigen Angaben.
*/
double_lst
distribution(hst_def hst)
{
  int        i;
  double_lst dst;

  assert(hst.set_cnt);

  init_double_lst(&dst);

  dst.cnt = hst.hst.cnt;
  CALLOC(double, dst.cnt, dst.dbl);

  for (i = 0; i < dst.cnt; i++) {
    dst.dbl[i] = hst.hst.dbl[i]/(double)hst.set_cnt;
  }

  return dst;                               /* p(d(i)) = H(d(i))/N */
}



/* Verteilungsfunktion einer Grundgesamtheit berechnen.

Input:  - Verteilungsdichte der Grundgesamtheit
Output: - Verteilungsfunktion der Grundgesamtheit
*/
static double_lst
distribution_func(double_lst dst)
{
  int        i;
  double_lst dstf;

  assert(dst.cnt);

  init_double_lst(&dstf);

  dstf.cnt = dst.cnt;
  CALLOC(double, dstf.cnt, dstf.dbl);

  for (i = 1, dstf.dbl[0] = dst.dbl[0]; i < dstf.cnt; i++) {
    dstf.dbl[i] = dstf.dbl[i-1] + dst.dbl[i];
  }

  return dstf;
}



/* Standardabweichung einer Grundgesamtheit berechnen,
   E([d-E(d)]^2) = sum[{d(i)-mv}^2*H(d(i))]/N, E(d) = Mittelwert.

Input:  - Verteilung der Grundgesamtheit
        - Mittelwert der Grundgesamtheit
Output: - Standardabweichung der Grundgesamtheit
Precondition: implizit Array von int 0-dst.cnt
*/
static double
expect_variance(double_lst dst, double mv)
{
  int    i;
  double z, sum;

  for (i = 0, sum = 0.0; i < dst.cnt; i++) {
    z = (double)i - mv;
    sum += z*z*dst.dbl[i];
  }

  return sum;
}



/* Kompatibilittsfunktion fr Symbolabstnde.

Input:  - Wert, Offset
Output: - Funktionswert
*/
static int
distance_compatibility(int dist, double dt)
{
  assert(dt>=0.0);

  if ((double)dist > dt) return 1;
  if ((double)dist == dt) return 0;
  else return -1;
}



/* Aktualisierungsoperator fr Symbolabstnde (#Abstnde = #Symbole - 1).

Input:  - Liste
Output: - Liste auf Stand
Precondition & Postcondition: 
 - Liste enthlt d(i,i+1) sortiert nach i = 0,1,2,...
 - Liste enthlt keine negativen Werte
*/
static void
distance_update(int_lst dlst, double dt)
{
  int d0, d1, i;

  assert(dlst.cnt);

  if (dlst.cnt == 1) {                                    /* #Symbole = 2 */
    dlst.nr[0] += distance_compatibility(dlst.nr[0] + INT_MIN+1, dt);
  }
  else {                                                  /* #Symbole > 2 */
    d0 = dlst.nr[0];
    dlst.nr[0] += distance_compatibility(d0 + INT_MIN+1, dt) +
                  distance_compatibility(d0-dlst.nr[1], dt);

    for (i = 1; i < dlst.cnt-1; i++) {                    /* #Symbole > 3 */
      d1 = dlst.nr[i];
      dlst.nr[i] += distance_compatibility(d1-d0, dt) +
                    distance_compatibility(d1-dlst.nr[i+1], dt);
      d0 = d1;
    }

    d1 = dlst.nr[dlst.cnt-1];
    dlst.nr[dlst.cnt-1] += 
          distance_compatibility(d1 + INT_MIN+1, dt) +
          distance_compatibility(d1-d0, dt);
  }

  for (i = 0; i < dlst.cnt; i++) {                        /* conditioning */
    dlst.nr[i] = XPMAX(dlst.nr[i],0); 
  }
}



/* Relaxation of a distance list.

Revision: 19.08.2002 E.Sack

Input:  - distance list
        - distance threshold
        - iteration number
Output: - relaxed distance list
*/
int_lst
relax_list(int_lst ilst, double dt, int iteration) 
{
  int     i;
  int_lst olst;

  init_int_lst(&olst);

  if (!ilst.cnt) {
    return olst;
  }

  copy_int_lst(&ilst, &olst);

  for (i = 0; i < iteration; i++) {
    distance_update(olst, dt);
  }

  return olst;
}



/* Ein Bild wird einer Dilatation unterworfen.
   Kernel:   111
             111
             111

Revision: 03.03.2002 E.Sack

Input : - Bild
Output: - Bild auf Stand
*/
void 
dilate_in_box(pic_def pic)
{
  int     d, z, s;
  pic_def zp;

  assert(pic.ar.width && pic.ar.height);

  init_pic_def(&zp);

  zp.ar.width = pic.ar.width + 2;
  zp.ar.height = pic.ar.height + 2;

  CALLOC(PICT, zp.ar.width*zp.ar.height, zp.mem);

  for (z = 0; z < pic.ar.height; z++) {
    for (s = 0; s < pic.ar.width; s++) {
      d = pic.mem[s + z*pic.ar.width];

      if (d == BACKGROUND) continue;

      zp.mem[s+1 + (z+1)*zp.ar.width] = d;  /* Mittelpunkt */
      zp.mem[s   + (z+1)*zp.ar.width] = d;
      zp.mem[s+2 + (z+1)*zp.ar.width] = d;
      zp.mem[s+1 + (z  )*zp.ar.width] = d;
      zp.mem[s+1 + (z+2)*zp.ar.width] = d;
      zp.mem[s   + (z  )*zp.ar.width] = d;
      zp.mem[s+2 + (z  )*zp.ar.width] = d;
      zp.mem[s   + (z+2)*zp.ar.width] = d;
      zp.mem[s+2 + (z+2)*zp.ar.width] = d;
    }
  }

  copy_bmps(zp, pic, pic.ar, 1, 1, 0, 0);

  FREEMEM(zp.mem);
}



static void
four1(double data[], ulong nn, int isign)
{
  ulong  n, mmax, m, j, istep, i;
  double wtemp, wr, wpr, wpi, wi, theta, tempr, tempi;

  n = nn << 1;
  j = 1;

  for (i = 1; i < n; i += 2) {
    if (j > i) {
      XPSWAP(tempr, data[j-1], data[i-1]);
      XPSWAP(tempr, data[j], data[i]);
    }

    m = n >> 1;

    while (m >= 2 && j > m) {
      j -= m;
      m >>= 1;
    }

    j += m;
  }

  mmax = 2;
  while (n > mmax) {
    istep = mmax << 1;
    theta = (double)isign*(CONST_2PI/(double)mmax);
    wtemp = sin(0.5*theta);
    wpr = -2.0*wtemp*wtemp;
    wpi = sin(theta);
    wr = 1.0;
    wi = 0.0;

    for (m = 1; m < mmax; m += 2) {
      for (i = m; i <= n; i += istep) {
	j = i + mmax;
	tempr = wr*data[j-1] - wi*data[j];
	tempi = wr*data[j] + wi*data[j-1];
	data[j-1] = data[i-1] - tempr;
	data[j]   = data[i] - tempi;
	data[i-1] += tempr;
	data[i]   += tempi;
      }

      wr = (wtemp=wr)*wpr - wi*wpi + wr;
      wi = wi*wpr + wtemp*wpi + wi;
    }

    mmax = istep;
  }
}



/* 1-dimensional (inverse) FFT of real values.

Input : - real 1D-Array
        - length of array
        - Flag (1: FFT, -1: iFFT)
Output: - complex 1D-Array
Precondition : - length of array MUST be power of 2.
               - 1D-Array with real numbers.
Postcondition: - Output of complex numbers of the FFT in the 1D-Array, real 
                 and imaginary parts in consecutive locations, except the
                 first.
*/
void
realft(double data[], ulong n, int isign) 
{
  ulong  i, i1, i2, i3, i4, np3;
  double c1 = 0.5, c2, h1r, h1i, h2r, h2i;
  double wr, wi, wpr, wpi, wtemp, theta;

  theta = CONST_2PI/2.0/(double)(n>>1);

  if (isign == 1) {
    c2 = -0.5;
    four1(data, n>>1, 1);
  }
  else {
    c2 = 0.5;
    theta = -theta;
  }

  wtemp = sin(0.5*theta);
  wpr = -2.0*wtemp*wtemp;
  wpi = sin(theta);
  wr = 1.0 + wpr;
  wi = wpi;
  np3 = n + 1;

  for (i = 2; i <= (n>>2); i++) {
    i4 = 1 + (i3=np3-(i2=1+(i1=i+i-2)));
    h1r = c1*(data[i1] + data[i3]);
    h1i = c1*(data[i2] - data[i4]);
    h2r = -c2*(data[i2] + data[i4]);
    h2i =  c2*(data[i1] - data[i3]);
    data[i1] = h1r + wr*h2r - wi*h2i;
    data[i2] = h1i + wr*h2i + wi*h2r;
    data[i3] = h1r - wr*h2r + wi*h2i;
    data[i4] = -h1i + wr*h2i + wi*h2r;
    wr = (wtemp=wr)*wpr - wi*wpi + wr;
    wi = wi*wpr + wtemp*wpi + wi;
  }

  if (isign == 1) {
    data[0] = (h1r=data[0]) + data[1];
    data[1] = h1r - data[1];
  }
  else {
    data[0] = c1*((h1r=data[0]) + data[1]);
    data[1] = c1*(h1r - data[1]);
    four1(data, n>>1, -1);
  }
}



/* 2-dimensional (inverse) FFT of complex values.

Input : - 2D-Array
        - Flag (1: FFT, -1: iFFT)
        - Array of length of Dimensions
Output: - 2D-Array updated
Precondition : - length of Dimensions MUST all be power of 2.
               - 2D-Array with real numbers, length of array twice the 
                 product of the length of the dimensions.
Postcondition: - Output of complex numbers of the FFT in the 2D-Array, real 
                 and imaginary parts in consecutive locations.
*/
static void
four2(double data[], ulong nn[], int isign)
{
  int    idim;
  ulong  i1, i2, i3, i2rev, i3rev, ip1, ip2, ip3, ifp1, ifp2;
  ulong  ibit, k1, k2, n, nprev, nrem, ntot;
  double tempi, tempr, theta, wi, wpi, wpr, wr, wtemp;

  assert(isign == 1 || isign == -1);

  ntot = nn[0]*nn[1];            /* compute total number of complex values */

  nprev = 1;

  for (idim = 1; idim >= 0; idim--) {     /* main loop over the dimensions */
    n = nn[idim];
    nrem = ntot/(n*nprev);
    ip1 = nprev << 1;

    nprev *= n;

    ip2 = ip1*n;
    ip3 = ip2*nrem;
    i2rev = 1;

    for (i2 = 1; i2 <= ip2; i2 += ip1) {                   /* bit-reversal */
      if (i2 < i2rev) {
	for (i1 = i2; i1 <= i2+ip1-2; i1 += 2) {
	  for (i3 = i1; i3 <= ip3; i3 += ip2) {
	    i3rev = i2rev + i3 - i2;
	    XPSWAP(tempr, data[i3-1], data[i3rev-1]);
	    XPSWAP(tempr, data[i3], data[i3rev]);
	  }
	}
      }

      ibit = ip2 >> 1;
      while (ibit >= ip1 && i2rev > ibit) {
	i2rev -= ibit;
	ibit >>= 1;
      }

      i2rev += ibit;
    }

    ifp1 = ip1; /* here begins the Danielson-Lanczos section of the routine */
    while (ifp1 < ip2) {
      ifp2 = ifp1 << 1;
      theta = (double)isign*CONST_2PI/(double)(ifp2/ip1);/* trig. recurrence*/
      wtemp = sin(0.5*theta);
      wpr = -2.0*wtemp*wtemp;
      wpi = sin(theta);
      wr = 1.0;
      wi = 0.0;

      for (i3 = 1; i3 <= ifp1; i3 += ip1) {
	for (i1 = i3; i1 <= i3+ip1-2; i1 += 2) {
	  for (i2 = i1; i2 <= ip3; i2 += ifp2) {
	    k1 = i2;
	    k2 = k1 + ifp1;
	    tempr = wr*data[k2-1] - wi*data[k2];
	    tempi = wr*data[k2]   + wi*data[k2-1];
	    data[k2-1] = data[k1-1] - tempr;
	    data[k2]   = data[k1]   - tempi;
	    data[k1-1] += tempr;
	    data[k1]   += tempi;
	  }
	}

	wr = (wtemp = wr)*wpr - wi*wpi + wr;   /* trigonometric recurrence */
	wi = wi*wpr + wtemp*wpi + wi;
      }

      ifp1 = ifp2;
    }
  }
}



/* 2-dimensional (inverse) FFT of real values (image X x Y). Y (height) 
   refers to the first frequency index, X (width) refers to the last 
   frequency index. Frequencies for X dimension are returned for the 
   positive half including the second array, frequencies for Y dimension
   are returned in wrap around order. See "Numerical Recipes in C" for
   details.

Input : - 2D-Array
        - 1D-Array
        - Length of Dimensions
        - Flag (1: FFT, -1: iFFT)
Output: - 2D- and 1D-Array updated
Precondition : - length of Dimensions MUST all be power of 2.
               - 2D-Array with real numbers.
               - 1D-Array allocated with 2*nn2
Postcondition: - Output of complex numbers of the FFT in the 2D-Array, real 
                 and imaginary parts in consecutive locations.
*/
void
rlft2(double **data, double *speq, ulong nn2, ulong nn3, int isign)
{
  ulong  i2 ,i3, j2, j3, nn[2], ii3;
  double theta, wi, wpi, wpr, wr, wtemp;
  double c1, c2, h1r, h1i, h2r, h2i;

  assert(!(nn2&1));                                        /* minimal test */
  assert(!(nn3&1));                                        /* minimal test */

  /* test of contiguity of data which is important                         */
  assert((ulong)(1+&data[nn2-1][nn3-1]-&data[0][0]) == nn2*nn3);

  c1 = 0.5;
  c2 = -0.5*(double)isign;
  theta = (double)isign*CONST_2PI/(double)nn3;
  wtemp = sin(0.5*theta);
  wpr = -2.0*wtemp*wtemp;
  wpi = sin(theta);
  nn[0] = nn2;
  nn[1] = nn3 >> 1;

  if (isign == 1) {
    four2(&data[0][0], nn, isign);
    for (i2 = 0, j2 = 0; i2 < nn2; i2++) {
      speq[j2++] = data[i2][0];
      speq[j2++] = data[i2][1];
    }
  }

  wr = 1.0;
  wi = 0.0;

  for (ii3 = 0, i3 = 1; i3 <= (nn3>>2)+1; i3++, ii3 += 2) {
    for (i2 = 0; i2 < nn2; i2++) {
      if (i3 == 1) {
	j2 = (i2 != 0 ? ((nn2-i2-1)<<1)+2 : 0);
	h1r =  c1*(data[i2][0] + speq[j2]);
	h1i =  c1*(data[i2][1] - speq[j2+1]);
	h2i =  c2*(data[i2][0] - speq[j2]);
	h2r = -c2*(data[i2][1] + speq[j2+1]);
	data[i2][0] = h1r + h2r;
	data[i2][1] = h1i + h2i;
	speq[j2]    = h1r - h2r;
	speq[j2+1]  = h2i - h1i;
      }
      else {
	j2 = (i2 != 0 ? nn2-i2 : 0);
	j3 = nn3 + 2 - (i3<<1);
	h1r =  c1*(data[i2][ii3]   + data[j2][j3]);
	h1i =  c1*(data[i2][ii3+1] - data[j2][j3+1]);
	h2i =  c2*(data[i2][ii3]   - data[j2][j3]);
	h2r = -c2*(data[i2][ii3+1] + data[j2][j3+1]);
	data[i2][ii3]   =  h1r + wr*h2r - wi*h2i;
	data[i2][ii3+1] =  h1i + wr*h2i + wi*h2r;
	data[j2][j3]    =  h1r - wr*h2r + wi*h2i;
	data[j2][j3+1]  = -h1i + wr*h2i + wi*h2r;
      }
    }

    wr = (wtemp = wr)*wpr - wi*wpi + wr;
    wi = wi*wpr + wtemp*wpi + wi;
  }

  if (isign == -1) {
    four2(&data[0][0], nn, isign);
  }
}



/* Multiplication of complex input A with complex conjugate input B.

Input : - complex input A
        - complex input B
Output: - complex input A updated
Precondition: - inputs may be identical
*/
static void
power_spectral_density_2D(fft_def fd1, fft_def fd2)
{
  int    j, width = fd1.ar.width, height = fd1.ar.height, len;
  double *sp1, *sp2, r, i;

  assert(fd1.ar.width == fd2.ar.width);
  assert(fd1.ar.height == fd2.ar.height);

  sp1 = &fd1.data[0][0];                  /* pointer simplifies loop usage */
  sp2 = &fd2.data[0][0];                

  len = height*(width>>1);          /* complex values needs double storage */

  for (j = 0; j < len; j++) {
    r = sp1[0]*sp2[0] + sp1[1]*sp2[1];                        /* real part */
    i = sp1[1]*sp2[0] - sp1[0]*sp2[1];                   /* imaginary part */
    sp1[0] = r;
    sp1[1] = i;
    sp1 += 2;           
    sp2 += 2;
  }

  sp1 = &fd1.spec[0];
  sp2 = &fd2.spec[0];

  for (j = 0; j < height; j++) {
    r = sp1[0]*sp2[0] + sp1[1]*sp2[1];
    i = sp1[1]*sp2[0] - sp1[0]*sp2[1];
    sp1[0] = r;
    sp1[1] = i;
    sp1 += 2;
    sp2 += 2;
  }
}



/* Renorming for correlation.

Input : - correlation data
Output: - correlation data updated
*/
static void
renorm_2D(fft_def fd)
{
  int    j, k;
  double r = 2.0/(double)fd.ar.height/(double)fd.ar.width;

  assert(fd.ar.height);
  assert(fd.ar.width);

  for (k = 0; k < fd.ar.height; k++) {
    for (j = 0; j < fd.ar.width; j++) {
      fd.data[k][j] *= r;                                     /* renorming */
    }
  }
}



/* Correlation by means of FFT and iFFT.

Input : - image
        - reference
        - flag (0: not FFT'ed, 1: already FFT'ed)
Output: - image updated
*/
void
corr2d(fft_def fd1, fft_def fd2, int pre_fd1_fft, int pre_fd2_fft)
{
  if (!pre_fd1_fft) {                                /* image to be FFT'ed */
    assert(!(!pre_fd2_fft && 
	   ((fd1.data == fd2.data) || (fd1.spec == fd2.spec))));
    rlft2(fd1.data, fd1.spec, (ulong)fd1.ar.height, (ulong)fd1.ar.width, 1);
  }

  if (!pre_fd2_fft) {                             /* referenz to be FFT'ed */
    assert(!(!pre_fd1_fft && 
	   ((fd1.data == fd2.data) || (fd1.spec == fd2.spec))));
    rlft2(fd2.data, fd2.spec, (ulong)fd2.ar.height, (ulong)fd2.ar.width, 1);
  }

  power_spectral_density_2D(fd1, fd2);                 /* frequency domain */

  /* iFFT */
  rlft2(fd1.data, fd1.spec, (ulong)fd1.ar.height, (ulong)fd1.ar.width, -1);

  renorm_2D(fd1);             /* end of correlation: renorming inverse FFT */
}



/* Power spectral density of a 2-D signal by means of FFT.

Input : - image
        - flag (0: not FFT'ed, 1: already FFT'ed)
Output: - image updated
*/
static void
power_density_auto(fft_def fd1, int pre_fd1_fft)
{
  if (!pre_fd1_fft) {                                /* image to be FFT'ed */
    rlft2(fd1.data, fd1.spec, (ulong)fd1.ar.height, (ulong)fd1.ar.width, 1);
  }

  power_spectral_density_2D(fd1, fd1);                 /* frequency domain */
}



/* Cross power spectral density of two 2-D signals by means of FFT.

Input : - image A and B
        - flag a and b (0: not FFT'ed, 1: already FFT'ed)
Output: - image A updated
*/
static void
power_density_cross(fft_def fd1, int pre_fd1_fft, 
		    fft_def fd2, int pre_fd2_fft)
{
  assert(fd1.ar.width == fd2.ar.width);     /* can be precalculated always */
  assert(fd1.ar.height == fd2.ar.height);

  if (!pre_fd1_fft) {                                /* image to be FFT'ed */
    rlft2(fd1.data, fd1.spec, (ulong)fd1.ar.height, (ulong)fd1.ar.width, 1);
  }

  if (!pre_fd2_fft) {                                /* image to be FFT'ed */
    rlft2(fd2.data, fd2.spec, (ulong)fd2.ar.height, (ulong)fd2.ar.width, 1);
  }

  power_spectral_density_2D(fd1, fd2);                 /* frequency domain */
}



/* Generates autocorrelation function of an real integer list of length S
   with all positive non-redundant S lags. The input list has to be zero-
   padded to at least 2S-1. The case of the Sth lag and of further lags is
   excluded because that values always are zero, e.g. S = 3:

   list:  a b c 0 0   (extension with zeros)
   lag=0  a b c 
   lag=1    a b c 
   lag=2      a b c 
   lag=3        a b c (unnecessary, because only correspondence 0 in the list)
   corr.: A B C 0

   As using the FFT algorithm one needs to extend the list with zeros up to 
   the power of two (zero padding), hence the list above becomes

   list:  a b c 0 0 0 0 0  (zero padding to N = 8)
   lag=0  a b c 
   lag=1    a b c 
   lag=2      a b c 
   corr.: A B C 0 0 0 0 0  (again, first S=3 lags are of interest).
                |-->... redundant if generated by the FFT

   The FFT algorithm is the only reason that more lags are computed that we
   really need. See also crosscorrelation function CCF.

Revision: 18.05.2003 E.Sack

Input : - integer list
        - pointer to maximal relevant index
Output: - autocorrelation
        - index updated
Postcondition: - lags greater than S-1 are redundant and should be ignored.
*/
fft_def
autocorrelation_of_double_lst(double_lst lst)
{
  int      j;
  fft_def  fm;
  area_def pow2;

  init_fft_def(&fm);

  XPOW2(2*lst.cnt-1, pow2.width);          /* all positive, reasonable lags */
  XPOW2(1, pow2.height);                   /* with zero padding             */

  fm = alloc_fft_def(pow2.width, pow2.height);

  for (j = 0; j < lst.cnt; j++) {
    fm.data[0][j] = lst.dbl[j];                             /* set up list */
  }

  power_density_auto(fm, 0);                 /* positive Fx of complex PSD */

  /* iFFT */
  rlft2(fm.data, fm.spec, (ulong)fm.ar.height, (ulong)fm.ar.width, -1);

  renorm_2D(fm);              /* end of correlation: renorming inverse FFT */

  return fm;
}



/* Calculates correlation features of signal A by signal B if FFT is used. 

Revision: 18.05.2003 E.Sack

Input : - signal A
        - signal B
        - CCF of A by B
        - pointer to features
Output: - features updated as flagged
        - flag (0: no error, 1: error occurred)
Postcondition: no redundant lags if redundant interval is empty.
*/
int
correlation_features_1D(double_lst alst, double_lst blst, fft_def corrf,
                        ival_def *positive_lag, ival_def *redundant_lag, 
			ival_def *negative_lag)
{
  int n_of_FFT = corrf.ar.width, alen = alst.cnt, blen = blst.cnt;

  if (corrf.ar.width < alen+blen-1) {
    return 1;
  }

  positive_lag->start = 0;
  positive_lag->end   = n_of_FFT - blen;

  redundant_lag->start = alen+blen;
  redundant_lag->end   = positive_lag->end;

  negative_lag->start = positive_lag->end + 1;
  negative_lag->end   = n_of_FFT - 1;

  return 0;
}



/* Generates crosscorrelation function CCF of signal A by signal B. 

Revision: 18.05.2003 E.Sack

Input : - signal A
        - signal B
Output: - crosscorrelation function with length N
Postcondition: lag = l   if 0 <= l <= lmax    (l periodic lag),
               lag = l-N if lmax < l <= N-1   (N different lags),
               lag = a+b <= l <= lmax         (redundant lags),
               (a+b-1 <= N, lmax = N-b, a,b signal length).
*/
fft_def
crosscorrelation_of_double_lst(double_lst alst, double_lst blst)
{
  int      j, width, dummy_height;
  fft_def  afm, bfm;

  init_fft_def(&afm);
  init_fft_def(&bfm);

  XPOW2(alst.cnt+blst.cnt-1, width);               /* minimal lmax is a+b-1 */
  XPOW2(1, dummy_height);                              /* with zero padding */

  afm = alloc_fft_def(width, dummy_height);     /* FFT needs equal lengths */
  bfm = alloc_fft_def(width, dummy_height);          /* allocate N samples */

  for (j = 0; j < alst.cnt; j++) {
    afm.data[0][j] = alst.dbl[j];             /* set up list for N samples */
  }

  for (j = 0; j < blst.cnt; j++) {
    bfm.data[0][j] = blst.dbl[j];             /* set up list for N samples */
  }

  power_density_cross(afm, 0, bfm, 0);
  free_fft_def(&bfm);                                         /* free asap */

  /* iFFT */
  rlft2(afm.data, afm.spec, (ulong)afm.ar.height, (ulong)afm.ar.width, -1);

  renorm_2D(afm);             /* correlation ends by renorming inverse FFT */

  return afm;
}



/* One-sided power spectral density (PSD) of a real number list with all 
   positive frequencies Fx. The complex PSD is sorted. The first half
   contains all real numbers, the second half contains all imaginary numbers
   of the complex value. The Nyquist frequency values are ignored.
   
Revision: 11.05.2003 E.Sack

Input : - real number list
Output: - PSD
*/
fft_def
psd_of_double_lst(double_lst lst)
{
  int      j, k;
  fft_def  fm;
  area_def pow2;

  init_fft_def(&fm);

  XPOW2(lst.cnt, pow2.width);
  XPOW2(1, pow2.height);                               /* with zero padding */

  fm = alloc_fft_def(pow2.width, pow2.height);

  for (j = 0; j < lst.cnt; j++) {
    fm.data[0][j] = lst.dbl[j];                              /* set up list */
  }

  power_density_auto(fm, 0);                  /* positive Fx of complex PSD */

  for (k = 0; k < fm.ar.height; k++) {          /* renorming for real input */
    for (j = 0; j < fm.ar.width; j++) {
      fm.data[k][j] *= 2.0;                      /* factor of one-sided PSD */
    }
  }

  /* separation of real and imaginary parts in consecutive order due to     */
  /* simplify storage model for complex numbers.                            */
  separate_complex_array(&fm.data[0][0], pow2.width);  /* no Nyquist frequ. */

  return fm;
}



/* Separates the real and imaginary parts of a complex array. The aliased 
   

Revision: 11.05.2003 E.Sack

Input : - pointer to array
        - array length
Output: - array updated
Precondition: - the complex numbers are stored in the input array with the
                real part followed by the imaginary part.
Postcondition: - the first half of the array contains the real part of all
                 complex numbers, the second half contains the complex part
                 of the numbers.
*/
static void
separate_complex_array(double *data, int length)
{
  int    j, half = length>>1, im_temp_len = (length>>1) - 1;
  double *im_temp = 0;

  assert(data);
  assert(length >= 0);
  assert(!(length&1));                              /* not a complex array */

  CALLOC(double, im_temp_len, im_temp);

  for (j = 0; j < im_temp_len; j++) {
    im_temp[j] = data[2*j+1];           /* save imaginary part except last */
  }

  for (j = 1; j <= im_temp_len; j++) {
    data[j] = data[2*j];                   /* store real part except first */
  }

  for (j = 0; j < im_temp_len; j++) {
    data[j+half] = im_temp[j];                  /* restore imaginary parts */
  }

  FREEMEM(im_temp);
}



/* Die Anzahl der Vordergrundpixel in einem Bildausschnitt werden gezhlt.

Input : - Bild
        - Ausschnitt
        - Offset
Output: - Anzahl Vordergrundpixel
*/
int
pic_mass(pic_def src, area_def cut, int sx, int sy) 
{
  int s, z, mass;

  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (z = 0, mass = 0; z < cut.height; z++) {
    for (s = 0; s < cut.width; s++) {
      if (src.mem[s+sx + (z+sy)*src.ar.width] == BACKGROUND) continue;
      else mass++;
    }
  }

  return mass;
}



/* Calculates from a grey-scaled image the histogram of the difference
   of true and false of the slopes of three pixel values evaluating 
   a < b < c. All 8 directions of a 8-neighborhood are considered.

Revision: 21.12.2004 E.Sack

Input : - image
        - maximal grey value
Output: - histogram
Precondition: - image size should be at least 3x3 pixels.  
Postcondition: - the histogram is guaranteed to be non-negative.
               - the threshold value is set, see Usage.
Usage: - the function is used to find an usable threshold of a grey-scaled
         image for binarisation.
       - the histogram built can be seen as the difference of two histograms.
*/
hst_def
slope_mean(pic_def pic, int mgv) 
{
  int     x, y, w = pic.ar.width, h = pic.ar.height;
  hst_def histo;
  double  mv, min;

  #define N(X,Y)  pic.mem[(X)   + ((Y)-1)*w]
  #define O(X,Y)  pic.mem[(X)   +  (Y)   *w]
  #define S(X,Y)  pic.mem[(X)   + ((Y)+1)*w]
  #define E(X,Y)  pic.mem[(X)+1 +  (Y)   *w]
  #define W(X,Y)  pic.mem[(X)-1 +  (Y)   *w]
  #define NE(X,Y) pic.mem[(X)+1 + ((Y)-1)*w]
  #define SW(X,Y) pic.mem[(X)-1 + ((Y)+1)*w]
  #define SE(X,Y) pic.mem[(X)+1 + ((Y)+1)*w]
  #define NW(X,Y) pic.mem[(X)-1 + ((Y)-1)*w]
  #define CHECK(X,Y,M)  (((X)<(M) && (M)<(Y)) || ((X)>(M) && (M)>(Y)))
  #define UPDOWN(X,Y,M) (CHECK(X,Y,M) ? 1.0:-1.0)
  #define COUNT(X,Y,M) mv = (double)((X) + (Y))/2.0;      \
		     histo.hst.dbl[(int)mv] += UPDOWN(X,Y,M) /* the formula */

  init_hst_def(&histo);
  histo.hst.cnt = mgv+1;
  CALLOC(double, histo.hst.cnt, histo.hst.dbl);

  /* the histogram */
  for (y = 1; y < h-1; y++) {               /* at least three pixels needed */
    for (x = 1; x < w-1; x++) {
      COUNT( N(x,y),  S(x,y), O(x,y));         /* 8-neighborhood calculated */
      COUNT( E(x,y),  W(x,y), O(x,y));
      COUNT(NE(x,y), SW(x,y), O(x,y));
      COUNT(NW(x,y), SE(x,y), O(x,y));
    }
  }

  min = histo.hst.dbl[0];                                           /* init */
  for (y = 1; y < histo.hst.cnt; y++) {
    min = XPMIN(min, histo.hst.dbl[y]);
  }

  if (min < 0.0) {           /* the algorithm delivers not a real histogram */
    for (y = 0; y < histo.hst.cnt; y++) {
      histo.hst.dbl[y] -= min;          /* a real histogram is non-negative */
    }
  }

  /* the histogram features */
  histo.th_bin = max_index_double_left(histo.hst.dbl, histo.hst.cnt);
  histo.max_bin = histo.th_bin;
  histo.maxval = histo.hst.dbl[histo.max_bin];
  histo.start_bin = 0;                       /* for reference grey value 0 */
  histo.set_cnt = AREA_SIZE(pic.ar);

  return histo;

  #undef O
  #undef N
  #undef NE
  #undef E
  #undef SE
  #undef S
  #undef SW
  #undef W
  #undef NW
  #undef UPDOWN
  #undef COUNT
  #undef CHECK
}



/* With the method of Otsu (J.R.PARKER S. 120) an attempt will be made to
   find the best level out of the histogram of the grey values to separate
   the dark pixels from the light ones.

Input : - pointer to distribution
Output: - histogram according to Otsu
*/
hst_def
threshold_otsu(double_lst dst) 
{
  int        t;
  double     *mv = 0, mvt, vat, x, y;
  double_lst dstf; 
  hst_def    histv;

  init_hst_def(&histv);

  histv.hst.cnt = dst.cnt;
  CALLOC(double, histv.hst.cnt, histv.hst.dbl);
  CALLOC(double, dst.cnt, mv);

  for (t = 0; t < dst.cnt; t++) {
    mv[t] = expect_mean_value(dst, t);                  /* leveled mean    */
  }

  mvt = mv[dst.cnt-1];                                  /* Global mean     */
  vat = expect_variance(dst, mvt);                      /* Global Variance */
  dstf = distribution_func(dst);                  /* distribution function */

  for (t = 0; t < dst.cnt; t++) {
    x = mvt*dstf.dbl[t] - mv[t];
    x = x*x;

    y = dstf.dbl[t]*(1.0-dstf.dbl[t]);
    if (y > 0) x = x/y/vat;
    else x = 0.0;

    histv.hst.dbl[t] = x;
  }

  FREEMEM(mv);
  free_double_lst(&dstf);

  return histv;
}



/* Computation of the sum of the squares of differences of two lists.

Input : - pointer to lists
Output: - sum
*/
double
sum_of_square(double *a, double *b, int cnt) 
{
  int    i;
  double diff, sum = 0.0;

  assert(a);
  assert(b);
  assert(cnt>0);

  for (i = 0; i < cnt; i++) {
    diff = a[i] - b[i];
    sum += diff*diff;
  }

  return sum;
}



/* Power of a digital, time discrete signal (theorem of Parseval). As
   the value range is big enaugh, the digital signal is as good as a 
   analog one.

Revision: 16.04.2003 E.Sack

Input : - signal
Output: - signal power 
*/
double
power_of_signal(double_lst signal) 
{
  int    i;
  double temp, sum = 0.0;

  if (!signal.dbl) {                                 /* consistency check */
    assert(!signal.cnt);
  }
  else {
    assert(signal.cnt > 0);
  }

  for (i = 0; i < signal.cnt; i++) {
    temp = signal.dbl[i];
    sum += temp*temp;
  }

  return sum;
}



/* Performs a morphological erosion operation on an image.

Revision: 10.06.2004 E.Sack

Input : - image
Output: - image updated
        - flag (0: no foreground pixel found, image unchanged,
                1: foreground pixel found,  image eroded)
*/
int
erode_in_box(pic_def pic)
{
  int     d, z, zz, s, found = 0;
  PICT    *ptr;
  pic_def zp1, zp2;

  assert(pic.ar.width && pic.ar.height);

  init_pic_def(&zp1);
  init_pic_def(&zp2);

  zp1.ar.width = pic.ar.width + 4;
  zp1.ar.height = pic.ar.height + 4;
  CALLOC(PICT, zp1.ar.width*zp1.ar.height, zp1.mem);

  ptr = zp1.mem;
  d = AREA_SIZE(zp1.ar);
  while (d--) {
    *ptr++ = FOREGROUND;                                           /* init */
  }

  zp2.ar.width = pic.ar.width + 2;
  zp2.ar.height = pic.ar.height + 2;
  CALLOC(PICT, AREA_SIZE(zp2.ar), zp2.mem);

  copy_bmps(pic, zp2, pic.ar, 0, 0, 1, 1);

  for (z = 0; z < zp2.ar.height; z++) {
    zz = z*zp2.ar.width;

    for (s = 0; s < zp2.ar.width; s++) {
      d = zp2.mem[s + zz];

      if (d != BACKGROUND) {
	continue;
      }

      zp1.mem[s+1 + (z+1)*zp1.ar.width] = d;                     /* center */
      zp1.mem[s   + (z+1)*zp1.ar.width] = d;
      zp1.mem[s+2 + (z+1)*zp1.ar.width] = d;
      zp1.mem[s+1 + (z  )*zp1.ar.width] = d;
      zp1.mem[s+1 + (z+2)*zp1.ar.width] = d;          
      zp1.mem[s   + (z  )*zp1.ar.width] = d;
      zp1.mem[s+2 + (z  )*zp1.ar.width] = d;
      zp1.mem[s   + (z+2)*zp1.ar.width] = d;
      zp1.mem[s+2 + (z+2)*zp1.ar.width] = d;

      found = 1;
    }
  }

  copy_bmps(zp1, pic, pic.ar, 2, 2, 0, 0);

  FREEMEM(zp1.mem);
  FREEMEM(zp2.mem);

  return found;
}



/* Scales down an image into a frame retaining aspect ratio.

Revision: 14.01.2005 E.Sack

Input : - source image
        - destination size
        - flag (TRUE: scale image respecting aspect else don't respect)
Output: - scaled down image
Postcondition: - if the destination size is greater than the source image, 
                 the source image is copied to the output image unscaled.
               - if the source image is empty, the output is empty as well.
*/
pic_def
scale_down_image(pic_def source_img, area_def dest_ar, gboolean has_aspect)
{
  int     i_x, i_y, i_yy, j_x, j_y, j_yy, istart_x, iend_x, istart_y, iend_y;
  double  red, green, blue, norm, aspect, fstep_x, fstep_y;
  double  fstart_x, fstart_y;
  col_def pxl;
  pic_def out;

  init_pic_def(&out);

  if (AREA_SIZE(source_img.ar) <= 0) {
    return out;                               /* no input buffer available */
  }

  if (AREA_SIZE(dest_ar) <= 0) {
    return out;                                 /* no output size declared */
  }

  assert(source_img.col);                          /* color image required */

  if (source_img.ar.width <= dest_ar.width && 
      source_img.ar.height <= dest_ar.height) {
    out = xdup_pic_def(source_img);
    return out;              /* size of source image is satisfying request */
  }

  aspect = (double)source_img.ar.width/(double)source_img.ar.height;

  if ((double)dest_ar.width <= (double)dest_ar.height * aspect) {
    out.ar.width = dest_ar.width;
    out.ar.height = (int)(((double)dest_ar.width)/aspect + 0.5);

    fstep_x = (double)source_img.ar.width/(double)dest_ar.width;
  }
  else {
    out.ar.width = (int)(((double)dest_ar.height)*aspect + 0.5);
    out.ar.height = dest_ar.height;
    fstep_x = (double)source_img.ar.height/(double)dest_ar.height;
  }

  if (has_aspect == TRUE) {
    fstep_y = fstep_x;    /* scale and retain aspect ratio of source image */
  }
  else {                                        /* just horizontal scaling */
    out.ar.height = dest_ar.height;
    fstep_y = (double)source_img.ar.height/(double)dest_ar.height;
  }

  CALLOC(col_def, AREA_SIZE(out.ar), out.col);              /* color image */

  fstart_y = 0.0;                                    /* whole source image */
  for (j_y = 0; j_y < out.ar.height; j_y++) { /* destination vertical line */
    j_yy = j_y*out.ar.width;

    istart_y = (int)(fstart_y + 0.5);
    iend_y = XPMIN((int)(fstart_y + fstep_y + 0.5), source_img.ar.height);

    fstart_x = 0.0;
    for (j_x = 0; j_x < out.ar.width; j_x++) {/* destination horizontal line */
      istart_x = (int)(fstart_x + 0.5);
      iend_x = XPMIN((int)(fstart_x + fstep_x + 0.5), source_img.ar.width);

      norm = red = green = blue = 0.0;                     /* set to black */
      for (i_y = istart_y; i_y < iend_y; i_y++) {                  /* tile */
	i_yy = i_y*source_img.ar.width;

	for (i_x = istart_x; i_x < iend_x; i_x++) {
	  pxl = source_img.col[i_x + i_yy];
	  red   += (double)pxl.r;
	  green += (double)pxl.g;
	  blue  += (double)pxl.b;
	  norm += 1.0;
	}
      }

      norm = 1.0/(norm == 0.0 ? 1.0 : norm);
      pxl.r = (PICT)(red   * norm);
      pxl.g = (PICT)(green * norm);
      pxl.b = (PICT)(blue  * norm);

      out.col[j_x + j_yy] = pxl;                         /* squeezed pixel */

      fstart_x += fstep_x;
    }

    fstart_y += fstep_y;
  }

  return out;
}



/* Scales down an image while putting the result in a prepared output image.

Revision: 14.01.2005 E.Sack

Input : - source image
        - destination image
        - source start coordinates
        - down-size value
Output: - destination image updated
Precondition:  - the output image can have an unconstrained size, however
                 must be consistent.
               - the down-size value v has to fulfil 0.0 < v <= 1.0
Postcondition: - if the destination image is greater than the result, the 
                 residual part of the destination image is not updated.
               - if the destination image is less than the result, the 
                 residual part of the result is ignored.
Usage: - This procedure is used to get a minimum of memory consumption when
         resizing an image. E.g. that means only displayed image data needs
         to be stored. The dynamic memory allocation is made outside. 
*/
void
xplab_scale_down_image(pic_def source_img, pic_def dest_img, 
		       gfloat src_start_x, gfloat src_start_y, 
		       gfloat magnitude)
{
  int     istart_x, iend_x, istart_y, iend_y, i_x, i_y, i_yy, j_x, j_y, j_yy;
  double  red, green, blue, norm, fstep, fstart_x, fstart_y;
  col_def pxl;

  if (AREA_SIZE(source_img.ar) <= 0) {
    return;                                   /* no input buffer available */
  }

  if (AREA_SIZE(dest_img.ar) <= 0) {
    return;                                   /* no output buffer prepared */
  }

  assert(source_img.col);
  assert(dest_img.col);
  assert(src_start_x>=0.0);
  assert(src_start_y>=0.0);
  assert(magnitude<=1.0);
  assert(magnitude>0.0);

  fstep = 1.0/magnitude;

  fstart_y = src_start_y;
  for (j_y = 0; j_y < dest_img.ar.height; j_y++) {
    j_yy = j_y*dest_img.ar.width;

    istart_y = (int)(fstart_y + 0.5);
    iend_y = XPMIN((int)(fstart_y + fstep + 0.5), source_img.ar.height);

    fstart_x = src_start_x;
    for (j_x = 0; j_x < dest_img.ar.width; j_x++) {
      istart_x = (int)(fstart_x + 0.5);
      iend_x = XPMIN((int)(fstart_x + fstep + 0.5), source_img.ar.width);

      norm = red = green = blue = 0.0;                     /* set to black */
      for (i_y = istart_y; i_y < iend_y; i_y++) {
	i_yy = i_y*source_img.ar.width;

	for (i_x = istart_x; i_x < iend_x; i_x++) {
	  pxl = source_img.col[i_x+i_yy];
	  red   += (double)pxl.r;
	  green += (double)pxl.g;
	  blue  += (double)pxl.b;
	  norm += 1.0;
	}
      }

      if (norm == 0.0) {
	break;         /* leave rest of destination image background unset */
      }

      norm = 1.0/norm;
      pxl.r = (PICT)(red   * norm);
      pxl.g = (PICT)(green * norm);
      pxl.b = (PICT)(blue  * norm);

      dest_img.col[j_x + j_yy] = pxl;

      fstart_x += fstep;
    }

    fstart_y += fstep;
  }
}



/* Scales up an image while putting the result in a prepared output image.

Revision: 11.12.2004 E.Sack

Input : - source image
        - destination image
        - source start coordinates
        - up-size value
        - visible area values
Output: - destination image updated
Postcondition: - the displayed drawing is very sensitive to the visible
                 area values.
Usage: - see counterpart xplab_scale_down_image(). 
*/
void
xplab_scale_up_image(pic_def source_img, pic_def dest_img, 
		     gfloat src_start_x, gfloat src_start_y, 
		     gfloat magnitude, int netto_x, int netto_y)
{
  int     yy;
  gfloat  x, y, mx, my, sx, sy;
  col_def pxl;

  sx = src_start_x;
  sy = src_start_y;
  x = y = 0.0;

  while (1) {
    if (sx >= (gfloat)source_img.ar.width || 
	sy >= (gfloat)source_img.ar.height) {
      return;                                           /* not yet updated */
    }

    pxl = source_img.col[(int)sx+(int)sy*source_img.ar.width];

    mx = x;
    my = y;                                               /* fill a square */
    while (y < my+magnitude && (int)y < netto_y) {
      yy = (int)y*dest_img.ar.width;
      x = mx;
      while (x < mx+magnitude && (int)x < netto_x) {
	dest_img.col[(int)x+yy] = pxl;
	x += 1.0;
      }
      y += 1.0;
    }                                                     /* square filled */

    if ((int)x >= netto_x && (int)y >= netto_y) {
      break;                                   /* very last square plotted */
    }

    if ((int)x >= netto_x) {                   /* last square in x plotted */
      x = 0.0;                                         /* next square in y */
      sx = src_start_x;
      sy += 1.0;
    }
    else {                                             /* next square in x */
      sx += 1.0;
      y = my;
    }
  }
}



/* Generation of a histogram of an array.

Revision: 16.12.2004 E.Sack

Input : - pointer to integer array
        - array size
Output: - histogram
Postcondition: - histogram values starts with the smallest value of the array.
*/
hst_def
histogram_auto(int *val, int cnt)
{
  int     i, min, max;
  double  sum;
  hst_def histo;

  assert(cnt > 0);

  init_hst_def(&histo);

  min = val[0];                                                    /* init */
  max = val[0];

  for (i = 1; i < cnt; i++) {
    min = XPMIN(val[i], min);
    max = XPMAX(val[i], max);
  }

  histo.start_bin = min;                      /* histogram start reference */
  histo.hst.cnt = max - min + 1;              /* histogram size            */
  histo.set_cnt = cnt;                        /* sample size               */

  assert(histo.hst.cnt > 0);

  CALLOC(double, histo.hst.cnt, histo.hst.dbl);

  for (i = 0; i < cnt; i++) {                           /* build histogram */
    histo.hst.dbl[val[i]-min]++;                           /* fit to array */
  }

  for (i = 0, histo.maxval = 0.0, sum = 0.0; i < histo.hst.cnt; i++) {
    if (histo.maxval < histo.hst.dbl[i]) {
      histo.maxval  = histo.hst.dbl[i];
      histo.max_bin = i;
    }

    sum += histo.hst.dbl[i];                      /* integral of histogram */
  }

  assert(sum == (double)histo.set_cnt);   /* integral equal to sample size */

  return histo;
}



/* Calculates the histogram of an 8bit-type buffer.

Revision: 10.12.2004 E.Sack

Input : - pointer to buffer
        - buffer size
Output: - histogram
*/
hst_def
histogram_8bit(PICT *img_buf, area_def bsize) 
{
  int     i = 0;
  double  sum;
  hst_def histo;

  assert(img_buf);

  init_hst_def(&histo);

  histo.hst.cnt = PICT_MAX + 1;                     /* 8bit histogram size */
  CALLOC(double, histo.hst.cnt, histo.hst.dbl);
  assert(histo.hst.dbl[0] == 0.0);

  histo.set_cnt = AREA_SIZE(bsize);                         /* sample size */
  histo.start_bin = 0;                        /* histogram start reference */

  for (i = 0; i < histo.set_cnt; i++) {
    histo.hst.dbl[img_buf[i]]++;                        /* build histogram */
  }

  for (i = 0, histo.maxval = 0.0, sum = 0.0; i < histo.hst.cnt; i++) {
    if (histo.maxval < histo.hst.dbl[i]) {
      histo.maxval = histo.hst.dbl[i];
      histo.max_bin  = i;
    }

    sum += histo.hst.dbl[i];                      /* integral of histogram */
  }

  assert(sum == (double)histo.set_cnt);   /* integral equal to sample size */

  return histo;
}



/* Calculates the histogram of an 24bit-type buffer.

Revision: 15.12.2004 E.Sack

Input : - pointer to buffer
        - buffer size
        - pointer to histograms
Output: - histograms updated
*/
void
histogram_rgb(col_def *img_buf, area_def bsize, hst_def *hred, 
	      hst_def *hgreen, hst_def *hblue) 
{
  int    i = 0;
  double sum;

  assert(img_buf);
  assert(hred);
  assert(hgreen);
  assert(hblue);

  free_hst_def(hred);
  free_hst_def(hgreen);
  free_hst_def(hblue);

  hred->hst.cnt = PICT_MAX + 1;                     /* 8bit histogram size */
  CALLOC(double, hred->hst.cnt, hred->hst.dbl);
  assert(hred->hst.dbl[0] == 0.0);
  hred->set_cnt = AREA_SIZE(bsize);                         /* sample size */
  hred->start_bin = 0;                        /* histogram start reference */

  hgreen->hst.cnt = PICT_MAX + 1;
  CALLOC(double, hgreen->hst.cnt, hgreen->hst.dbl);
  assert(hgreen->hst.dbl[0] == 0.0);
  hgreen->set_cnt = AREA_SIZE(bsize);
  hgreen->start_bin = 0;                      /* histogram start reference */

  hblue->hst.cnt = PICT_MAX + 1;
  CALLOC(double, hblue->hst.cnt, hblue->hst.dbl);
  assert(hblue->hst.dbl[0] == 0.0);
  hblue->set_cnt = AREA_SIZE(bsize);
  hblue->start_bin = 0;                       /* histogram start reference */

  for (i = 0; i < hred->set_cnt; i++) {
    hred->  hst.dbl[img_buf[i].r]++;                   /* build histograms */
    hgreen->hst.dbl[img_buf[i].g]++;
    hblue-> hst.dbl[img_buf[i].b]++;
  }

  for (i = 0, hred->maxval = 0.0, sum = 0.0; i < hred->hst.cnt; i++) {
    if (hred->maxval < hred->hst.dbl[i]) {
      hred->maxval = hred->hst.dbl[i];
      hred->max_bin  = i;
    }

    sum += hred->hst.dbl[i];                      /* integral of histogram */
  }

  assert(sum == (double)hred->set_cnt);   /* integral equal to sample size */

  for (i = 0, hgreen->maxval = 0.0, sum = 0.0; i < hgreen->hst.cnt; i++) {
    if (hgreen->maxval < hgreen->hst.dbl[i]) {
      hgreen->maxval = hgreen->hst.dbl[i];
      hgreen->max_bin  = i;
    }

    sum += hgreen->hst.dbl[i];                    /* integral of histogram */
  }

  assert(sum == (double)hgreen->set_cnt); /* integral equal to sample size */

  for (i = 0, hblue->maxval = 0.0, sum = 0.0; i < hblue->hst.cnt; i++) {
    if (hblue->maxval < hblue->hst.dbl[i]) {
      hblue->maxval = hblue->hst.dbl[i];
      hblue->max_bin  = i;
    }

    sum += hblue->hst.dbl[i];                     /* integral of histogram */
  }

  assert(sum == (double)hblue->set_cnt);  /* integral equal to sample size */
}
