/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_preferences_fe.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_batch.h"
#include "xplab_dchannel.h"
#include "xplab_recall_fe.h"
#include "xplab_cluster.h"
#include "xplab_dbio.h"

/* local prototypes ********************************************************/
static void read_cli_batch(int argc, char *argv[], prog_ctrl *ctrl);
static void open_batch_log(batch_def *batch, int verbos);
static void close_batch_log(FILE *log_stream, int verbs);

/* Functions **************************************************************/

/* Converts command line arguments to XPLAB arguments.
   To be honest, this function is only to show, how to fetch arguments in
   this phase.

Revision: 09.04.2004 E.Sack

Input : - CLI argument count
        - CLI argument vector
        - pointer to control structure
Output: - control structure updated
        - phase results
*/
static void
read_cli_batch(int argc, char *argv[], prog_ctrl *ctrl) 
{
  int c;

  optind = 0;                                              /* reset getopt */

  while (1) {
    c = getopt(argc, argv, BATCH_OPTION_LIST GUI_OPTION_LIST);

    if (c == -1) {
      break;
    }

    switch (c) {
    case 'F': ctrl->cfg.cdomain = 1; break;/* hidden feature for debugging purposes */
    default: break;
    }
  }
}



/* Main routine of XPLAB. 

Revision: 23.01.2004 E.Sack

Input : - CLI argument count
        - CLI argument vector
        - pointer to control structure
Output: - recognition results
*/
void
xplab_batch(int argc, char *argv[], prog_ctrl *bctrl)
{
  char   *tbuf = NULL;
  time_t start_time = time((time_t *)NULL);

  open_batch_log(&xctrl.cfg.batch, xctrl.cfg.verbose > BE_QUIET);

  read_cli_batch(argc, argv, bctrl);

  switch (bctrl->cfg.batch.phase) {
  case BATCH_PHASE_RB_CHAIN:                       /* processing RB chains */
    cb_recall_start_dump(NULL, NULL);
    break;
      
  case BATCH_PHASE_MATCHING:                                   /* matching */
    cb_recall_start_matching(NULL, NULL);
    break;
      
  case BATCH_PHASE_CLUSTER:                                  /* clustering */
    cb_clustering_start(NULL, NULL);
    break;
      
  case BATCH_PHASE_NO:
    break;

  default: assert(0); break;
  }

  bctrl->cfg.batch.phase = BATCH_PHASE_NO;

  if (bctrl->cfg.duration) {
    tbuf = print_time_difference(start_time);
    fprintf(stdout, "%s\n", tbuf);
  }

  string_to_log(NULL, "\n", "", "");                        /* deflate log */

  FREEMEM(tbuf);
  close_batch_log(xctrl.cfg.batch.log_stream, xctrl.cfg.verbose > BE_QUIET);
}



/* Opens batch log file.

Revision: 08.04.2004 E.Sack

Input : - batch job settings
        - verbose flag
Output: - settings updated
        - log opened as flagged
*/
static void
open_batch_log(batch_def *batch, int verbose)
{
  int  err;

  if (batch->blog == BATCH_LOG_NO) {
    return;                                               /* no log wanted */
  }

  err = check_name(BATCHNAME_DEFAULT);

  if (err) {
    return;                                                /* no file name */
  }

  if (batch->log_stream != NULL) {
    return;                                            /* open stream once */
  }

  batch->log_stream = xplab_open_file(BATCHNAME_DEFAULT, 
				      batch->blog==BATCH_LOG_APPND?"a":"w");

  if (batch->log_stream) {
    if (verbose) {
      fprintf(stdout, "batch log file '" BATCHNAME_DEFAULT "' opened\n");
    }
  }
}



/* Closes the batch log.

Revision: 08.04.2004 E.Sack

Input : - pointer to stream
        - verbose flag
Output: - stream closed 
*/
static void
close_batch_log(FILE *log_stream, int verbose)
{ 
  int clos;

  if (log_stream == NULL) {
    return;
  }

  clos = xplab_close_file(log_stream, BATCHNAME_DEFAULT);
  log_stream = NULL;

  if (verbose) {
    if (!clos) {
      fprintf(stdout, "batch log file '" BATCHNAME_DEFAULT "' closed\n");
    }
    else {
      fprintf(stdout, 
	      "WARNING: cannot close batch log file '" BATCHNAME_DEFAULT "'\n");
    }
  }
}



/* Writes to a stream.

Revision: 08.04.2004 E.Sack

Input : - pointer to stream
        - pointer to string
Output: - stream updated as flagged
        - flag (0: no stream defined, 1: stream defined)
*/
int
write_to_batch_log(FILE *log_stream, const char *string)
{
  if (log_stream == NULL) {
    return 0;
  }

  if (string) {
    fprintf(log_stream, "%s", string);
  }

  return 1;
}



