/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_recall_be.h"
#include "xplab_maintenance_be.h"
#include "xplab_dbio.h"
#include "xplab_image.h"
#include "xplab_output.h"
#include "xplab_training_fe.h"
#include "xplab_preferences_fe.h"
#include "xplab_preferences_be.h"
#include "xplab_cluster.h"
#include "xplab_selection.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
#define XPLAB_CLUSTERING_TITLE_ITEM "clustering"
#define XPLAB_CLUSTERING_WINDOW_WIDTH   300
#define XPLAB_CLUSTERING_WINDOW_HEIGHT  0

typedef struct {
  GtkWidget *top_clustering, *iteration_label, **caller;
  gint      db_key;
  gboolean  cactive, cstop;
  pic_def   cimg;
  prog_ctrl ctrl;
  run_def   crun, irun;
} clust_def;

/* local prototypes ********************************************************/
static void init_clust_def(clust_def *cl);
static void free_clust_def(clust_def *cl);

static cls_lst xcluster_generation(cand_lst hit, pic_def rpic, pic_def rorg, 
				   cls_lst old, hash_def *hash, char *filename, 
				   CLUSTER_NB_BIT p_cluster, int cluster_write, 
				   run_def *crun, kern_def kernel, 
				   int cluster_thold, int *cfileno);
static cls_lst xpattern_cluster(match_def mat, pic_def rpic, pic_def rorg, rb_lst *rblst, 
				cls_lst old, hash_lst *gen, char *filename,
				int fdomain, CLUSTER_NB_BIT p_cluster,
				int cluster_write, kern_def kernel, 
				int cluster_thold, int single_line, 
				int *cfileno, int single_nb);
static gen_lst build_pedigrees(hash_lst cluster, cls_lst cl);
static gen_def build_generation(hash_lst gen, cls_lst cl, int cid, int pid,
				cand_def ckr);
static void add_hash_lst(hash_lst *gen, hash_def hash);
static void write_generation_rpt(cls_lst old, cls_lst new);
static void write_cluster_rpt(hash_lst gen, cls_lst cl, char *input_name,
			      char *db_name);
static int untrusty_candidate(cand_lst hit);
static int classify_twins(const char *cname, ALIGMNT zline, nb_def nb, cls_lst cl);
static int part_count(pic_def pic, sector_def fr, int invtd);
static int compare_bvl(nb_def nb1, nb_def nb2);
static int classify_bvl(nb_def src, nb_def *dst, int strong);
static void xclustering(match_def mat, db_def db_old, int tps_old, 
			rb_lst *rblst, pic_def rpic, char *filename, 
			int fdomain, CLUSTER_NB_BIT p_cluster, 
			double cluster_level, double cluster_step, 
			int cluster_force, int cluster_write, 
			kern_def kernel, int cluster_thold, int single_line,
			int single_nb);
static int check_cluster_resources(db_def db, char *tps_id, 
				   char *image_filename, pic_def image, 
				   rb_lst rblst);

/* Functions assembling widgets */
static void clustering_setup(prog_ctrl ctrl);
static void clustering_window(gdouble corrlevel);
static void clustering_menu_bar(GtkWidget *vbox);
static void clustering_menu_entry_file(GtkWidget *menu_bar);
static void clustering_iteration(GtkWidget *vbox, gdouble corrlevel);
static void clustering_control(GtkWidget *vbox);
static void clustering_file_menu_item_quit(GtkWidget *file_menu);

/* Handlers/Callbacks */
static void cb_clustering_stop(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_clustering_phase(GTK_CONVENTION_MENUITEM_RTS_VOID);

/* local data **************************************************************/

static clust_def cl;                               /* clustering structure */

/* Functions ***************************************************************/

static void
init_clust_def(clust_def *cl)
{
  assert(cl);

  cl->top_clustering = NULL;                                 /* zero pointer */
  cl->iteration_label = NULL;                                /* zero pointer */
  cl->caller = NULL;                                         /* zero pointer */

  cl->db_key = 0;
  cl->cactive = cl->cstop = FALSE;

  init_pic_def(&cl->cimg);
  init_prog_ctrl(&cl->ctrl);

  init_run_def(&cl->crun);
  init_run_def(&cl->irun);
}

static void
free_clust_def(clust_def *cl)
{
  assert(cl);
                           /* don't free GTK stuff */

  free_pic_def(&cl->cimg);
  free_prog_ctrl(&cl->ctrl);

  init_clust_def(cl);
}



/* Initializes the clustering phase.

Revision: 02.05.2004 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_clustering(void)
{
  init_clust_def(&cl);

  cl.crun.pactive = &cl.cactive;
  cl.crun.pstop = &cl.cstop;
}



/* Frees the clustering phase.

Revision: 28.11.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_free_clustering(void)
{
  free_clust_def(&cl);

  xplab_init_clustering();
}



/* Cluster composition of similar pattern in a document.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - DB
        - TPS index
        - pointer to RB list
        - RB image
        - pointer to image filename
        - domain
        - cluster flag
        - minimal correlation value for clustering
        - correlation value decreasing step
        - flag for decreasing correlation value
        - flag for writing clustered images
        - clustering kernel
        - clustering extension level
        - flag (0: multi line, 1: single line)
        - flag (0: multi NB, 1: single NB)
Output: - new DB dumped
        - clustered images dumped as flagged
Precondition: - RB list sorted in descending order of size.  
*/
static void
xclustering(match_def mat, db_def db_old, int tps_old, rb_lst *rblst, 
	    pic_def rpic, char *filename, int fdomain, 
	    CLUSTER_NB_BIT p_cluster, double cluster_level, 
	    double cluster_step, int cluster_force, int cluster_write, 
	    kern_def kernel, int cluster_thold, int single_line, 
	    int single_nb)
{
  int      err, tps_all, tps_new = -1, iteration = 0, cfileno = 0;
  db_def   db_all, db_new;
  cls_lst  cl_new;
  char     *ident = NULL, *buf = NULL;
  double   cl_old = mat.corrlevel;
  hash_lst gen;
  pic_def  rorg;                                       /* contains original */

  init_pic_def(&rorg);
  init_hash_lst(&gen);
  init_cls_lst(&cl_new);
  init_db_def(&db_all);
  init_db_def(&db_new);

  if (xctrl.cfg.verbose > VERBOSE_DEFAULT) {
    string_to_log(main_log, "clustering using TPS ID '", db_old.tpl.tps[tps_old].ident, "' ");
    string_to_log(main_log, "in DB '", db_old.db_fname, "'\n");
    double_to_log("threshold ", mat.corrlevel, "\n");
  }

  if (cluster_force) {
    int_to_label(cl.iteration_label, "iteration ", iteration, "");
  }
  else {
    xplab_label_set_text(cl.iteration_label, "(no decreasing CV threshold)");
  }

  xreport("threshold %f\n", mat.corrlevel);

  /* start and initializing clustering algorithm                           */
  rorg.ar = rpic.ar;
  CALLOC(PICT, AREA_SIZE(rorg.ar), rorg.mem);
  copy_bmps(rpic, rorg, rorg.ar, 0, 0, 0, 0);

  tps_all = db_all.tpl.cnt++;
  CALLOC(tps_def, db_all.tpl.cnt, db_all.tpl.tps);                 /* init */
  copy_tps_def(&db_old.tpl.tps[tps_old], &db_all.tpl.tps[tps_all], FALSE);

  add_string(db_old.tpl.tps[tps_old].ident, &ident);
  tps_new = create_tps_def(&db_new.tpl, &ident);
  if (tps_new < 0) {                               /* something went wrong */
    free_pic_def(&rorg);
    free_db_def(&db_all);
    free_db_def(&db_new);
    return;
  }

  assert(tps_new==0);
  assert(db_new.tpl.cnt==1);
  err = strcmp(db_old.tpl.tps[tps_old].ident, ident);
  assert(!err);
  FREEMEM(ident);

  copy_media_att(db_old.tpl.tps[tps_old].media, 
		 &db_new.tpl.tps[tps_new].media);

  gen_class_references(db_all.tpl.tps[tps_all].cl, mat.mborder, 
		       mat.maskweight);

  cl_new = xpattern_cluster(mat, rpic, rorg, rblst, 
			    db_all.tpl.tps[tps_all].cl,
			    &gen, filename, fdomain, p_cluster,
			    cluster_write, kernel, cluster_thold, single_line,
			    &cfileno, single_nb);

  while (*cl.crun.pstop == FALSE) {
    if (cl_new.cnt) {                                    /* growing cluster */
      cat_cls_lst(cl_new, &db_new.tpl.tps[tps_new].cl);      /* only new NB */
      cat_cls_lst(cl_new, &db_all.tpl.tps[tps_all].cl);           /* all NB */

      free_cls_lst(&cl_new);

      if (!rblst->cnt) {
	break;                                  /* no more pattern existing */
      }
    }
    else if (1.0e-16 >= cluster_level + cluster_step - mat.corrlevel &&
	     cluster_force) {
      mat.corrlevel -= cluster_step;                      /* next threshold */
    }
    else {
      break;
    }

    iteration++;

    int_to_label(cl.iteration_label, "iteration ", iteration, "");
    xplab_progress_set_value(&cl.irun, (gdouble)mat.corrlevel, 1.0);
    xreport("threshold %f\n", mat.corrlevel);

    gen_class_references(db_all.tpl.tps[tps_all].cl, mat.mborder, 
			 mat.maskweight);
    cl_new = xpattern_cluster(mat, rpic, rorg, rblst, 
			      db_all.tpl.tps[tps_all].cl, 
			      &gen, filename, fdomain, p_cluster,
			      cluster_write, kernel, cluster_thold,
			      single_line, &cfileno, single_nb);
  }

  mat.corrlevel = cl_old;                                /* reset original */

  xreport("%d iteration%s\n", iteration, iteration==1?"":"s");
  if (!gen.cnt) {
    xreport("*** no cluster found ***\n\n");
  }

  cat_string(filename, CLUSTER_DEFAULT, "", &buf);

  dump_pic_def(rpic, buf, rpic.ar, 0, 0,          /* write clustered input */
	       xctrl.cfg.verbose > BE_QUIET, PNM_FORMAT_UNDEF);
  FREEMEM(buf);

  cat_string(db_old.db_fname, ".new", "", &db_new.db_fname);
  write_db_file(db_new);                                    /* dump new DB */

  cat_string(db_old.db_fname, ".all", "", &db_all.db_fname);
  write_db_file(db_all);                                  /* dump total DB */

  write_cluster_rpt(gen, db_all.tpl.tps[tps_all].cl, filename, 
		    db_all.db_fname);
  write_generation_rpt(db_old.tpl.tps[tps_old].cl, 
		       db_new.tpl.tps[tps_new].cl);

  free_pic_def(&rorg);
  free_db_def(&db_all);
  free_db_def(&db_new);
  free_hash_lst(&gen);
}



/* Builds with recognized NBs a new class list using an old one. The old
   list represents the parents, the new list contains their children. To
   get the reference between parents and children a so-called hash list is
   carried along with it. To crop out new NBs after recognition two images
   are needed, the unchanged original RB image and a cropped working one.
   After cropping pattern a new RB list is built again.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - working RB image
        - original RB image
        - pointer to RB list
        - old class list
        - pointer to generation hash list
        - pointer to image file name
        - domain
        - cluster flag
        - flag for writing clustered images
        - clustering kernel
        - clustering extension level
        - flag (0: multi line, 1: single line)
        - pointer to filename number extension
        - flag (0: multi NB, 1: single NB)
Output: - new class list
        - RB list updated
Precondition: - old class list contains NB references
*/
static cls_lst
xpattern_cluster(match_def mat, pic_def rpic, pic_def rorg, rb_lst *rblst, 
		 cls_lst old, hash_lst *gen, char *filname, int fdomain,
		 CLUSTER_NB_BIT p_cluster, int cluster_write, 
		 kern_def kernel, int cluster_thold, int single_line, 
		 int *cfileno, int single_nb)
{
  cand_lst hit_lst;
  cls_lst  new;                                          /* new class list */
  hash_def hash;                            /* to track cluster generation */

  assert(old.cnt);

  init_cls_lst(&new);
  init_hash_def(&hash);

  hit_lst = xpattern_matching(mat, single_line, *rblst, old, &rpic, fdomain,
			      &cl.crun, single_nb, 0);

  if (hit_lst.cnt && *cl.crun.pstop == FALSE) {
    new = xcluster_generation(hit_lst, rpic, rorg, old, &hash, filname,
			      p_cluster, cluster_write, &cl.crun, kernel, 
			      cluster_thold, cfileno);

    create_rb_list(rblst, rpic);
    add_hash_lst(gen, hash);                      /* create new generation */
  }

  free_cand_lst(&hit_lst);

  return new;
}



/* Adds a hash entry to a hash list. The hash list contains the history 
   of parent-child generation of a clustering cycle.

Revision: 27.08.2002 E.Sack

Input : - pointer to generation hash list
        - generation hash
Output: - generation hash list updated
Postcondition: - pointer just copied
*/
static void
add_hash_lst(hash_lst *cluster, hash_def hash)
{
  if (!hash.cnt) {
    return;
  }

  if (cluster->cnt) {
    REALLOC(hash_def, cluster->cnt+1, cluster->nid);
  }
  else {
    CALLOC(hash_def, cluster->cnt+1, cluster->nid);
  }

  cluster->nid[cluster->cnt] = hash;

  cluster->cnt++;
}



/* Wenn mindestens einer der beiden Nachbarn links und rechts einer
   erkannten NB (r) in der RB unerkannt (n) ist, wird die erkannte NB
   als unsicher markiert (r->u). Ein unsicherer Kandidat (u) selbst erzeugt
   dagegen keinen weiteren unsicheren Kandidaten:
   rrr -> rrr, nrr ->nur, rrn -> run, nrn -> nun, nur -> nur.

Input : - Kandidatenliste
Output: - Kandidaten-Nummernliste
        - Anzahl unsicherer Kandidaten
Precondition: kein Kandidat als unsicher markiert.  */
static int
untrusty_candidate(cand_lst hit)
{
  int     j, i, rb, stop = 0, *cand = 0, unsure = 0;
  int_lst lst;

  assert(hit.cnt);

  init_int_lst(&lst);
  CALLOC(int, hit.cnt, lst.nr);                      /* allocate max. size */
  CALLOC(int, hit.cnt, cand);

  while (1) {
    for (i = 0, stop = 1; i < hit.cnt; i++) {
      if (cand[i]) continue;                           /* candidate booked */

      stop = 0;
      rb = hit.c[i].rbn;                              /* remember RB index */
      lst.cnt = 0;

      for (j = i; j < hit.cnt; j++) {  /* search for candidates of same RB */
	if (hit.c[j].rbn == rb) {
	  assert(!cand[j]);
          cand[j] = 1;                                   /* mark candidate */
	  lst.nr[lst.cnt++] = j;            /* gathering candidate numbers */
	}
      }
      break;
    }

    if (stop) break;                              /* all candidates booked */

    sort_nb(hit, lst);               /* sort NB of the RB against location */

    for (i = 0; i < lst.cnt; i++) {            /* search unsafe candidates */
      assert(hit.c[lst.nr[i]].n != SUBSTUC_FLAG);
      if (hit.c[lst.nr[i]].n < 0) continue;
      if (i) {                                            /* left neighbor */
	if (hit.c[lst.nr[i-1]].n == SUBSTUC_FLAG) continue;
        if (hit.c[lst.nr[i-1]].n < 0) {
	  hit.c[lst.nr[i]].n = SUBSTUC_FLAG;
	  unsure++;
	  continue;
	}
      }
      if (i < lst.cnt-1) {                               /* right neighbor */
        if (hit.c[lst.nr[i+1]].n < 0) {
	  hit.c[lst.nr[i]].n = SUBSTUC_FLAG;
	  unsure++;
	  continue;
	}
      }
    }
  }

  FREEMEM(cand);
  free_int_lst(&lst);

  return unsure;
}



/* Generation of new NBs.
   With the candidate list and the appendant RB image the pattern will be
   copied and stored as NB in a list, which will be expanded accordingly.

Revision: 27.03.2002 E.Sack

Input : - candidate list
        - working RB image
        - original RB image
        - old class list
        - pointer to generation hash list
        - pointer to document filename
        - cluster flags
        - pointer to control structure
        - pointer to cluster kernel
        - cluster threshold
Output: - new class list
        - control structure updated
Postcondition: - the pattern of the accepted NB are erased in RP image.
               - if candidate list is empty, the new class list is empty, too.
*/
static cls_lst
xcluster_generation(cand_lst hit, pic_def rpic, pic_def rorg, cls_lst old, 
		    hash_def *hash, char *filename, CLUSTER_NB_BIT p_cluster,
		    int cluster_write, run_def *crun, kern_def kernel, 
		    int cluster_thold, int *cfileno)
{
  int     d, i, no, co, cn, nn, new_pic, identic, twin, twinsum = 0, unsure;
  int     new_nb = 0, idsum = 0;
  nb_def  nbz;
  cls_lst new; 

  assert(cfileno);

  init_nb_def(&nbz);
  init_cls_lst(&new);

  if (!hit.cnt) {
    return new;
  }

  unsure = untrusty_candidate(hit);          /* sort out unsure candidates */

  xplab_progress_configure(crun, "clustering");
  xplab_progress_set_value(crun, 1.0, (gdouble)hit.cnt);

  for (i = 0, new_pic = 0; i < hit.cnt && *crun->pactive==TRUE; i++) {/*all candidates */
  xplab_progress_set_value(crun, (gdouble)(i+1), (gdouble)hit.cnt);

    if (hit.c[i].n < 0) {                    /* only recognized candidates */
      continue;
    }

    switch (p_cluster) {
    case CLUSTER_NB_NEW:                   /* only take over new candidate */
      if (hit.c[i].m.cv == 1.0) {                 /* identic candidate ... */
        clear_nb(rpic, rorg, hit.c[i], old);      /* ... erased            */
	new_pic = 1;
	continue;                                 /* ... not taken over    */
      }
      break;
    case CLUSTER_NB_OLD:                   /* only take over old candidate */
      if (hit.c[i].m.cv != 1.0) {                 /* new candidate ...     */
        clear_nb(rpic, rorg, hit.c[i], old);      /* ... erased            */
	new_pic = 1;
	continue;                                 /* ... not taken over    */
      }
      break;
    case CLUSTER_NB_NEW|CLUSTER_NB_OLD: /* take over old and new candidate */
      break;
    default: assert(0); break;
    }

    nbz.pat = cut_mask(rpic, rorg, hit.c[i], old, kernel, /*get NB pattern */
		       cluster_thold);
    clear_nb(rpic, rorg, hit.c[i], old);                /* erase candidate */

    if (!nbz.pat.mem) {                  /* NB empty, nothing to take over */
      continue;
    }

    co = hit.c[i].c;                                          /* short cut */
    no = hit.c[i].n;

    d = check_part_count(old.cls[co].nbl.nb[no].pat, nbz.pat);
    if (d) {                                  /* never take over fractions */
      free_nb_def(&nbz);
      continue;
    }

    twin = classify_twins(old.cls[co].att.class, 
			  old.cls[co].att.alignment, nbz, old);
    twin += classify_twins(old.cls[co].att.class, 
			   old.cls[co].att.alignment, nbz, new);
    twinsum += twin;
    new_pic = 1;

    if (twin) {                                   /* never take over twins */
      free_nb_def(&nbz);
      continue;
    }

    idsum = search_identic_nb(old.cls[co].att.class, nbz, old);
    idsum += search_identic_nb(old.cls[co].att.class, nbz, new);
    if (idsum) {       /* never take over identic pattern of other classes */
      free_nb_def(&nbz);
      continue;
    }

    switch (p_cluster) {
    case CLUSTER_NB_NEW:                   /* only take over new candidate */
      identic = search_nb(old.cls[co].att.class, 
			  old.cls[co].att.alignment, nbz, old);
      identic += search_nb(old.cls[co].att.class, 
			   old.cls[co].att.alignment, nbz, new);
      if (identic) {
	break;                  /* pattern already in class, don't take it */
      }

    case CLUSTER_NB_OLD:                   /* only take over old candidate */
    case CLUSTER_NB_NEW|CLUSTER_NB_OLD: /* take over old and new candidate */
      nbz.aux = gen_hash_entry(hash, hit.c[i], &old.cls[co].nbl.nb[no].aux);

      cn = search_class(old.cls[co].att.class, new);

      if (cn >= 0) {                          /* class already in new list */
	REALLOC(nb_def, new.cls[cn].nbl.cnt+1, new.cls[cn].nbl.nb);
      }
      else {                                /* class not found in new list */
        cn = new.cnt++;

        if (cn) REALLOC(cls_def, cn+1, new.cls); /* new list was not empty */
        else CALLOC(cls_def, cn+1, new.cls);     /* new list was empty     */

        init_cls_def(&new.cls[cn]);          /* take over class attributes */
	copy_cls_att(old.cls[co].att, &new.cls[cn].att);

	CALLOC(nb_def, new.cls[cn].nbl.cnt+1, new.cls[cn].nbl.nb);
      }

      nn = new.cls[cn].nbl.cnt++;                             /* short cut */
      init_nb_def(&new.cls[cn].nbl.nb[nn]);

      move_nb_def(&nbz, &new.cls[cn].nbl.nb[nn], FALSE);/* enroll NB pattern */
      new_nb++;
      break;

    default: assert(0); break;
    }

    free_nb_def(&nbz);
  }

  xreport("[%d unsure candidate%s]", unsure, unsure==1?"":"s");
  xreport("[%d twin%s]", twinsum, twinsum==1?"":"s");
  xreport("[%d identit%s]", idsum, idsum==1?"y":"ies");
  xreport("[%d new candidate%s]\n", new_nb, new_nb==1?"":"s");

  if (new_pic && cluster_write)
  {
    char *buf = NULL;

    add_string(filename, &buf);
    int_to_string(CLUSTER_DEFAULT FILE_EXT, (*cfileno)++, "", &buf);
    dump_pic_def(rpic, buf, rpic.ar, 0, 0, xctrl.cfg.verbose > BE_QUIET, 
		 PNM_FORMAT_UNDEF);
    FREEMEM(buf);
  }

  return new;
}



/* Compares the blob number of two images.

Revision: 03.03.2002 E.Sack

Input : - image a
        - image b
Output: - Flag (0: no difference, 1: difference)
*/
int
check_part_count(pic_def pic_a, pic_def pic_b)
{
  int        part_a, part_b;
  sector_def fr;
  pic_def    z;                       /* buffer for non destroying pattern */

  init_pic_def(&z);

  z.ar = pic_a.ar;                                             /* pattern A */
  CALLOC(PICT, AREA_SIZE(z.ar), z.mem);
  copy_bmps(pic_a, z, z.ar, 0, 0, 0, 0);

  AREA_TO_SECTOR(z.ar, fr);
  part_a = part_count(z, fr, 0);

  free_pic_def(&z);

  z.ar = pic_b.ar;                                             /* pattern B */
  CALLOC(PICT, AREA_SIZE(z.ar), z.mem);
  copy_bmps(pic_b, z, z.ar, 0, 0, 0, 0);

  AREA_TO_SECTOR(z.ar, fr);
  part_b = part_count(z, fr, 0);

  free_pic_def(&z);

  return part_b-part_a;
}



/* Calculates the number of separate foreground regions in a image
   while seeking recursively for contours. Contours may contain other 
   contours.

Revision: 01.06.2004 E.Sack

Input : - image
        - clipping
        - flag (0: BMP not inverted, 1: BMP inverted)
Output: - number of foreground regions
Postcondition: image changed
*/
static int 
part_count(pic_def pic, sector_def fr, int invtd) 
{
  int         d, cnt = 0;
  h_contour_lst ctr;

  if (invtd) {                                              /* BMP inverted */
    while (1) {                                             /* all inverted */
      ctr = xgen_h_contour(pic, fr, XPLAB_IGNORED_PIXEL, 0, 0);

      if (!ctr.cnt) {
	break;
      }

      set_h_contour(ctr, BACKGROUND, FOREGROUND, pic);/* get foreground */
      set_h_contour(ctr, XPLAB_IGNORED_PIXEL, BACKGROUND, pic);/* get background */
      free_h_contour_lst(&ctr);
      cnt += part_count(pic, fr, 0);          /* ignore inverted background */
    }

    return cnt;
  }
  else {                                                /* BMP not inverted */
    ctr = xgen_h_contour(pic, fr, FOREGROUND, 0, 0);       /* look for FGND */

    if (!ctr.cnt) {                                 /* no foreground pixels */
      return 0;
    }

    cnt = 1;

    d = set_h_contour(ctr, BACKGROUND, XPLAB_IGNORED_PIXEL, pic); /* in contour all */
    set_h_contour(ctr, FOREGROUND, BACKGROUND, pic);      /* color inverted */

    if (d) {
      invtd = 1;                                      /* next inner contour */
    }
    else {                              /* no background pixels in contour  */
      invtd = 0;                        /* old contour erased, next contour */
    }
  }

  free_h_contour_lst(&ctr);
  cnt += part_count(pic, fr, invtd);

  return cnt;
}



/* Searches in a class for a certain NB.

Revision: 27.07.2002 E.Sack

Input : - pointer to class name
        - alignment flag
        - NB
        - class list
Output: - flag (0: NB not found, 1: NB found)
*/
int
search_nb(const char *cname, ALIGMNT zline, nb_def nb, cls_lst clst)
{
  int c, n, identic;

  for (c = 0; c < clst.cnt; c++) {
    if (strcmp(clst.cls[c].att.class, cname))  {
      continue;             /* not identic because of different class name */
    }

    if (clst.cls[c].att.alignment != zline) {
      continue;              /* not identic because of different alignment */
    }

    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      identic = compare_bvl(clst.cls[c].nbl.nb[n], nb);

      if (identic == 2) {
        return 1;                 /* identic NB found (pattern & b-values) */
      }
    }
  }

  return 0;
}



/* Searches in a NB list for a certain pattern.

Revision: 27.07.2002 E.Sack

Input : - pointer to class name
        - NB
        - class list
Output: - flag (0: NB not found, 1: NB found)
*/
int
search_identic_nb(const char *cname, nb_def nb, cls_lst clst)
{
  int c, n, identic;

  for (c = 0; c < clst.cnt; c++) {
    if (!strcmp(clst.cls[c].att.class, cname))  {
      continue;                               /* own class not investigated */
    }

    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      identic = compare_bmp(clst.cls[c].nbl.nb[n], nb);

      if (identic) {
        return 1;                                  /* identic pattern found */
      }
    }
  }

  return 0;
}



/* Searches in a NB list for a certain twin.

Revision: 27.07.2002 E.Sack

Input : - pointer to class name
        - alignment flag
        - NB
        - class list
Output: - flag (0 twin not found, 1 twin found)
*/
static int
classify_twins(const char *cname, ALIGMNT zline, nb_def nb, cls_lst clst)
{
  int c, n, identic;

  for (c = 0; c < clst.cnt; c++) {
    if (strcmp(clst.cls[c].att.class, cname))  {
      continue;                /* no twins because of different class names */
    }

    if (clst.cls[c].att.alignment != zline) {
      continue;                 /* no twins because of different alignments */
    }

    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      identic = classify_bvl(nb, &clst.cls[c].nbl.nb[n], 1);

      if (identic) {                     /* twin found, NB-b-values merged */
        return 1;   
      }
    }
  }

  return 0;                                              /* twin not found */
}



/* Creates a hash entry. Unique birth IDs for parent and child are generated
   and enrolled in a hash entry. With the IDs the NB references are found
   in the class list. The affinity of parents and children is documented
   only in the hash entry. Birth IDs have nothing to do with other numbers
   and are unique.

Revision: 02.10.2004 E.Sack

Input : - pointer to hash entry
        - candidate features
        - pointer to parent ID
Output: - child ID
        - hash entry updated
        - parent ID updated
*/
int
gen_hash_entry(hash_def *hash, cand_def kr, int *pid)
{
  static int hash_id = XPLAB_INIT_HASH_ID;                /* initialise ID */
  int        cid;

  if (hash->cnt) {
    REALLOC(int, hash->cnt+1, hash->cid);
    REALLOC(int, hash->cnt+1, hash->pid);
    REALLOC(cand_def, hash->cnt+1, hash->kr); 
  }
  else {
    CALLOC(int, hash->cnt+1, hash->cid);
    CALLOC(int, hash->cnt+1, hash->pid);
    CALLOC(cand_def, hash->cnt+1, hash->kr);
  }

  init_cand_def(&hash->kr[hash->cnt]);

  cid = ++hash_id;                                    /* generate child ID */
  hash->cid[hash->cnt] = cid;             /* enroll child ID in hash entry */

  if (*pid < XPLAB_INIT_HASH_ID) {    /* if parent has no ID, generate one */
    *pid = ++hash_id;
  }

  hash->pid[hash->cnt] = *pid;           /* enroll parent ID in hash entry */

  copy_cand_def(&kr, &hash->kr[hash->cnt]);/* take over birthdata of child */

  hash->cnt++;

  return cid;                     /* child ID strictly linked to parent ID */
}



/* Copys a region, which is defined by a mask, to a NB.

Revision: 03.10.2002 E.Sack

Input : - image
        - original RB image
        - candidate
        - class list
Output: - NB updated
Postcondition: - NB pattern may be empty
*/
pic_def
cut_mask(pic_def rpic, pic_def rorg, cand_def cand, cls_lst clst, 
	 kern_def kernel, int cluster_thold) 
{
  int     d;
  pic_def copy;

  assert(cand.c>=0 && cand.c<clst.cnt);
  assert(cand.n>=0 && cand.n<clst.cls[cand.c].nbl.cnt);

  copy = 
    xcopy_masked_area(clst.cls[cand.c].nbl.nb[cand.n].ref.pic,/* NB pattern */
		      clst.cls[cand.c].nbl.nb[cand.n].ref.msk, /* NB mask   */
                      cand.roi,                        /* RB image clipping */
                      rpic,                            /* RB image          */
                      rorg,                            /* original RB image */
                      cand.m.x,                 /* NB offset relative to RB */
                      cand.m.y,
		      kernel, cluster_thold);

  pixel_erase(copy, NBH4C, 0);                        /* no frayed pattern */

  d = xcopy_and_fit_pic(copy, &copy);         /* minimal frame gives new NB */

  if (!d) {                                             /* frame was empty */
    free_pic_def(&copy);
  }

  return copy;
}



/* Clears a region, which is defined by a mask, in an image.

Revision: 03.10.2002 E.Sack

Input : - image
        - original RB image
        - candidate
        - class list
Output: - image updated
*/
void
clear_nb(pic_def rpic, pic_def rorg, cand_def cand, cls_lst clst) 
{
  assert(cand.c>=0 && cand.c<clst.cnt);
  assert(cand.n>=0 && cand.n<clst.cls[cand.c].nbl.cnt);

  erase_masked_area(clst.cls[cand.c].nbl.nb[cand.n].ref.pic, /* NB pattern */
    	            clst.cls[cand.c].nbl.nb[cand.n].ref.msk, /* NB mask    */
                    cand.roi,                         /* RB image clipping */
                    rpic,                             /* RB image          */
                    rorg,                             /* original RB image */
                    cand.m.x,                  /* NB offset relative to RB */
                    cand.m.y);

  pixel_erase(rpic, NBH8, 0);
}
 


/* Writes the clustering results.

Revision: 27.08.2002 E.Sack

Input : - old class list
        - new class list
Output: - results dumped to report file
*/
static void
write_generation_rpt(cls_lst old, cls_lst new)
{
  int   onr, nnr, d, childless = 0, parent = 0, children = 0, hope;
  char  buf1[MAXCHAR], buf2[MAXCHAR];
  ulong max;

  if (!xctrl.set.report_stream) {
    return;
  }

  max = strlen(old.cls[0].att.class);                /* format calculation */
  for (onr = 1; onr < old.cnt; onr++) {
    max = XPMAX(max, strlen(old.cls[onr].att.class));
  }

  xreport("\nNB generation report\n");
  xreport("<class>     <name>  <#parents>  ==>  "
                      "<number of successors>\n");

  for (onr = 0; onr < old.cnt; onr++) {
    g_snprintf(buf1, sizeof(buf1), "Class[%d]", onr);
    g_snprintf(buf2, sizeof(buf2), "%-*s %-*s ", 12, buf1, (int)max, 
	     old.cls[onr].att.class);
    g_snprintf(buf1, sizeof(buf1), "%-*s %5d", (int)(14+max), buf2, 
	     old.cls[onr].nbl.cnt);
    xreport("%s  ==>  ", buf1);

    hope = old.cls[onr].nbl.cnt;

    for (nnr = 0, d = 1; nnr < new.cnt; nnr++) {
      d = strcmp(old.cls[onr].att.class, new.cls[nnr].att.class);
      if (!d) {                                                   /* equal */
	xreport("%d\n", new.cls[nnr].nbl.cnt);
	parent += hope;                                      /* parent sum */
	children += new.cls[nnr].nbl.cnt;                  /* children sum */
	break;
      }
    }

    if (d) {
      xreport("no successors\n");
    }
  }

  childless = old.cnt - parent;

  if (parent) {
    xreport("\n%d NB with %d successor%s\n", parent, children, 
	    children==1?"":"s");
    xreport("%d NB without successors\n", childless);
  }
  else {
    xreport("all NB without successors\n");
  }
}



/* Writes the pedigrees of the cluster.

Revision: 27.07.2002 E.Sack

Input : - generation hash list
        - class list
        - pointer to image file name
        - pointer to database file name
Output: - pedigrees dumped to disc
*/
static void
write_cluster_rpt(hash_lst cluster, cls_lst clst, char *input_name, 
		  char *db_name)
{
  int     d, l, h, clos;
  gen_lst gel;                           /* generation lists for all new NB */
  FILE    *cfile;
  char    *cls_name = NULL, *settings = 0;

  gel = build_pedigrees(cluster, clst);

  cat_string(input_name, FILE_CLS, "", &cls_name);

  cfile = xplab_open_file(cls_name, "w");
  if (cfile && xctrl.cfg.verbose > VERBOSE_DEFAULT) {
    string_to_log(main_log, "Cluster file '", cls_name, " opened\n");
  }

  settings = print_settings(cl.ctrl, CRF_HEADER, db_name);
  d = fprintf(cfile, "%s", settings);assert(d > 0);/* write to cluster file */
  FREEMEM(settings);

  d = fprintf(cfile, "\n\n\nList of Generations:\n");          assert(d > 0);
  d = fprintf(cfile, "\n");                                    assert(d > 0);
  d = fprintf(cfile, "Parent -> Child      K            "
                     "lag      kernel\n\n");                   assert(d > 0);

  for (l = gel.cnt-1; l >=0; l--) {   /* begin with last list for clearness */
    for (h = 0; h < gel.gd[l].cnt; h++) {
      if (!h) {
	d = fprintf(cfile, "Class[%d] = %s\n",
                    gel.gd[l].c[h], clst.cls[gel.gd[l].c[h]].att.class);
	assert(d > 0);

      }
      else {
        assert(gel.gd[l].c[h-1] == gel.gd[l].c[h]);          /* same class! */

	d = fprintf(cfile, "%4d  %4d           %f     %-3d %-3d  ",
                    gel.gd[l].n[h-1], gel.gd[l].n[h],
                    gel.gd[l].hl[h].m.cv, gel.gd[l].hl[h].m.x, 
		    gel.gd[l].hl[h].m.y);
	assert(d > 0);

	d = fprintf(cfile, "(%d-%d)x(%d-%d)\n",
	     gel.gd[l].hl[h].knl.column.start, gel.gd[l].hl[h].knl.column.end,
	     gel.gd[l].hl[h].knl.line.start, gel.gd[l].hl[h].knl.line.end);
	assert(d > 0);
      }
    }

    d = fprintf(cfile, "\n"); assert(d > 0);
  }

  clos = xplab_close_file(cfile, cls_name);

  if (!clos && xctrl.cfg.verbose > VERBOSE_DEFAULT) {
    string_to_log(main_log, "Cluster file '", cls_name, " closed\n");
  }

  free_gen_lst(&gel);
  FREEMEM(cls_name);
}



/* Builds the generations of all NBs. The new received NB are supplied with
   marks, which indicates the generations with the help of the candidates
   list.

Revision: 27.07.2002 E.Sack

Input : - generation hash list
        - class list
Output: - generation list
Precondition: inputs must be consistent
*/
static gen_lst
build_pedigrees(hash_lst cluster, cls_lst cl)
{
  int     l, h, gcnt;
  gen_lst gen;

  init_gen_lst(&gen);

  for (l = cluster.cnt-1; l >=0; l--) {            /* begin with last list */
    assert(cluster.nid[l].cnt);
    for (h = 0; h < cluster.nid[l].cnt; h++) { /* hash list refers candid. */
      gcnt = gen.cnt++;                                        /* shortcut */

      if (gcnt) REALLOC(gen_def, gcnt+1, gen.gd);
      else CALLOC(gen_def, gcnt+1, gen.gd);
      init_gen_def(&gen.gd[gcnt]);

      gen.gd[gcnt] = build_generation(cluster, cl,          /* parent list */
                                      cluster.nid[l].cid[h], 
                                      cluster.nid[l].pid[h], 
                                      cluster.nid[l].kr[h]);
    }
  }

  return gen;
}



/* Builds the generation of a NB regarding hash marks.

Revision: 27.07.2002 E.Sack

Input : - generation hash list
        - class list
        - child ID
        - parent ID
        - candidate
Output: - child birth dates
*/
static gen_def
build_generation(hash_lst gen, cls_lst clst, int cid, int pid, cand_def ckr)
{
  int      ce = -1, ne = -1, c, n = -1, gcnt, stop;
  gen_def  gd;
  cand_def dummy;

  assert(cid);

  init_gen_def(&gd);

  if (pid) {                    /* search in hash list for PID recursively */
    for (c = 0, stop = 0; c < gen.cnt && !stop; c++) {
      for (n = 0; n < gen.nid[c].cnt; n++) {
        if (gen.nid[c].cid[n] != pid) continue;

        gd = build_generation(gen, clst, pid, gen.nid[c].pid[n],
			      gen.nid[c].kr[n]);
        stop = 1;
        break;
      }
    }

    if (!stop) {/* if no pid in hash list found it is the root in database */
      gd = build_generation(gen, clst, pid, 0, dummy);
    }
  }

  /* search in DB for CID.                                                 */
  for (c = 0, stop = 0; c < clst.cnt && !stop; c++) {
    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      if (clst.cls[c].nbl.nb[n].aux != cid) {
	continue;
      }

      ce = c;
      ne = n;
      stop = 1;
      break;
    }
  }

  assert(stop);                                 /* CID muss vorhanden sein */

  gcnt = gd.cnt++;

  if (gcnt) {
    REALLOC(int, gcnt+1, gd.c); 
    REALLOC(int, gcnt+1, gd.n);
    REALLOC(cand_def, gcnt+1, gd.hl);
  }
  else {
    CALLOC(int, gcnt+1, gd.c);
    CALLOC(int, gcnt+1, gd.n);
    CALLOC(cand_def, gcnt+1, gd.hl); 
  } 

  init_cand_def(&gd.hl[gcnt]);

  /* Klasse und NB-Nummer des Kindes aus der DB eintragen.                  */
  gd.c[gcnt] = ce;
  gd.n[gcnt] = ne;

  if (pid) {        /* Eltern vorhanden, Geburtsdaten des Kindes eintragen. */
    copy_cand_def(&ckr, &gd.hl[gcnt]);
  }

  return gd;
}



/* Zwei NB werden auf Identitt ihrer Muster und ihrer b-Werte untersucht.

Input : - NB
Output: - Flag (0 NB nicht identisch,
                1 NB-Muster identisch & NB-b-Werte nicht identisch,
                2 NB-Muster identisch & NB-b-Werte identisch)
*/
static int
compare_bvl(nb_def nb1, nb_def nb2)
{
  int l, len, identic, b, m;

  identic = compare_bmp(nb1, nb2);

  if (!identic) {
    return 0;                                /* NB-Muster nicht identisch */
  }

  if (!nb1.pat.bvl && !nb2.pat.bvl) {
    return 2;                          /* b-values not at all: NB identic */
  }

  if (!nb1.pat.bvl || !nb2.pat.bvl) {
    return 1; /* no b-values: NB-pattern identic & NB-b-values not identic */
  }

  len = nb1.pat.ar.width * nb1.pat.ar.height;

  for (l = 0; l < len; l++) {
    b = nb1.pat.bvl[l];                                      /* short cut */
    m = nb1.pat.mem[l];                                      /* short cut */
    assert(!b || (b && m));                      /* consistence violation */

    b = nb2.pat.bvl[l];                                      /* short cut */
    m = nb2.pat.mem[l];                                      /* short cut */
    assert(!b || (b && m));                      /* consistence violation */

    if (nb1.pat.bvl[l] != nb2.pat.bvl[l]) {
      return 1;           /* NB-pattern identic & NB-b-values not identic */
    }
  }

  return 2;                   /* NB-pattern identic & NB-b-values identic */
}



/* Zwei NB werden auf Identitt ihrer Muster untersucht und bei 
   bereinstimmung werden ihre b-Werte regelbasiert zusammengefasst.

Revision: 03.03.2002 E.Sack

Input : - Quelle
        - Senke
        - Flag (0 schwache B-Wert Bedingung, 1 starke B-Wert Bedingung)
Output: - Flag (0 b-values not merged,
                1 b-values merged)
Postcondition: source never changed.
*/
static int
classify_bvl(nb_def src, nb_def *dst, int strong)
{
  int l, len = 0, identic, b, m;

  identic = compare_bmp(src, *dst);

  if (!identic) {
    return 0;                    /* NB-Muster nicht identisch: not merged */
  }

  len = src.pat.ar.width * src.pat.ar.height;

  if (strong) {
    if (!src.pat.bvl) {                         /* source b-values are zero */
      return 1;                                                   /* merged */
    }

    if (!dst->pat.bvl) {                         /* no destination b-values */
      CALLOC(PICT, len, dst->pat.bvl);       /* create destination b-values */

      for (l = 0; l < len; l++) {
        dst->pat.bvl[l] = src.pat.bvl[l];                           /* copy */
      }

      return 1;                                                   /* merged */
    }
  }
  else {
    if (!dst->pat.bvl) {                   /* destination b-values are zero */
      return 1;/* source b-values may be merged or source redundant: merged */
    }

    if (!src.pat.bvl) {                               /* no source b-values */
      FREEMEM(dst->pat.bvl);         /* destination b-values zeroed: merged */
      return 1;            
    }
  }

  for (l = 0; l < len; l++) {
    m = src.pat.mem[l];                                      /* short cut */
    b = src.pat.bvl[l];                                      /* short cut */
    assert(!b || m);                             /* consistence violation */

    m = dst->pat.mem[l];                                     /* short cut */
    b = dst->pat.bvl[l];                                     /* short cut */
    assert(!b || m);                             /* consistence violation */

    if (strong) {                                         /* merging rule */
      dst->pat.bvl[l] = XPMAX(dst->pat.bvl[l], src.pat.bvl[l]);
    }
    else {
      dst->pat.bvl[l] = XPMIN(dst->pat.bvl[l], src.pat.bvl[l]);
    }
  }

  return 1;                                            /* b-values merged */
}



/* Adds an item to a menu.

Revision: 28.10.2003 E.Sack

Input : - menu
Output: - menu item
*/
void
xplab_clustering_menu(GtkWidget *match_menu, GtkWidget **caller)
{
  GtkWidget *match_menu_cluster;

  match_menu_cluster = gtk_menu_item_new_with_label(XPLAB_CLUSTERING_TITLE_ITEM XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(match_menu), match_menu_cluster);

  g_signal_connect(G_OBJECT(match_menu_cluster), "activate", 
		   G_CALLBACK(cb_clustering_phase), (gpointer)caller);
}



/* Starts clustering phase.

Revision: 28.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_clustering_phase(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  assert(user_data);
  assert(menuitem);

  raised = xplab_window_raise(&cl.top_clustering, 1);

  if (raised == FALSE) {
    xplab_free_clustering();    /* must free here because of sudden run stop */
    cl.caller = (GtkWidget **)user_data;              /* last succeeded call */
    clustering_window(xctrl.cfg.mad.corrlevel);             /* setup widgets */
  }
}



/* Creates top window for clustering.

Revision: 29.07.2005 E.Sack

Input : - -
Output: - top window
*/
static void
clustering_window(gdouble corrlevel)
{
  int       err;
  char      *buf = NULL;
  GtkWidget *vbox;

  /* set the clustering top level widget                                     */
  cl.top_clustering = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(cl.top_clustering), XPLAB_CLUSTERING_TITLE_ITEM);
  gtk_window_set_default_size(GTK_WINDOW(cl.top_clustering), 
			      XPLAB_CLUSTERING_WINDOW_WIDTH, 
			      XPLAB_CLUSTERING_WINDOW_HEIGHT);

  /* put into top level widget a vertical box                                */
  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_add(GTK_CONTAINER(cl.top_clustering), vbox);

  clustering_menu_bar(vbox);              /* put into vertical box a menubar */

  clustering_iteration(vbox, corrlevel);                    /* display panel */
  clustering_control(vbox);                                 /* control panel */

  gtk_widget_show_all(cl.top_clustering);

  g_signal_connect(G_OBJECT(cl.top_clustering), "delete-event", 
		   G_CALLBACK(cb_clustering_top_delete), NULL);

  xplab_set_window_icon(cl.top_clustering);                   /* window icon */

  err = check_name(xctrl.set.image_filename);

  if (err) {
    xplab_label_set_text(cl.crun.plabel, "WARNING: no image file selected");
  }
  else {
    db_def db_tmp;

    db_tmp = xplab_request_db(NULL, NULL, XPLAB_DB_RO, NULL);     /* DB copy */
    err = check_name(db_tmp.db_fname);

    if (err) {
      xplab_label_set_text(cl.crun.plabel, "WARNING: no database file selected");
    }
    else {                                                        /* success */
      cat_string("image: ", xctrl.set.image_filename, ", ", &buf);
      cat_string("DB: ", db_tmp.db_fname, "", &buf);

      xplab_label_set_text(cl.crun.plabel, buf);
    }

    free_db_def(&db_tmp);
  }

  FREEMEM(buf);
}



/* Deletes top window.

Revision: 28.11.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_clustering_top_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_clustering_quit(NULL, NULL);

  return FALSE;
}



/* Quits clustering phase.

Revision: 23.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_clustering_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  cb_clustering_stop(NULL, NULL);                         /* sudden run stop */

  if (GTK_IS_WIDGET(cl.top_clustering) == TRUE) {
    xplab_window_raise(cl.caller, xctrl.cfg.raise);
    xplab_widget_destroy(&cl.top_clustering);

    /* cannot free data here because of sudden run stop                      */

  }
}



/* Stops clustering phase.

Revision: 28.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_clustering_stop(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  stop_run_def(&cl.crun);
}



/* Builds a menu bar.

Revision: 29.07.2005 E.Sack

Input : - container
Output: - container updated
*/
static void
clustering_menu_bar(GtkWidget *vbox)
{
  GtkWidget *menu_bar;

  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

  clustering_menu_entry_file(menu_bar);                       /* menu file */
  xplab_menu_help(menu_bar, &cl.top_clustering);              /* menu help */
}



/* Builds an entry for a menu bar.

Revision: 29.07.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
clustering_menu_entry_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *file_menu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  file_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), file_menu);

  xplab_preferences_menu(file_menu, &cl.top_clustering);/*item preferences */
  xplab_selection_menu(file_menu, &cl.top_clustering);   /* item selection */

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep);

  clustering_file_menu_item_quit(file_menu);                  /* item quit */
}



/* Adds an item to a menu.

Revision: 29.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
clustering_file_menu_item_quit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_quit;

  file_menu_quit = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_quit);

  g_signal_connect(G_OBJECT(file_menu_quit), "activate", 
		   G_CALLBACK(cb_clustering_quit), NULL);
}



/* Builds the clustering control panel.

Revision: 29.07.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
clustering_control(GtkWidget *vbox1)
{
  GtkWidget *hbox1, *hbox, *vbox;

  /* put into inbox a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

  /* put into horizontal box a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox1), vbox, TRUE, TRUE, 0);

  /* put into vertical box a label */
  cl.crun.plabel = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), cl.crun.plabel, FALSE, FALSE, 0);

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a start button */
  xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_clustering_start), 
		      SINGLE_BUTTON_START,
		      DESC1_CLUSTER_START, DESC2_CLUSTER_START, NULL);

  /* put into horizontal box a stop button */
  xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_clustering_stop), 
		      SINGLE_BUTTON_STOP,
		      DESC1_STOP_ACTION, DESC2_STOP_ACTION, NULL);

  /* put into horizontal box a progress bar */
  cl.crun.pbar = xplab_progress(hbox);
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(cl.crun.pbar), " ");/* set height */
}



/* Builds the clustering display panel.

Revision: 30.11.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
clustering_iteration(GtkWidget *vbox1, gdouble corrlevel)
{
  GtkWidget *hbox1, *hbox, *vbox;

  /* put into inbox a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

  /* put into horizontal box a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox1), vbox, TRUE, TRUE, 0);

  /* put into vertical box a label */
  cl.iteration_label = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), cl.iteration_label, FALSE, FALSE, 0);

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a progress bar */
  cl.irun.pbar = xplab_progress(hbox);
  xplab_progress_configure(&cl.irun, "CV threshold");
  xplab_progress_set_value(&cl.irun, corrlevel, 1.0);
}



/* Starts pattern matching.

Revision: 28.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_clustering_start(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (*cl.crun.pactive == TRUE) {
    return;
  }

  start_run_def(&cl.crun);

  clustering_setup(xctrl);

  stop_run_def(&cl.crun);
}



/* Clustering phase of pattern matching methods.

Revision: 17.08.2005 E.Sack

Input : - settings
Output: - clustering results
*/
static void
clustering_setup(prog_ctrl ctrl)
{
  int    tps_idx, err;
  db_def db_tmp;

  copy_prog_ctrl(&ctrl, &cl.ctrl);                        /* global -> local */

  db_tmp = xplab_request_db(XPLAB_CLUSTERING_TITLE_ITEM, &cl.db_key, 
			    XPLAB_DB_RW, NULL);        /* lock DB during run */

  err = check_name(db_tmp.db_fname);

  if (err) {
    xplab_label_set_text(cl.crun.plabel, "database locked - cannot run");
    free_db_def(&db_tmp);
    return;
  }
  else {
    check_aux(db_tmp.tpl, 1);                       /* make sure DB is clean */
  }

  free_pic_def(&cl.cimg);
  cl.cimg = xplab_request_image(P1, NULL, NULL, 0, NULL).pimg; /* image copy */

  tps_idx = check_cluster_resources(db_tmp, cl.ctrl.cfg.tps_id, 
				    cl.ctrl.set.image_filename, 
				    cl.cimg, cl.ctrl.set.rblst);
  if (tps_idx < 0) {
    string_to_label(cl.crun.plabel, "resource missing - cannot run", "","");
    err = xplab_release_db(&cl.db_key, 1);                      /* unlock DB */
    assert(!err);
    free_db_def(&db_tmp);
    return;
  }

  xclustering(cl.ctrl.cfg.mad, db_tmp, tps_idx, &cl.ctrl.set.rblst, cl.cimg, 
	      cl.ctrl.set.image_filename, cl.ctrl.cfg.cdomain, 
	      cl.ctrl.cfg.cluster_nb, cl.ctrl.cfg.cluster_level, 
	      cl.ctrl.cfg.cluster_step, cl.ctrl.cfg.cluster_force, 
	      cl.ctrl.cfg.cluster_write, cl.ctrl.cfg.ckernel, 
	      cl.ctrl.cfg.cluster_thold, cl.ctrl.cfg.single_line, 
	      cl.ctrl.cfg.single_nb);

  time_to_log(main_log, XPLAB_CLUSTERING_TITLE_ITEM " finished after ", 
	      cl.crun.start_time, "\n");

  err = xplab_release_db(&cl.db_key, 1);                        /* unlock DB */
  assert(!err);

  free_db_def(&db_tmp);
}



/* Wrapper: Checks the cluster resources.

Revision: 02.12.2003 E.Sack

Input : - DB
        - pointer to TPS ID
        - pointer to image file name string
        - pointer to replay file name string
        - image
        - RB list
        - cluster kernel
Output: - TPS index (-1: error)
        - message output
*/
static int
check_cluster_resources(db_def db, char *tps_id, char *image_filename, 
		        pic_def image, rb_lst rblst)
{
  int tps_idx;

  tps_idx = check_recall_resources(db, tps_id, image_filename, 
				   image, rblst);
  return tps_idx;
}
