/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/
/*
Revision: 20.04.2003 E.Sack
*/

/* This module mainly lists generic functions to ensure better
   consistency. The aim is to maintain only one piece of code to fit
   changes in data structures.                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <float.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_debug.h"

/* Functions/Methods *******************************************************/

float **
alloc_fmatrix(int nx, int ny)
{
  int   i;
  float **m = 0;

  CALLOC(float, ny+1, *m);      /* allocate pointer to rows */

  for (i = 0; i < ny+1; i++) {
    CALLOC(float, nx+1, m[i]);      /* allocate rows */
  }

  return m;
}



/* Allocation of 2-D arrays (images) for FFT. Convention is (without loss
   of generality): X (width) dimension is the last dimension as suggested 
   in "Numerical Recipes in C", Y (height) dimension is the first dimension. 
   This order has the advantage of using the definition for 1-D data in a 
   row as well by setting the Y dimension length to 1.

Revision: 11.05.2003 E.Sack

Input : - image size
Output: - arrays allocated
*/
fft_def
alloc_fft_def(int w, int h)
{
  int     i;
  double  *dum = 0;               /* guarants consecutive memory allocation */
  fft_def fd;

  init_fft_def(&fd);

  assert(w>1);
  assert(h>1);

  fd.ar.width  = w;            /* at least 2 memories for complex value,    */
  fd.ar.height = h;            /* eventually zero-padded                    */

  CALLOC(double, 2*fd.ar.height, fd.spec);

  CALLOC(double, fd.ar.width*fd.ar.height, dum);
  CALLOC(double *, fd.ar.height+1, fd.data);                   /* see below */
  for (i = 0; i < fd.ar.height; i++) {
    fd.data[i] = &dum[i*fd.ar.width];
  }
  fd.data[fd.ar.height] = dum;   /* that is the reason for increased height */

  return fd;
}


kwh_def
alloc_kwh_def(area_def ar)
{
  int     i;
  lag_lst *dum = 0;
  kwh_def kd;

  init_kwh_def(&kd);

  kd.ar = ar;

  CALLOC(lag_lst, kd.ar.width*kd.ar.height, dum);
  CALLOC(lag_lst *, kd.ar.height+1, kd.ll);
  for (i = 0; i < kd.ar.height; i++) {
    kd.ll[i] = &dum[i*kd.ar.width];
  }
  kd.ll[kd.ar.height] = dum;

  return kd;
}


cv_def_i**
alloc_cv_def_i_matrix(int vector_len)
{
  int   i;
  cv_def_i *dum = 0, **matrix = 0;

  assert(vector_len>0);

  CALLOC(cv_def_i, vector_len*vector_len, dum);
  CALLOC(cv_def_i *, vector_len+1, matrix);
  for (i = 0; i < vector_len; i++) {
    matrix[i] = &dum[i*vector_len];
  }
  matrix[vector_len] = dum;

  return matrix;
}


col_def
convert_rgb_coef(rgb_coef in)
{
  col_def out;

  out.r = (PICT)in.r;
  out.g = (PICT)in.g;
  out.b = (PICT)in.b;

  return out;
} 


int
compare_col_def(col_def a, col_def b)
{
  return (int)((a.r == b.r) && (a.g == b.g) && (a.b == b.b));
} 


void 
init_batch_def(batch_def *bd) 
{
  assert(bd);

  bd->log_stream = NULL;                                   /* zero pointer */

  bd->phase = BATCH_PHASE_NO;
  bd->blog = BATCH_LOG_NO;
}


void
xdefault_batch_def(batch_def *bd)
{
  assert(bd);

  bd->phase = BATCH_PHASE_NO;
  bd->blog = BATCH_LOG_NO;
}


void 
init_kern_def(kern_def kd) 
{
  kern_def kernel = CKERNEL_NULL;

  assert(kd);

  copy_kern_def(kernel, kd);
}


void 
xdefault_kern_def(kern_def kd) 
{
  kern_def kernel = CKERNEL_DEFAULT;

  assert(kd);

  copy_kern_def(kernel, kd);
}


void 
init_prog_ctrl(prog_ctrl *ctrl) 
{
  assert(ctrl);

  init_set_def(&ctrl->set);
  init_cfg_def(&ctrl->cfg);
}


void
xdefault_prog_ctrl(prog_ctrl *pc)
{
  assert(pc);

  xdefault_set_def(&pc->set);
  xdefault_cfg_def(&pc->cfg);
}


void 
init_match_def(match_def *md) 
{
  assert(md);

  md->mborder    = 0;
  md->nborder    = 0;
  md->nnbmatch   = 0;
  md->maskweight = 0;
  md->corrlevel  = 0.0;
  md->corrbroken = 0.0;
  md->overlap    = 0.0;
  md->nb_lap     = 0.0;
  md->interleave = 0.0;
}


void
xdefault_match_def(match_def *md)
{
  assert(md);

  md->mborder    = MBORDER_DEFAULT;
  md->nborder    = NBORDER_DEFAULT;
  md->nnbmatch   = NNBMATCH_DEFAULT;
  md->maskweight = MASKWEIGHT_DEFAULT;
  md->corrlevel  = LEVEL_DEFAULT;
  md->corrbroken = BROKEN_LEVEL_DEFAULT;
  md->overlap    = OVERLAP_DEFAULT;
  md->nb_lap     = NB_LAP_DEFAULT;
  md->interleave = INTERLEAVE_DEFAULT;
}


void
init_rgb_coef(rgb_coef *rgb)
{
  assert(rgb);

  rgb->r = rgb->g = rgb->b = 0.0;
}


void
xdefault_rgb_coef(rgb_coef *rgb)
{
  assert(rgb);

  rgb->r = rgb->g = rgb->b = RGB_NORM_DEFAULT;
}


void 
init_cfg_def(cfg_def *cd) 
{
  assert(cd);

  cd->subst_rb = NULL;                                     /* zero pointer */
  cd->subst_nb = NULL;                                     /* zero pointer */
  cd->tps_id   = NULL;                                     /* zero pointer */

  cd->report_write      = 0;
  cd->rb_chain          = RB_CHAIN_DEFAULT;
  cd->histo_write       = 0;
  cd->verbose           = 0;
  cd->raise             = 0;
  cd->tips              = 0;
  cd->match_method      = MATCHING_METHOD_FIRST;
  cd->line_method       = METHOD_FIRST;
  cd->renew_method      = 0;
  cd->timestamp         = 0;
  cd->writeauto         = 0;
  cd->cdomain           = 0;
  cd->syspc             = 0.0;
  cd->rawbit            = 0;
  cd->nb_border_x       = 0;
  cd->nb_border_y       = 0;
  cd->duration          = 0;
  cd->support_dots      = 0;
  cd->support_dots_auto = 0;
  cd->overwrite_image   = 0;
  cd->replay_write      = 0;
  cd->training_write    = 0;
  cd->single_line       = 0;
  cd->single_nb         = 0;
  cd->dc_store          = 0;
  cd->cluster_force     = 0;
  cd->cluster_write     = 0;
  cd->cluster_nb        = CLUSTER_NB_NEW;
  cd->histo_thold       = 0;
  cd->cluster_level     = 0.0;
  cd->cluster_step      = 0.0;
  cd->cluster_thold     = 0;

  init_kern_def(cd->ckernel); 
  init_rgb_coef(&cd->rgb_norm);
  init_batch_def(&cd->batch);
  init_int2_lst(&cd->dc_1d);
  init_int2_lst(&cd->dc_2d);
  init_match_def(&cd->mad);
}


void
xdefault_cfg_def(cfg_def *cd)
{
  assert(cd);

  cd->report_write      = REPORT_WRITE_DEFAULT;
  cd->rb_chain          = RB_CHAIN_DEFAULT;
  cd->histo_write       = HISTO_WRITE_DEFAULT;
  cd->verbose           = VERBOSE_DEFAULT;
  cd->raise             = RAISE_DEFAULT;
  cd->tips              = TIPS_DEFAULT;
  cd->match_method      = MATCHING_METHOD_1;                /* main method */
  cd->line_method       = METHOD_FIRST;            /* don't force a method */
  cd->renew_method      = RENEW_METHOD_DEFAULT;
  cd->timestamp         = TIMESTAMP_DEFAULT;
  cd->writeauto         = WRITEAUTO_DEFAULT;
  cd->cdomain           = DOMAIN_DEFAULT;
  cd->syspc             = SYSPC_DEFAULT;
  cd->rawbit            = RAWBIT_DEFAULT;
  cd->nb_border_x       = NB_BORDER_DEFAULT;
  cd->nb_border_y       = NB_BORDER_DEFAULT;
  cd->duration          = DURATION_DEFAULT;
  cd->support_dots      = SUPPORT_DOTS_DEFAULT;
  cd->support_dots_auto = SUPPORT_DOTS_AUTO_DEFAULT;
  cd->overwrite_image   = OVERWRITE_IMAGE_DEFAULT;
  cd->replay_write      = REPLAY_WRITE_DEFAULT;
  cd->training_write    = TRAINING_WRITE_DEFAULT;
  cd->single_line       = SINGLE_LINE_DEFAULT;
  cd->single_nb         = SINGLE_NB_DEFAULT;
  cd->dc_store          = DC_STORE_DEFAULT;
  cd->cluster_force     = CLUSTER_FORCE_DEFAULT;
  cd->cluster_write     = CLUSTER_WRITE_DEFAULT;
  cd->cluster_nb        = CLUSTER_NB_NEW;
  cd->histo_thold       = THOLD_DEFAULT;
  cd->cluster_level     = CLUSTER_LVL_DEFAULT;
  cd->cluster_step      = CLUSTER_STEP_DEFAULT;
  cd->cluster_thold     = CHOLD_DEFAULT;

  cd->subst_rb = default_subst_rb();
  cd->subst_nb = default_subst_nb();
  cd->tps_id   = default_tps_id();
  xdefault_rgb_coef(&cd->rgb_norm);
  xdefault_batch_def(&cd->batch);
  xdefault_match_def(&cd->mad);

  xdefault_kern_def(cd->ckernel); 
}


void
init_set_def(set_def *sd)
{
  assert(sd);

  sd->report_stream   = NULL;                               /* zero pointer */
  sd->image_filename  = NULL;                               /* zero pointer */
  sd->user_options    = NULL;                               /* zero pointer */
  sd->xprog           = NULL;                               /* zero pointer */
  sd->gtk_rc          = NULL;                               /* zero pointer */
  sd->config_filespec = NULL;                               /* zero pointer */
  sd->xpid            = 0;
  sd->dobatch         = FALSE;

  init_rb_lst(&sd->rblst);
  init_hst_def(&sd->grey_hist);
  init_hst_def(&sd->th_hist);
}


void
xdefault_set_def(set_def *sd)
{
  assert(sd);

  sd->xpid = (int)getpid();

  sd->config_filespec = default_config_basename(NULL);
}


void
init_cli_def(cli_def *ci)
{
  assert(ci);

  ci->help = 0;
  ci->license = 0;
}


void
init_area_def(area_def *ad)
{
  assert(ad);

  ad->width = 0;
  ad->height = 0;
}


void
init_idx_def(idx_def *id) 
{
  assert(id);

  id->tps_idx = -1;
  id->cls_idx = -1;
  id->nb_idx = -1;
}


void
init_c_def(c_def *cd)
{
  assert(cd);

  cd->x = -1;
  cd->y = -1;
}


void 
init_cluster_def(cluster_def *cd)
{
  guint kernel[CKERNEL_SIZE] = CKERNEL_DEFAULT;
  guint i;

  assert(cd);

  cd->ctyp  = CLUSTER_NB_NEW;
  cd->cwrit = 0;
  cd->cforc = 0;
  cd->clvl  = CLUSTER_LVL_DEFAULT;
  cd->cstp  = CLUSTER_STEP_DEFAULT;
  cd->cthd  = CHOLD_DEFAULT;

  for (i = 0; i < CKERNEL_SIZE; i++) {
    cd->cknl[i] = kernel[i];
  }
}


void 
init_key_def(key_def *kd)
{
  assert(kd);

  kd->key = 0;
  kd->wac = 0;
  kd->cmt = NULL;                                          /* zero pointer */
  kd->aux = NULL;                                          /* zero pointer */
}


void 
init_key_lst(key_lst *kl)
{
  assert(kl);

  kl->cnt = 0;
  kl->kar = 0;                    /* zero pointer */
}


void 
init_part_lst(part_lst *pl)
{
  assert(pl);

  pl->cnt = 0;
  pl->fp  = 0;                    /* zero pointer */
}


void 
init_sort_ilst(sort_ilst *sl)
{
  assert(sl);

  sl->cnt = 0;
  sl->il  = 0;                    /* zero pointer */
}


void 
init_sort_dlst(sort_dlst *sl)
{
  assert(sl);

  sl->cnt = 0;
  sl->dl  = 0;                    /* zero pointer */
}


void 
init_hst_def(hst_def *hd) 
{
  assert(hd);

  hd->maxval = 0.0;
  hd->max_bin = 0;
  hd->th_bin = THOLD_DEFAULT;                                       /* n/a */
  hd->start_bin = 0;
  hd->set_cnt = 0;

  init_double_lst(&hd->hst);
}


void 
init_stc_def(stc_def *st)
{
  assert(st);

  st->mv = 0.0;
  st->sd = 0.0;
}


void 
init_hst_info(hst_info *hi)
{
  assert(hi);

  init_stc_def(&hi->st); 
  init_hst_def(&hi->hd); 
}


void 
init_int_lst(int_lst *il) 
{
  assert(il);

  il->cnt = 0;
  il->nr  = 0;                /* zero pointer */
}


void 
init_double_lst(double_lst *dl) 
{
  assert(dl);

  dl->cnt = 0;
  dl->dbl = 0;                /* zero pointer */
}


void 
init_int2_lst(int2_lst *i2l) 
{
  assert(i2l);

  i2l->cnt = 0;
  i2l->n   = 0;                /* zero pointer */
}


void 
init_text_def(text_def *td) 
{
  assert(td);

  td->cnt = 0;
  td->tl  = 0;                /* zero pointer */
}


void 
init_text_ring(text_ring *tr) 
{
  assert(tr);

  tr->idx = -1;
  init_text_def(&tr->td);
}


void 
init_int_lstm(int_lstm *im) 
{
  assert(im);

  im->cnt = 0;
  im->il  = 0;                /* zero pointer */
}


void 
init_pic_dbl(pic_dbl *pd) 
{
  assert(pd);

  pd->ar.width  = 0;
  pd->ar.height = 0;
  pd->mem       = 0;          /* zero pointer */
}


void 
init_inh_def(inh_def *id) 
{
  assert(id);

  id->inh = INHIBITED_FALSE;
  init_sector_def(&id->isd);
}


void 
init_pic_def(pic_def *pd) 
{
  assert(pd);

  pd->org.x     = 0;
  pd->org.y     = 0;
  pd->ar.width  = 0;             /* retain 0 */
  pd->ar.height = 0;             /* retain 0 */
  pd->mem       = NULL;          /* zero pointer */
  pd->bvl       = NULL;          /* zero pointer */
  pd->col       = NULL;          /* zero pointer */
}


void 
init_img_ext(img_ext *ie) 
{
  assert(ie);

  init_aux_lst(&ie->il);
  init_pic_def(&ie->eimg);
}


void 
init_img_typ(img_typ *it) 
{
  assert(it);

  it->typ = PNM_FORMAT_UNDEF;

  init_pic_def(&it->pimg);
}


/* Returns state of structure.

Revision: 15.01.2005 E.Sack

Input : - structure
Output: - state of structure
Postcondition: - this function checks the consistency of the structure as 
                 well because this is required for the state.
Usage:  - used whenever a structure is checked for its state. This prevents 
          the test of structure members what is highly casual.
*/
DATA_STATE
state_pic_def(pic_def pd) 
{
  PICT pix;
  int  org = pd.org.x || pd.org.y;
  int  arsiz = AREA_SIZE(pd.ar);
  int  d_is_def = pd.org.x || pd.org.y || pd.ar.width || pd.ar.height;
  int  p_is_null = (pd.mem == NULL) && (pd.bvl == NULL) && (pd.col == NULL);

  if (p_is_null) {
    assert(!d_is_def);                                   /* data hierarchy */
    assert(arsiz == 0);                                  /* data hierarchy */
    assert(org == 0);                                    /* data hierarchy */
    return STATE_IS_INIT;
  }

  assert(arsiz > 0);                                     /* data hierarchy */

  if (pd.mem != NULL) {
    XPSWAP(pix,pd.mem[0],pd.mem[arsiz-1]);               /* data hierarchy */
    XPSWAP(pix,pd.mem[arsiz-1],pd.mem[0]);               /* reversion      */
  }

  if (pd.bvl != NULL) {
    XPSWAP(pix,pd.bvl[0],pd.bvl[arsiz-1]);               /* data hierarchy */
    XPSWAP(pix,pd.bvl[arsiz-1],pd.bvl[0]);               /* reversion      */
  }

  if (pd.col != NULL) {
    XPSWAP(pix,pd.col[0].r,pd.col[arsiz-1].r);           /* data hierarchy */
    XPSWAP(pix,pd.col[arsiz-1].r,pd.col[0].r);           /* reversion      */
  }

  return STATE_IS_DEFINED_PARTIALLY;
}


void 
init_col_def(col_def *cl) 
{
  assert(cl);

  cl->r = 0;
  cl->g = 0;
  cl->b = 0;
}


void 
init_period_def(period_def *pd)
{
  assert(pd);

  pd->l_phase = 0;
  pd->h_phase = 0;
}


void 
init_linespacing_par(linespacing_par *lp)
{
  assert(lp);

  init_period_def(&lp->period);

  lp->opos = -1;
}


void 
init_inline_def(inline_def *il)
{
  assert(il);

  il->method = METHOD_FIRST;
  
  init_linespacing_par(&il->linespacing);
}


void 
init_media_att(media_att *md) 
{
  assert(md);

  init_inline_def(&md->in_line);

  md->dpi = 0;
}


void 
init_cls_lst(cls_lst *cl) 
{
  assert(cl);

  cl->cnt = 0;
  cl->cls = 0;             /* zero pointer */
}


void 
init_tps_lst(tps_lst *tpl) 
{
  assert(tpl);

  tpl->cnt = 0;
  tpl->tps = 0;             /* zero pointer */
}


DATA_STATE
state_tps_lst(tps_lst tpl) 
{
  if (tpl.tps == NULL) {
    assert(!tpl.cnt);                                  /* data consistency */

    return STATE_IS_INIT;
  }

  assert(tpl.cnt >= 0);                                /* data consistency */

  return STATE_IS_DEFINED_FULLY;
}


void 
init_nb_lst(nb_lst *nl) 
{
  assert(nl);

  nl->cnt = 0;
  nl->nb  = 0;             /* zero pointer */
}


void 
init_aux_lst(aux_lst *al) 
{
  assert(al);

  al->cnt = 0;
  al->dat = 0;             /* zero pointer */
}


void 
init_rb_chn_lst(rb_chn_lst *rcl) 
{
  assert(rcl);

  rcl->cnt = 0;
  rcl->rcd = 0;             /* zero pointer */
}


void 
init_nb_chn_lst(nb_chn_lst *ncl) 
{
  assert(ncl);

  ncl->cnt = 0;
  ncl->cd  = 0;             /* zero pointer */
}


void 
init_nb_chn_def(nb_chn_def *ncd) 
{
  assert(ncd);

  ncd->cdx = 0;
  ncd->ndx = 0;
  ncd->x = 0;
  ncd->y = 0;
}


void 
init_cc_lst(cc_lst *cl) 
{
  assert(cl);

  cl->cnt = 0;
  cl->cc  = 0;             /* zero pointer */
}


void 
init_tps_def(tps_def *tps) 
{
  assert(tps);

  tps->ident = 0;          /* zero pointer */

  init_media_att(&tps->media);
  init_cls_lst(&tps->cl);
}


void 
init_ival_def(ival_def *ivl) 
{
  assert(ivl);

  ivl->start = 0;                 /* interval length zero and minimum-safe */
  ivl->end   = -1;                /* interval length zero and maximum-safe */
}


void 
init_ival_lst(ival_lst *il) 
{
  assert(il);

  il->cnt = 0;
  il->ival = 0;                                           /* zero pointer */
}


void 
init_sector_def(sector_def *sec) 
{
  assert(sec);

  init_ival_def(&sec->line);
  init_ival_def(&sec->column);
}


void 
init_sector_lst(sector_lst *sl) 
{
  assert(sl);

  sl->cnt = 0;
  sl->sdl = 0;                                           /* zero pointer */
}


void 
init_cls_att(cls_att *ca) 
{
  assert(ca);

  ca->class = 0;                                          /* zero pointer */
  ca->sign = 0;                                           /* zero pointer */
  ca->crem = 0;                                           /* zero pointer */

  ca->alignment = UDEFLINE;
  ca->printable = 0;
  ca->fixed     = 0;
}


void 
init_inh_lst(inh_lst *il) 
{
  assert(il);

  il->cnt = 0;
  il->ind = 0;                                            /* zero pointer */
}


DATA_STATE
state_inh_lst(inh_lst il) 
{
  if (il.ind == NULL) {
    assert(!il.cnt);                                   /* data consistency */

    return STATE_IS_INIT;
  }

  assert(il.cnt >= 0);                                 /* data consistency */

  return STATE_IS_DEFINED_FULLY;
}


void 
init_cls_def(cls_def *cls) 
{
  assert(cls);

  init_cls_att(&cls->att);
  init_nb_lst(&cls->nbl);
  init_cc_lst(&cls->ccl);
  init_inh_lst(&cls->inl);
}


void
init_nb_ref(nb_ref *nref) 
{
  assert(nref);

  init_pic_def(&nref->pic);
  init_pic_def(&nref->fld);
  init_pic_def(&nref->wtd);
  init_pic_def(&nref->msk);
  init_pic_def(&nref->dlt);
  nref->nbm = 0.0;
  nref->fldm = 0.0;
  nref->dltm = 0.0;
  nref->wtdm = 0.0;
  nref->mskm = 0.0;
  nref->cmin = 0.0;
}


void 
init_nb_def(nb_def *nb) 
{
  assert(nb);

  init_pic_def(&nb->pat);
  init_nb_chn_lst(&nb->cnb);
  init_nb_chn_lst(&nb->nnb);
  init_nb_ref(&nb->ref);
  init_nb_aux(&nb->dat);

  nb->aux = XPLAB_INIT_AUX_NB_DEF;
}


void 
init_nb_aux(nb_aux *na) 
{
  assert(na);

  na->cidx = na->nidx = XPLAB_IDX_NONE;

  init_c_def(&na->afs);
  init_sector_def(&na->sec);
}


void 
init_pattern_color(pattern_color *pc)
{
  assert(pc);

  init_col_def(&pc->bgd_in);
  init_col_def(&pc->bgd_out);
  init_col_def(&pc->fgd);
  init_col_def(&pc->sel);
  init_col_def(&pc->sbl);
  init_col_def(&pc->tnd);
  init_col_def(&pc->res);
  init_col_def(&pc->ign);
}


void 
init_cc_def(cc_def *cc) 
{
  assert(cc);

  cc->ar.width  = 0;
  cc->ar.height = 0;
  cc->patt      = NULL;                                    /* zero pointer */
}


void 
init_rb_lst(rb_lst *rl) 
{
  assert(rl);

  rl->rar.width  = 0;
  rl->rar.height = 0;
  rl->cnt        = 0;
  rl->rb         = NULL;                                   /* zero pointer */
}


void 
init_rb_chn_def(rb_chn_def *rc) 
{
  assert(rc);

  rc->chain_name = 0;             /* zero pointer */

  init_sector_def(&rc->chain_frame);
  init_rb_lst(&rc->rb_chain);
  init_cand_lst(&rc->bhclst);
  rc->caux = 0;
}


void 
init_tiles_def(tiles_def *rl) 
{
  assert(rl);

  init_rb_chn_def(&rl->doc);

  rl->border   = 0;

  init_rb_chn_lst(&rl->rch);
}


void 
init_baseline_lst(baseline_lst *bl) 
{
  assert(bl);

  bl->lspace = BASELINE_INIT;

  init_int_lst(&bl->bline);
}


void 
init_rb_def(rb_def *rb) 
{
  assert(rb);

  rb->raux = RB_STATUS_INIT;
  rb->rld = RB_STATUS_INIT;

  init_sector_def(&rb->sc);
  init_ival_lst(&rb->gridy);
}


void 
init_h_contour_lst(h_contour_lst *ctr) 
{
  assert(ctr);

  ctr->type = CONTOUR_UNDEF;
  ctr->cnt = 0;
  ctr->hcl = 0;               /* zero pointer */
}


void 
init_v_contour_lst(v_contour_lst *ctr) 
{
  assert(ctr);

  ctr->type = CONTOUR_UNDEF;
  ctr->cnt = 0;
  ctr->vcl = 0;               /* zero pointer */
}


void 
init_db_def(db_def *db) 
{
  assert(db);

  db->db_fname = 0;            /* zero pointer */
  db->db_version = 0;          /* zero pointer */

  init_tps_lst(&db->tpl);
}


DATA_STATE
state_db_def(db_def db) 
{
  DATA_STATE state_tpl = state_tps_lst(db.tpl);

  if ((db.db_fname == NULL) && (db.db_version == NULL) && 
      (state_tpl == STATE_IS_INIT)) {
    return STATE_IS_INIT;
  }

  assert(db.db_fname != NULL);                           /* data hierarchy */
  assert(db.db_version != NULL);                         /* data hierarchy */

  if (state_tpl == STATE_IS_INIT) {
    return STATE_IS_DEFINED_PARTIALLY;
  }

  return STATE_IS_DEFINED_FULLY;
}


void 
init_h_contour(h_contour *cln) 
{
  assert(cln);

  cln->yco = 0;
  init_ival_def(&cln->xil);
}


void 
init_v_contour(v_contour *cln) 
{
  assert(cln);

  cln->xco = 0;
  init_ival_def(&cln->yil);
}


void 
init_cand_lst(cand_lst *kl) 
{
  assert(kl);

  kl->cnt = 0;
  kl->c   = 0;                  /* zero pointer */
}


void 
init_cv_def_i(cv_def_i *kd) 
{
  assert(kd);

  kd->cv = 0.0;
  kd->x = 0;
  kd->y = 0;
}


void 
init_cand_def(cand_def *kr) 
{
  assert(kr);

  kr->mtd = MATCHING_METHOD_FIRST;

  init_cv_def_i(&kr->m);
  kr->rbn = 0;
  kr->n = 0;
  kr->c = 0;

  init_ival_def(&kr->line);

  init_sector_def(&kr->roi);
  init_sector_def(&kr->ext);
  init_sector_def(&kr->knl);
  init_sector_def(&kr->hdx);

  init_int_lst(&kr->crb);

  kr->alt = 0;               /* zero pointer */

  kr->aux = XPLAB_INIT_AUX_CAND_DEF;
}


void
init_reg_def(reg_def *rd) 
{
  assert(rd);

  rd->cnt = 0;
  rd->yrl = 0;               /* zero pointer */
}


void
init_y_reg_lst(y_reg_lst *yrl) 
{
  assert(yrl);

  yrl->cnt = 0;
  yrl->yrd = 0;               /* zero pointer */
}


void
init_y_reg_def(y_reg_def *yrd) 
{
  assert(yrd);

  yrd->y_range.start = 0;
  yrd->y_range.end   = 0;

  yrd->use = NOT_REGISTERED;

  yrd->pic_state = 0;                                  /* zero pointer */

  init_cand_lst(&yrd->vlst);
}


void
init_hash_lst(hash_lst *cdef) 
{
  assert(cdef);

  cdef->cnt = 0;
  cdef->nid = 0;               /* zero pointer */
}


void
init_kwh_lst(kwh_lst *kl) 
{
  assert(kl);

  kl->cnt = 0;
  kl->kwh = 0;               /* zero pointer */
}


void
init_rb_ref(rb_ref *rref) 
{
  assert(rref);

  rref->rpic = 0;               /* zero pointer */
  init_pic_def(&rref->wpic);
  init_pic_def(&rref->shdw1);
  init_pic_def(&rref->shdw2);
  init_kwh_lst(&rref->kl);
  init_reg_def(&rref->xreg);
  rref->pic_state = 0;           /* zero pointer */
}


void
init_cand_lst_ary(cand_lst_ary *cla) 
{
  assert(cla);

  cla->cnt  = 0;
  cla->clst = 0;                                     /* zero pointer */
}


void 
init_rc_def(rc_def *rc) 
{
  assert(rc);

  init_sector_def(&rc->sid);

  rc->k                = 0.0;

  rc->o.line.start     = INT_MAX;
  rc->o.column.start   = INT_MAX;
  rc->o.line.end       = INT_MIN;
  rc->o.column.end     = INT_MIN;

  rc->g                = 0;

  rc->lc               = 0;
  rc->ln               = 0;

  init_sector_def(&rc->ls);

  init_cand_lst(&rc->r);
}


void 
init_rc_lst(rc_lst *rl)
{
  assert(rl);

  rl->cnt = 0;
  rl->s   = 0;                  /* zero pointer */
}


void 
init_gen_lst(gen_lst *gl)
{
  assert(gl);

  gl->cnt = 0;
  gl->gd  = 0;                  /* zero pointer */
}


void 
init_gen_def(gen_def *dl)
{
  assert(dl);

  dl->cnt = 0;
  dl->c   = 0;                  /* zero pointer */
  dl->n   = 0;                  /* zero pointer */
  dl->hl  = 0;                  /* zero pointer */
}


void
init_hash_def(hash_def *hdef)
{
  assert(hdef);

  hdef->cnt = 0;
  hdef->cid = 0;                  /* zero pointer */
  hdef->pid = 0;                  /* zero pointer */
  hdef->kr  = 0;                  /* zero pointer */
}


void
init_fft_def(fft_def *fd)
{
  assert(fd);

  fd->ar.width = 0;
  fd->ar.height = 0;
  fd->data = 0;                    /* zero pointer */
  fd->spec = 0;                    /* zero pointer */
}


void
init_kwh_def(kwh_def *kw)
{
  assert(kw);

  kw->ar.width = 0;
  kw->ar.height = 0;
  kw->ll = 0;                     /* zero pointer */
}


void
init_lag_lst(lag_lst *ll)
{
  assert(ll);

  ll->cnt = 0;
  ll->ld  = 0;                                              /* zero pointer */
}


void 
copy_kern_def(kern_def src, kern_def dst) 
{
  gint i;

  for (i = 0; i < CKERNEL_SIZE; i++) {
    dst[i] = src[i];
  }
}


void
copy_int_array(int *src, int *dst, int len)
{
  assert(len>=0);
  assert(src);
  assert(dst);

   while (len--) {
    *dst++ = *src++;
  }
}


void
copy_rb_chn_def(rb_chn_def *src, rb_chn_def *dst)
{
  int len;

  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_rb_chn_def(dst);

  if (src->chain_name) {
    len = strlen(src->chain_name);
    assert(len > 0);                                 /* must be consistent */

    add_string(src->chain_name, &dst->chain_name);
  }

  dst->chain_frame = src->chain_frame;
  cat_rb_lst(src->rb_chain, &dst->rb_chain);
  cat_cand_lst(&src->bhclst, &dst->bhclst);
  dst->caux = src->caux;
}


void
copy_fft_def(fft_def src, fft_def dst)
{
  int     i;
  double  *iptr, *optr;

  assert(src.ar.width == dst.ar.width);
  assert(src.ar.height == dst.ar.height);

  iptr = src.spec;
  optr = dst.spec;
  for (i = 0; i < 2*dst.ar.height; i++) {
    *optr++ = *iptr++;
  }

  iptr = &src.data[0][0];
  optr = &dst.data[0][0];
  for (i = 0; i < dst.ar.width*dst.ar.height; i++) {
    *optr++ = *iptr++;                                /* array without gap */
  }
}


void 
copy_media_att(media_att src, media_att *dst) 
{   
  assert(dst);

  *dst = src;
}


void 
copy_cls_lst(cls_lst *src, cls_lst *dst, gboolean no_chain) 
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_cls_lst(dst);

  if (!src->cls) {                                          /* may be empty */
    assert(!src->cnt);                                /* must be consistent */
  }
  else {
    assert(src->cnt>0);                               /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(cls_def, dst->cnt, dst->cls);

    for (i = 0; i < dst->cnt; i++) {
      init_cls_def(&dst->cls[i]);
      copy_cls_def(&src->cls[i], &dst->cls[i], no_chain);
    }
  }
}


void 
copy_key_lst(key_lst *src, key_lst *dst) 
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_key_lst(dst);

  if (!src->kar) {                                          /* may be empty */
    assert(!src->cnt);                                /* must be consistent */
  }
  else {
    assert(src->cnt>0);                               /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(key_def, dst->cnt, dst->kar);

    for (i = 0; i < dst->cnt; i++) {
      init_key_def(&dst->kar[i]);
      copy_key_def(&src->kar[i], &dst->kar[i]);
    }
  }
}


void 
copy_cand_lst(cand_lst *src, cand_lst *dst) 
{
  int i;

  if (!src) {
    return;
  }

  assert(dst);
  assert(src != dst);

  free_cand_lst(dst);

  if (!src->c) {
    assert(!src->cnt);
  }
  else {
    assert(src->cnt>0);

    dst->cnt = src->cnt;
    CALLOC(cand_def, dst->cnt, dst->c);

    for (i = 0; i < dst->cnt; i++) {
      init_cand_def(&dst->c[i]);
      copy_cand_def(&src->c[i], &dst->c[i]);
    }
  }
}


void 
copy_rc_lst(rc_lst *src, rc_lst *dst) 
{
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_rc_lst(dst);

  if (!src->s) {                                     /* consistency check */
    assert(!src->cnt);
  }
  else {
    assert(src->cnt>0);

    dst->cnt = src->cnt;
    CALLOC(rc_def, dst->cnt, dst->s);

    for (i = 0; i < dst->cnt; i++) {
      init_rc_def(&dst->s[i]);
      copy_rc_def(&src->s[i], &dst->s[i]);
    }
  }
}


void
copy_rc_def(rc_def *src, rc_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);

  free_rc_def(dst);

  copy_sector_def(&src->sid, &dst->sid);

  dst->k = src->k;

  copy_sector_def(&src->o, &dst->o);

  dst->g  = src->g;
  dst->lc = src->lc;
  dst->ln = src->ln;

  copy_sector_def(&src->ls, &dst->ls);

  copy_cand_lst(&src->r, &dst->r);
}


void 
copy_db_def(db_def *src, db_def *dst, gboolean no_chain)
{   
  assert(src);
  assert(dst);
  assert(src != dst);

  free_db_def(dst);

  add_string(src->db_fname, &dst->db_fname);
  add_string(src->db_version, &dst->db_version);

  copy_tps_lst(&src->tpl, &dst->tpl, no_chain);
}


void 
copy_tps_lst(tps_lst *src, tps_lst *dst, gboolean no_chain)
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_tps_lst(dst);

  if (!src->tps) {                                          /* may be empty */
    assert(!src->cnt);                                /* must be consistent */
  }
  else {
    assert(src->cnt>0);                               /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(tps_def, dst->cnt, dst->tps);

    for (i = 0; i < dst->cnt; i++) {
      init_tps_def(&dst->tps[i]);
      copy_tps_def(&src->tps[i], &dst->tps[i], no_chain);
    }
  }
}


void 
copy_tps_def(tps_def *src, tps_def *dst, gboolean no_chain) 
{
  assert(src);
  assert(dst);
  assert(src != dst);

  free_tps_def(dst);

  add_string(src->ident, &dst->ident);

  copy_media_att(src->media, &dst->media);
  copy_cls_lst(&src->cl, &dst->cl, no_chain);
}


int
crop_double_lst(double_lst src, double_lst dst, int src_offset, 
		int dst_offset, int cnt)
{
  int s, d, copied = 0;

  assert(src.dbl);
  assert(dst.dbl);
  assert(src.cnt > 0);
  assert(dst.cnt > 0);
  assert(cnt > 0);

  assert(dst.cnt >= cnt+dst_offset);
  assert(src.cnt >= cnt+src_offset);

  s = src_offset;
  d = dst_offset;

  while (s < src.cnt && d < dst.cnt && copied < cnt) {
    dst.dbl[d++] = src.dbl[s++];
    copied++;
  }

  return copied;
}


int
copy_double(double *src, double *dst, int cnt)
{
  int copied = 0;

  assert(src);
  assert(dst);
  assert(cnt >= 0);

  while (copied < cnt) {
    *dst++ = *src++;
    copied++;
  }

  return copied;
}


void 
copy_nb_lst(nb_lst *src, nb_lst *dst, gboolean no_chain) 
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_nb_lst(dst);

  if (!src->nb) {                                     /* consistency check */
    assert(!src->cnt);
  }
  else {
    assert(src->cnt>0);

    dst->cnt = src->cnt;
    CALLOC(nb_def, dst->cnt, dst->nb);

    for (i = 0; i < dst->cnt; i++) {
      init_nb_def(&dst->nb[i]);
      copy_nb_def(&src->nb[i], &dst->nb[i], no_chain);
    }
  }
}


void 
copy_sector_lst(sector_lst *src, sector_lst *dst) 
{
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_sector_lst(dst);

  if (!src->sdl) {
    assert(!src->cnt);
  }
  else {
    assert(src->cnt>0);

    dst->cnt = src->cnt;
    CALLOC(sector_def, dst->cnt, dst->sdl);

    for (i = 0; i < dst->cnt; i++) {
      init_sector_def(&dst->sdl[i]);
      copy_sector_def(&src->sdl[i], &dst->sdl[i]);
    }
  }
}


void 
copy_inh_lst(inh_lst *src, inh_lst *dst) 
{
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_inh_lst(dst);

  if (!src->ind) {
    assert(!src->cnt);
  }
  else {
    assert(src->cnt>0);

    dst->cnt = src->cnt;
    CALLOC(inh_def, dst->cnt, dst->ind);

    for (i = 0; i < dst->cnt; i++) {
      init_inh_def(&dst->ind[i]);
      copy_inh_def(&src->ind[i], &dst->ind[i]);
    }
  }
}


void 
copy_cc_lst(cc_lst *src, cc_lst *dst) 
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_cc_lst(dst);

  if (!src->cc) {                                          /* may be empty */
    assert(!src->cnt);                               /* must be consistent */
  }
  else {
    assert(src->cnt>0);                              /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(cc_def, dst->cnt, dst->cc);

    for (i = 0; i < dst->cnt; i++) {
      init_cc_def(&dst->cc[i]);
      copy_cc_def(&src->cc[i], &dst->cc[i]);
    }
  }
}


void 
copy_nb_chn_def(nb_chn_def *src, nb_chn_def *dst) 
{   
  assert(src);
  assert(dst);

  *dst = *src;
}


void 
copy_nb_chn_lst(nb_chn_lst *src, nb_chn_lst *dst) 
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_nb_chn_lst(dst);

  if (!src->cd) {                                          /* may be empty */
    assert(!src->cnt);                               /* must be consistent */
  }
  else {
    assert(src->cnt>0);                              /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(nb_chn_def, dst->cnt, dst->cd);            /* pointers are NULL */

    for (i = 0; i < dst->cnt; i++) {
      init_nb_chn_def(&dst->cd[i]);
      copy_nb_chn_def(&src->cd[i], &dst->cd[i]);
    }
  }
}


void 
copy_double_lst(double_lst *src, double_lst *dst)
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_double_lst(dst);

  if (!src->dbl) {                                          /* may be empty */
    assert(!src->cnt);                                /* must be consistent */
  }
  else {
    assert(src->cnt>0);                               /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(double, dst->cnt, dst->dbl);

    for (i = 0; i < dst->cnt; i++) {
      dst->dbl[i] = src->dbl[i];
    }
  }
}


void 
copy_int_lst(int_lst *src, int_lst *dst)
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_int_lst(dst);

  if (!src->nr) {                                          /* may be empty */
    assert(!src->cnt);                               /* must be consistent */
  }
  else {
    assert(src->cnt>0);                              /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(int, dst->cnt, dst->nr);

    for (i = 0; i < dst->cnt; i++) {
      dst->nr[i] = src->nr[i];
    }
  }
}


void 
copy_int2_lst(int2_lst *src, int2_lst *dst)
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_int2_lst(dst);

  if (!src->n) {                                           /* may be empty */
    assert(!src->cnt);                               /* must be consistent */
  }
  else {
    assert(src->cnt>0);                              /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(int2, dst->cnt, dst->n);

    for (i = 0; i < dst->cnt; i++) {
      dst->n[i] = src->n[i];
    }
  }
}


void 
copy_ival_lst(ival_lst *src, ival_lst *dst)
{   
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  free_ival_lst(dst);

  if (!src->ival) {                                        /* may be empty */
    assert(!src->cnt);                               /* must be consistent */
  }
  else {
    assert(src->cnt>0);                              /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(ival_def, dst->cnt, dst->ival);

    for (i = 0; i < dst->cnt; i++) {
      dst->ival[i] = src->ival[i];
    }
  }
}


void 
copy_cls_def(cls_def *src, cls_def *dst, gboolean no_chain) 
{   
  assert(src);
  assert(dst);
  assert(src != dst);

  free_cls_def(dst);

  copy_cls_att(src->att, &dst->att);
  copy_nb_lst(&src->nbl, &dst->nbl, no_chain);
  copy_cc_lst(&src->ccl, &dst->ccl);
  copy_inh_lst(&src->inl, &dst->inl);
}


void
copy_cls_att(cls_att src, cls_att *dst)
{
  assert(dst);

  FREEMEM(dst->class);
  add_string(src.class, &dst->class);

  FREEMEM(dst->sign);
  add_string(src.sign, &dst->sign);

  FREEMEM(dst->crem);
  add_string(src.crem, &dst->crem);

  dst->alignment = src.alignment;
  dst->printable = src.printable;
  dst->fixed     = src.fixed;
}


void
copy_img_typ(img_typ *src, img_typ *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  dst->typ = src->typ;

  copy_pic_def(&src->pimg, &dst->pimg);
}


void
copy_pic_def(pic_def *src, pic_def *dst)
{
  int i, len;

  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  len = AREA_SIZE(src->ar);

  free_pic_def(dst);

  dst->org = src->org;
  dst->ar = src->ar;

  if (src->mem) {                                           /* may be empty */
    CALLOC(PICT, len, dst->mem);
    for (i = 0; i < len; i++) {
      dst->mem[i] = src->mem[i];
    }
  }

  if (src->bvl) {                                           /* may be empty */
    CALLOC(PICT, len, dst->bvl);
    for (i = 0; i < len; i++) {
      dst->bvl[i] = src->bvl[i];
    }
  }

  if (src->col) {                                           /* may be empty */
    CALLOC(col_def, len, dst->col);
    for (i = 0; i < len; i++) {
      dst->col[i] = src->col[i];
    }
  }
}


img_typ
xdup_img_typ(img_typ src)
{
  img_typ dupi;

  init_img_typ(&dupi);

  dupi.typ = src.typ;
  dupi.pimg = xdup_pic_def(src.pimg);

  return dupi;
}


pic_def
xdup_pic_def(pic_def src)
{
  int     i, len;
  pic_def dupp;

  init_pic_def(&dupp);

  len = AREA_SIZE(src.ar);

  dupp.org = src.org;
  dupp.ar = src.ar;

  if (src.mem) {                                           /* may be empty */
    CALLOC(PICT, len, dupp.mem);
    for (i = 0; i < len; i++) {
      dupp.mem[i] = src.mem[i];
    }
  }

  if (src.bvl) {                                           /* may be empty */
    CALLOC(PICT, len, dupp.bvl);
    for (i = 0; i < len; i++) {
      dupp.bvl[i] = src.bvl[i];
    }
  }

  if (src.col) {                                           /* may be empty */
    CALLOC(col_def, len, dupp.col);
    for (i = 0; i < len; i++) {
      dupp.col[i] = src.col[i];
    }
  }

  return dupp;
}


void
copy_nb_ref(nb_ref *src, nb_ref *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  copy_pic_def(&src->pic, &dst->pic);
  copy_pic_def(&src->fld, &dst->fld);
  copy_pic_def(&src->wtd, &dst->wtd);
  copy_pic_def(&src->msk, &dst->msk);
  copy_pic_def(&src->dlt, &dst->dlt);

  dst->nbm = src->nbm;
  dst->fldm = src->fldm;
  dst->dltm = src->dltm;
  dst->wtdm = src->wtdm;
  dst->mskm = src->mskm;
  dst->cmin = src->cmin;
}


void 
copy_nb_def(nb_def *src, nb_def *dst, gboolean no_chain) 
{
  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  free_nb_def(dst);

  copy_pic_def(&src->pat, &dst->pat);

  if (no_chain == FALSE) {
    copy_nb_chn_lst(&src->cnb, &dst->cnb);
    copy_nb_chn_lst(&src->nnb, &dst->nnb);
  }                                             /* else no NB chains wanted */

  copy_nb_ref(&src->ref, &dst->ref);
  copy_nb_aux(&src->dat, &dst->dat);

  dst->aux = src->aux;
}


void 
copy_nb_aux(nb_aux *src, nb_aux *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  free_nb_aux(dst);

  dst->cidx = src->cidx;
  dst->nidx = src->nidx;

  dst->afs = src->afs;
  dst->sec = src->sec;
}


void 
copy_cc_def(cc_def *src, cc_def *dst) 
{   
  int i, len;

  assert(src);
  assert(dst);
  assert(src != dst);         /* pointer required */

  len = XPMOD32(src->ar.width)*src->ar.height;

  free_cc_def(dst);

  dst->ar = src->ar;

  CALLOC(PICT, len, dst->patt);

  for (i = 0; i < len; i++) {
    dst->patt[i] = src->patt[i];
  }
}


void 
copy_baseline_lst(baseline_lst *src, baseline_lst *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);         /* pointer required */

  free_baseline_lst(dst);

  dst->lspace = src->lspace;

  copy_int_lst(&src->bline, &dst->bline);
}


void 
copy_rb_def(rb_def *src, rb_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);         /* pointer required */

  free_rb_def(dst);

  (*dst).raux = (*src).raux;
  (*dst).rld = (*src).rld;
  (*dst).sc = (*src).sc;

  copy_ival_lst(&(*src).gridy, &(*dst).gridy);
}


int                          /* interface to offer simple usage of casting */
compare_cand_def(const void *cd1, const void *cd2)
{
  cand_def c1 = *(const cand_def *)cd1;
  cand_def c2 = *(const cand_def *)cd2;

  if (c1.m.cv < c2.m.cv) {
    return -1;
  }
  else if (c1.m.cv > c2.m.cv) {
    return 1;
  }
  else {
    return 0;
  }
}


void 
copy_cand_def(cand_def *src, cand_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_cand_def(dst);

  dst->mtd = src->mtd;
  dst->m   = src->m;
  dst->rbn = src->rbn;
  dst->n   = src->n;
  dst->c   = src->c;

  dst->line = src->line;

  dst->roi = src->roi;
  dst->ext = src->ext;
  dst->knl = src->knl;
  dst->hdx = src->hdx;
 
  copy_int_lst(&src->crb, &dst->crb);

  if (src->alt) {
    if (!dst->alt) {
      CALLOC(cand_lst, 1, dst->alt);
      init_cand_lst(dst->alt);
    }

    copy_cand_lst(src->alt, dst->alt);
  }

  dst->aux = src->aux;
}


void 
copy_key_def(key_def *src, key_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);   /* pointer required */

  free_key_def(dst);

  dst->key = src->key;
  dst->wac = src->wac;
  dst->aux = src->aux;

  add_string(src->cmt, &dst->cmt);
}


void 
copy_inh_def(inh_def *src, inh_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);   /* pointer required */

  free_inh_def(dst);

  dst->inh = src->inh;

  copy_sector_def(&src->isd, &dst->isd);
}


void 
copy_sector_def(sector_def *src, sector_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  init_sector_def(dst);

  *dst = *src;
}


void
copy_prog_ctrl(prog_ctrl *src, prog_ctrl *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_prog_ctrl(dst);

  copy_set_def(&src->set, &dst->set);
  copy_cfg_def(&src->cfg, &dst->cfg);
}


void
copy_match_def(match_def *src, match_def *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_match_def(dst);

  dst->mborder    = src->mborder;                      /* explicitly coded */
  dst->nborder    = src->nborder;
  dst->nnbmatch   = src->nnbmatch;
  dst->maskweight = src->maskweight;
  dst->corrlevel  = src->corrlevel;
  dst->corrbroken = src->corrbroken;
  dst->overlap    = src->overlap;
  dst->nb_lap     = src->nb_lap;
  dst->interleave = src->interleave;
}


void
copy_cfg_def(cfg_def *src, cfg_def *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_cfg_def(dst);

  dst->report_write      = src->report_write;
  dst->rb_chain          = src->rb_chain;
  dst->histo_write       = src->histo_write;
  dst->verbose           = src->verbose;
  dst->raise             = src->raise;
  dst->tips              = src->tips;
  dst->match_method      = src->match_method;
  dst->line_method       = src->line_method;
  dst->renew_method      = src->renew_method;
  dst->timestamp         = src->timestamp;
  dst->writeauto         = src->writeauto;
  dst->cdomain           = src->cdomain;
  dst->syspc             = src->syspc;
  dst->rawbit            = src->rawbit;
  dst->nb_border_x       = src->nb_border_x;
  dst->nb_border_y       = src->nb_border_y;
  dst->dc_store          = src->dc_store;
  dst->duration          = src->duration;
  dst->support_dots      = src->support_dots;
  dst->support_dots_auto = src->support_dots_auto;
  dst->overwrite_image   = src->overwrite_image;
  dst->replay_write      = src->replay_write;
  dst->training_write    = src->training_write;
  dst->single_line       = src->single_line;
  dst->single_nb         = src->single_nb;
  dst->cluster_force     = src->cluster_force;
  dst->cluster_write     = src->cluster_write;
  dst->cluster_nb        = src->cluster_nb;
  dst->histo_thold       = src->histo_thold;
  dst->cluster_level     = src->cluster_level;
  dst->cluster_step      = src->cluster_step;
  dst->cluster_thold     = src->cluster_thold;
  dst->rgb_norm          = src->rgb_norm;

  add_string(src->subst_rb, &dst->subst_rb);
  add_string(src->subst_nb, &dst->subst_nb);
  add_string(src->tps_id, &dst->tps_id);

  copy_kern_def(src->ckernel, dst->ckernel); 
  copy_batch_def(&src->batch, &dst->batch);
  copy_int2_lst(&src->dc_1d, &dst->dc_1d);
  copy_int2_lst(&src->dc_2d, &dst->dc_2d);
  copy_match_def(&src->mad, &dst->mad);
}


void
copy_set_def(set_def *src, set_def *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                  /* pointer required */

  free_set_def(dst);

  dst->report_stream = src->report_stream;
  dst->xpid = src->xpid;
  dst->dobatch = src->dobatch;

  add_string(src->xprog, &dst->xprog);
  add_string(src->gtk_rc, &dst->gtk_rc);
  add_string(src->config_filespec, &dst->config_filespec);
  add_string(src->image_filename, &dst->image_filename);
  add_string(src->user_options, &dst->user_options);

  copy_rb_lst(&src->rblst, &dst->rblst);
  copy_hst_def(&src->grey_hist, &dst->grey_hist);
  copy_hst_def(&src->th_hist, &dst->th_hist);
}


void
copy_batch_def(batch_def *src, batch_def *dst)
{
  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  free_batch_def(dst);

  *dst = *src;
}


void
copy_rb_lst(rb_lst *src, rb_lst *dst)
{
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  free_rb_lst(dst);

  dst->rar = src->rar;

  if (!src->rb) {                                           /* may be empty */
    assert(!src->cnt);                                /* must be consistent */
  }
  else {
    assert(src->cnt>0);                               /* must be consistent */

    dst->cnt = src->cnt;
    CALLOC(rb_def, dst->cnt, dst->rb);

    for (i = 0; i < dst->cnt; i++) {
      init_rb_def(&dst->rb[i]);
      copy_rb_def(&src->rb[i], &dst->rb[i]);
    }
  }
}


void 
copy_hst_def(hst_def *src, hst_def *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);                                   /* pointer required */

  free_hst_def(dst);

  dst->maxval    = src->maxval;
  dst->max_bin   = src->max_bin;
  dst->th_bin    = src->th_bin;
  dst->start_bin = src->start_bin;
  dst->set_cnt   = src->set_cnt;

  copy_double_lst(&src->hst, &dst->hst);
}


void
move_double_lst(double_lst *src, double_lst *dst)
{
  assert(src->dbl);
  assert(src->cnt > 0);

  copy_double_lst(src, dst);

  free_double_lst(src);
}


void 
move_rc_lst(rc_lst *src, rc_lst *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);

  copy_rc_lst(src, dst);

  free_rc_lst(src);
}


void 
move_cls_def(cls_def *src, cls_def *dst, gboolean no_chain)
{
  assert(src);
  assert(dst);
  assert(src != dst);

  copy_cls_def(src, dst, no_chain);

  free_cls_def(src);
}


void 
move_nb_def(nb_def *src, nb_def *dst, gboolean no_chain) 
{   
  assert(src);
  assert(dst);
  assert(src != dst);

  copy_nb_def(src, dst, no_chain);

  free_nb_def(src);
} 


void 
move_cand_lst(cand_lst *src, cand_lst *dst) 
{
  assert(src);
  assert(dst);
  assert(src != dst);

  free_cand_lst(dst);

  dst->cnt = src->cnt;
  dst->c = src->c;                                          /* copy pointer */
}


void 
move_cand_def(cand_def *src, cand_def *dst) 
{   
  assert(src);
  assert(dst);
  assert(src != dst);

  copy_cand_def(src, dst);

  free_cand_def(src);
}


void 
move_key_def(key_def *src, key_def *dst) 
{   
  assert(src);
  assert(dst);
  assert(src != dst);

  copy_key_def(src, dst);

  free_key_def(src);
}


void
insert_int_lst(int_lst *il, int new_idx)
{
  assert(il);
  assert(new_idx>=0);
  assert(new_idx<=il->cnt);

  if (!il->nr) {                                      /* consistency check */
    assert(!il->cnt);
    il->cnt++;
    CALLOC(int, il->cnt, il->nr);
  }                                                    /* nothing to shift */
  else {
    int i;

    assert(il->cnt>0);
    il->cnt++;
    REALLOC(int, il->cnt, il->nr);

    switch (il->cnt) {
    case 0: 
      assert(0);
      break;

    case 1: 
      assert(0);
      break;

    default: 
      for (i = il->cnt-2; i >= new_idx; i--) {
	il->nr[i+1] = il->nr[i];
      }
      break;
    }
  }

  il->nr[new_idx] = 0;
}


void
insert_int_lstm(int_lstm *im, int new_idx)
{
  assert(im);
  assert(new_idx>=0);
  assert(new_idx<=im->cnt);

  if (!im->il) {                                      /* consistency check */
    assert(!im->cnt);
    im->cnt++;
    CALLOC(int_lst, im->cnt, im->il);
  }                                                    /* nothing to shift */
  else {
    int i;

    assert(im->cnt>0);
    im->cnt++;
    REALLOC(int_lst, im->cnt, im->il);

    switch (im->cnt) {
    case 0: 
      assert(0);
      break;

    case 1: 
      assert(0);
      break;

    default: 
      for (i = im->cnt-2; i >= new_idx; i--) {
	copy_int_lst(&im->il[i], &im->il[i+1]);
      }
      break;
    }
  }

  init_int_lst(&im->il[new_idx]);
}


int
misc_rb_chn_def(rb_chn_lst *rch, int src_idx, int dst_idx)
{
  assert(rch);
  assert(src_idx>=0);
  assert(src_idx<rch->cnt);
  assert(dst_idx>=0);
  assert(dst_idx<rch->cnt);
  assert(dst_idx!=src_idx);

  cat_rb_lst(rch->rcd[src_idx].rb_chain, &rch->rcd[dst_idx].rb_chain);

  delete_rb_chn_def(rch, src_idx);                   /* delete copied data */

  if (dst_idx > src_idx) {
    dst_idx--;                                    /* correct index as well */
  }

  return dst_idx;
}


int
cat_nb_def(nb_def src, nb_lst *dst_lst, gboolean no_chain) 
{
  int new_nnr_idx;

  assert(dst_lst);

  if (dst_lst->nb) {
    assert(dst_lst->cnt);
    REALLOC(nb_def, 1+dst_lst->cnt, dst_lst->nb);
  }
  else {
    assert(!dst_lst->cnt);                         /* at least initialized */
    CALLOC(nb_def, 1, dst_lst->nb);
  }

  init_nb_def(&dst_lst->nb[dst_lst->cnt]);
  copy_nb_def(&src, &dst_lst->nb[dst_lst->cnt], no_chain);

  new_nnr_idx = dst_lst->cnt++;

  return new_nnr_idx;
}


void
cat_nb_aux(nb_aux src, aux_lst *dst_lst) 
{
  assert(dst_lst);

  if (dst_lst->dat) {
    assert(dst_lst->cnt);
    REALLOC(nb_aux, 1+dst_lst->cnt, dst_lst->dat);
  }
  else {
    assert(!dst_lst->cnt);                           /* at least initialized */
    CALLOC(nb_aux, 1, dst_lst->dat);
  }

  init_nb_aux(&dst_lst->dat[dst_lst->cnt]);
  copy_nb_aux(&src, &dst_lst->dat[dst_lst->cnt]);

  dst_lst->cnt++;
}


void 
cat_rb_def(rb_def src, rb_lst *dst_lst) 
{
  assert(dst_lst);

  if (dst_lst->rb) {
    assert(dst_lst->cnt);
    REALLOC(rb_def, 1+dst_lst->cnt, dst_lst->rb);
  }
  else {
    assert(!dst_lst->cnt);                         /* at least initialized */
    CALLOC(rb_def, 1, dst_lst->rb);
  }

  init_rb_def(&dst_lst->rb[dst_lst->cnt]);
  copy_rb_def(&src, &dst_lst->rb[dst_lst->cnt]);

  dst_lst->cnt++;
}


void 
cat_rb_lst(rb_lst src, rb_lst *dst) 
{
  int i;

  assert(dst);

  dst->rar = src.rar;

  for (i = 0; i < src.cnt; i++) {
    cat_rb_def(src.rb[i], dst);
  }
}


void 
cat_cand_def(cand_def src, cand_lst *dst_lst)
{
  assert(dst_lst);

  if (dst_lst->c) {
    assert(dst_lst->cnt>0);
    REALLOC(cand_def, 1+dst_lst->cnt, dst_lst->c);
  }
  else {
    assert(!dst_lst->cnt);
    CALLOC(cand_def, 1, dst_lst->c);
  }

  init_cand_def(&dst_lst->c[dst_lst->cnt]);
  copy_cand_def(&src, &dst_lst->c[dst_lst->cnt]);

  dst_lst->cnt++;
}


void 
cat_cand_lst(cand_lst *src, cand_lst *dst) 
{
  int i;

  assert(src);
  assert(dst);
  assert(src != dst);

  for (i = 0; i < src->cnt; i++) {
    cat_cand_def(src->c[i], dst);
  }
}


void
delete_rb_chn_def(rb_chn_lst *rch, int idx)
{
  int i;

  assert(rch);
  assert(idx >= 0);
  assert(idx < rch->cnt);

  for (i = idx; i < rch->cnt-1; i++) {
    copy_rb_chn_def(&rch->rcd[i+1], &rch->rcd[i]);
  }

  free_rb_chn_def(&rch->rcd[rch->cnt-1]);

  rch->cnt--;
}


void
delete_cand_def(cand_lst *cl, int idx)
{
  int last;

  assert(cl);
  assert(idx>=0);
  assert(idx<cl->cnt);

  if (!cl->c) {                                    /* consistency check */
    assert(!cl->cnt);
    return;
  }
  else {
    assert(cl->cnt>0);
  }

  last = cl->cnt-1;

  if (idx == last) {
    free_cand_def(&cl->c[idx]);

    if (!last) {
      FREEMEM(cl->c);
    }
  }
  else {
    move_cand_def(&cl->c[last], &cl->c[idx]);
  }

  cl->cnt = last;
}


void 
free_prog_ctrl(prog_ctrl *ctrl) 
{
  assert(ctrl);

  free_set_def(&ctrl->set);
  free_cfg_def(&ctrl->cfg);

  init_prog_ctrl(ctrl); 
}


void 
free_match_def(match_def *md) 
{
  assert(md);

  init_match_def(md); 
}


void 
free_cfg_def(cfg_def *cd) 
{
  assert(cd);

  FREEMEM(cd->subst_rb);
  FREEMEM(cd->subst_nb);
  FREEMEM(cd->tps_id);

  free_batch_def(&cd->batch);
  free_int2_lst(&cd->dc_1d);
  free_int2_lst(&cd->dc_2d);
  free_match_def(&cd->mad);

  init_cfg_def(cd); 
}


void
free_set_def(set_def *sd) 
{
  assert(sd);

  sd->report_stream = sd->report_stream;         /* don't free system stuff */

  FREEMEM(sd->xprog);
  FREEMEM(sd->gtk_rc);
  FREEMEM(sd->config_filespec);
  FREEMEM(sd->image_filename);
  FREEMEM(sd->user_options);

  free_rb_lst(&sd->rblst);
  free_hst_def(&sd->grey_hist);
  free_hst_def(&sd->th_hist);

  init_set_def(sd); 
}


void 
free_key_def(key_def *kd)
{
  assert(kd);

  FREEMEM(kd->cmt);

  init_key_def(kd);
}


void 
free_key_lst(key_lst *kl)
{
  int i;

  assert(kl);

  if (!kl->kar) {                                     /* consistency check */
    assert(!kl->cnt);
  }
  else {
    assert(kl->cnt>0);
  }

  for (i = 0; i < kl->cnt; i++) {
    free_key_def(&kl->kar[i]);
  }
  FREEMEM(kl->kar);

  init_key_lst(kl);
}


void
free_baseline_lst(baseline_lst *bl)
{
  assert(bl);

  free_int_lst(&bl->bline);

  init_int_lst(&bl->bline);
}


void
free_h_contour_lst(h_contour_lst *cl)
{
  assert(cl);

  if (!cl->hcl) {                                    /* consistency check */
    assert(!cl->cnt);
  }
  else {
    assert(cl->cnt>0);
  }

  FREEMEM(cl->hcl);

  init_h_contour_lst(cl);
}


void
free_v_contour_lst(v_contour_lst *vl)
{
  assert(vl);

  if (!vl->vcl) {                                    /* consistency check */
    assert(!vl->cnt);
  }
  else {
    assert(vl->cnt>0);
  }

  FREEMEM(vl->vcl);

  init_v_contour_lst(vl);
}


void
free_sort_ilst(sort_ilst *sl) 
{
  assert(sl);

  if (!sl->il) {                                     /* consistency check */
    assert(!sl->cnt);
  }
  else {
    assert(sl->cnt>0);
  }

  FREEMEM(sl->il);

  init_sort_ilst(sl);
}


void
free_sort_dlst(sort_dlst *sl) 
{
  assert(sl);

  if (!sl->dl) {                                     /* consistency check */
    assert(!sl->cnt);
  }
  else {
    assert(sl->cnt>0);
  }

  FREEMEM(sl->dl);

  init_sort_dlst(sl);
}


void
free_cls_lst(cls_lst *cl) 
{
  int c;

  assert(cl);

  if (!cl->cls) {                                     /* consistency check */
    assert(!cl->cnt);
  }
  else {
    /* consistency checked elsewhere */
  }

  for (c = 0; c < cl->cnt; c++) {
    free_cls_def(&cl->cls[c]); 
  }
  FREEMEM(cl->cls);

  init_cls_lst(cl);
}


void
free_tps_def(tps_def *tps) 
{
  assert(tps);

  FREEMEM(tps->ident);

  free_cls_lst(&tps->cl);

  init_tps_def(tps);
}


void
free_hst_def(hst_def *hd) 
{
  assert(hd);

  free_double_lst(&hd->hst);

  init_hst_def(hd);
}


void
free_ival_lst(ival_lst *il) 
{
  assert(il);

  if (!il->ival) {                                   /* consistency check */
    assert(!il->cnt);
  }
  else {
    assert(il->cnt>0);

    FREEMEM(il->ival);
  }

  init_ival_lst(il);
}


void
free_sector_lst(sector_lst *sl) 
{
  assert(sl);

  FREEMEM(sl->sdl);

  init_sector_lst(sl);
}


void
free_fft_def(fft_def *fd) 
{
  assert(fd);

  FREEMEM(fd->spec);

  if (fd->ar.height) {
    FREEMEM(fd->data[fd->ar.height]);       /* memory in fact if allocated */
  }
  FREEMEM(fd->data);

  init_fft_def(fd);
}


void
free_cv_def_i_matrix(cv_def_i ***matrix, int vector_len) 
{
  assert(matrix);
  assert(vector_len>0);

  FREEMEM((*matrix)[vector_len]);                         /* memory in fact */
  FREEMEM(*matrix);
}


void
free_lag_lst_2D(lag_lst ***lag, area_def ar)
{
  int x, y;

  assert(lag);

  for (y = 0; y < ar.height && *lag; y++) {
    for (x = 0; x < ar.width; x++) {
      FREEMEM((*lag)[y][x].ld);
    }
  }

  if (*lag) {
    FREEMEM((*lag)[ar.height]);
    FREEMEM(*lag);
  }
}


void
free_kwh_def(kwh_def *kw) 
{
  assert(kw);

  free_lag_lst_2D(&kw->ll, kw->ar);

  init_kwh_def(kw);
}


void 
free_cls_att(cls_att *ca) 
{
  assert(ca);

  FREEMEM(ca->class);
  FREEMEM(ca->sign);
  FREEMEM(ca->crem);

  init_cls_att(ca);
}


void
free_cls_def(cls_def *cls) 
{
  assert(cls);

  free_cls_att(&cls->att);
  free_nb_lst(&cls->nbl);
  free_cc_lst(&cls->ccl);
  free_inh_lst(&cls->inl);

  init_cls_def(cls);
}


void
free_nb_def(nb_def *nb) 
{
  assert(nb);

  free_pic_def(&nb->pat);
  free_nb_chn_lst(&nb->cnb);
  free_nb_chn_lst(&nb->nnb);
  free_nb_ref(&nb->ref);

  init_nb_def(nb);
}


void
free_nb_aux(nb_aux *na) 
{
  assert(na);

  init_nb_aux(na);
}


void
free_hst_info(hst_info *hi)
{
  assert(hi);

  free_hst_def(&hi->hd);

  init_hst_info(hi);
}


void
free_cc_def(cc_def *cc) 
{
  assert(cc);

  FREEMEM(cc->patt);
  init_cc_def(cc);
}


void
free_lag_lst(lag_lst *ll) 
{
  assert(ll);

  FREEMEM(ll->ld);
  init_lag_lst(ll);
}


void
free_part_lst(part_lst *pl) 
{
  int i;

  assert(pl);

  if (pl->cnt) {
    assert(pl->fp);                             /* must be consistent */
  }

  for (i = 0; i < pl->cnt; i++) {
    free_fft_def(&pl->fp[i]);
  }
  FREEMEM(pl->fp);

  init_part_lst(pl);
}


void
free_pic_dbl(pic_dbl *pd) 
{
  assert(pd);

  FREEMEM(pd->mem);
  init_pic_dbl(pd);
}


void
free_pic_def(pic_def *pd) 
{
  assert(pd);

  FREEMEM(pd->mem);
  FREEMEM(pd->bvl);
  FREEMEM(pd->col);
  init_pic_def(pd);
}


void 
free_img_ext(img_ext *ie) 
{
  assert(ie);

  free_aux_lst(&ie->il);
  free_pic_def(&ie->eimg);

  init_img_ext(ie);
}


void 
free_img_typ(img_typ *it) 
{
  assert(it);

  free_pic_def(&it->pimg);

  init_img_typ(it);
}


void
free_int_lst(int_lst *il) 
{
  assert(il);

  FREEMEM(il->nr);
  init_int_lst(il);
}


void
free_int2_lst(int2_lst *i2l) 
{
  assert(i2l);

  FREEMEM(i2l->n);
  init_int2_lst(i2l);
}


void
free_int_lstm(int_lstm *im) 
{
  int i;

  assert(im);

  for (i = 0; i < im->cnt; i++) {
    free_int_lst(&im->il[i]);
  }
  FREEMEM(im->il);

  init_int_lstm(im);
}


void
free_text_ring(text_ring *tr) 
{
  free_text_def(&tr->td);
  init_text_ring(tr);
}


void
free_text_def(text_def *td) 
{
  int i;

  assert(td);

  if (!td->tl) {                                     /* consistency check */
    assert(!td->cnt);
  }
  else {
    assert(td->cnt>0);
  }

  for (i = 0; i < td->cnt; i++) {
    FREEMEM(td->tl[i]);
  }
  FREEMEM(td->tl);

  init_text_def(td);
}


void
free_double_lst(double_lst *dl) 
{
  assert(dl);

  FREEMEM(dl->dbl);
  init_double_lst(dl);
}


void
free_rb_def(rb_def *rd) 
{
  assert(rd);

  free_ival_lst(&rd->gridy);

  init_rb_def(rd);
}


void
free_rb_chn_def(rb_chn_def *rc) 
{
  assert(rc);

  FREEMEM(rc->chain_name);
  free_rb_lst(&rc->rb_chain);
  free_cand_lst(&rc->bhclst);

  init_rb_chn_def(rc);
}


void
free_inh_def(inh_def *id) 
{
  assert(id);

  init_inh_def(id);
}


void
free_tiles_def(tiles_def *rl) 
{
  assert(rl);

  free_rb_chn_def(&rl->doc);
  free_rb_chn_lst(&rl->rch);

  init_tiles_def(rl);
}


void
free_rb_chn_lst(rb_chn_lst *rcl) 
{
  int c;

  assert(rcl);

  if (rcl->cnt) {
    assert(rcl->rcd);                                /* must be consistent */
  }

  for (c = 0; c < rcl->cnt; c++) {
    free_rb_chn_def(&rcl->rcd[c]);
  }
  FREEMEM(rcl->rcd);

  init_rb_chn_lst(rcl);
}


void
free_batch_def(batch_def *bd) 
{
  assert(bd);

  init_batch_def(bd);
}


void
free_rb_lst(rb_lst *rl) 
{
  int r;

  assert(rl);

  if (!rl->rb) {                                      /* consistency check */
    assert(!rl->cnt);
  }
  else {
    assert(rl->cnt>0);
  }

  for (r = 0; r < rl->cnt; r++) {
    free_rb_def(&rl->rb[r]);
  }
  FREEMEM(rl->rb);

  init_rb_lst(rl);
}


void
free_tps_lst(tps_lst *tpl) 
{
  int c;

  assert(tpl);

  if (!tpl->tps) {
    assert(!tpl->cnt);                               /* must be consistent */
  }
  else {
    /* consistency checked elsewhere */
  }

  for (c = 0; c < tpl->cnt; c++) {
    free_tps_def(&tpl->tps[c]);
  }
  FREEMEM(tpl->tps);

  init_tps_lst(tpl);
}


void
free_nb_lst(nb_lst *nl) 
{
  int n;

  assert(nl);

  if (nl->cnt) {
    assert(nl->nb);                             /* must be consistent */
  }
  else {
    /* consistency checked elsewhere */
  }

  for (n = 0; n < nl->cnt; n++) {
    free_nb_def(&nl->nb[n]); 
  }
  FREEMEM(nl->nb);

  init_nb_lst(nl);
}


void
free_inh_lst(inh_lst *il) 
{
  int n;

  assert(il);

  if (il->ind) {                                   /* consistency */
    assert(il->cnt>0);
  }
  else {
    assert(!il->cnt);
  }

  for (n = 0; n < il->cnt; n++) {
    free_inh_def(&il->ind[n]); 
  }
  FREEMEM(il->ind);

  init_inh_lst(il);
}


void
free_aux_lst(aux_lst *al) 
{
  int n;

  assert(al);

  if (al->dat) {                                   /* consistency */
    assert(al->cnt>0);
  }
  else {
    assert(!al->cnt);
  }

  for (n = 0; n < al->cnt; n++) {
    free_nb_aux(&al->dat[n]); 
  }
  FREEMEM(al->dat);

  init_aux_lst(al);
}


void
free_cc_lst(cc_lst *cl) 
{
  int n;

  assert(cl);

  if (cl->cnt) {
    assert(cl->cc);                             /* must be consistent */
  }

  for (n = 0; n < cl->cnt; n++) {
    free_cc_def(&cl->cc[n]); 
  }
  FREEMEM(cl->cc);

  init_cc_lst(cl);
}


void
free_nb_chn_lst(nb_chn_lst *ncl) 
{
  assert(ncl);

  FREEMEM(ncl->cd);
  init_nb_chn_lst(ncl);
}


void
free_db_def(db_def *db) 
{
  assert(db);

  FREEMEM(db->db_fname);
  FREEMEM(db->db_version);

  free_tps_lst(&db->tpl);

  init_db_def(db);
}


void
free_nb_ref(nb_ref *nr)
{
  assert(nr);

  free_pic_def(&nr->pic);
  free_pic_def(&nr->fld);
  free_pic_def(&nr->wtd);
  free_pic_def(&nr->msk);
  free_pic_def(&nr->dlt);

  init_nb_ref(nr);
}


void 
free_cand_def(cand_def *kr)
{
  assert(kr);

  free_int_lst(&kr->crb);

  if (kr->alt) {
    free_cand_lst(kr->alt);
    FREEMEM(kr->alt);
  }

  init_cand_def(kr);
}


void
free_cand_lst(cand_lst *kl) 
{
  int i;

  assert(kl);

  if (kl->cnt) {
    assert(kl->c);                             /* must be consistent */
  }

  for (i = 0; i < kl->cnt; i++) {
    free_cand_def(&kl->c[i]); 
  }
  FREEMEM(kl->c);

  init_cand_lst(kl);
}


void
free_rc_def(rc_def *rd) 
{
  assert(rd);

  free_cand_lst(&rd->r);

  init_rc_def(rd);
}


void 
free_rc_lst(rc_lst *rlist) 
{
  int i;

  assert(rlist);

  if (rlist->s) {
    assert(rlist->cnt>0);                            /* must be consistent */
  }
  else {
    assert(!rlist->cnt);
  }

  for (i = 0; i < rlist->cnt; i++) {                   
    free_rc_def(&rlist->s[i]);
  }
  FREEMEM(rlist->s);

  init_rc_lst(rlist);
}


void 
free_kwh_lst(kwh_lst *kl) 
{
  int i;

  assert(kl);

  if (kl->cnt) {
    assert(kl->kwh);                             /* must be consistent */
  }

  for (i = 0; i < kl->cnt; i++) {                   
    free_kwh_def(&kl->kwh[i]);
  }
  FREEMEM(kl->kwh);

  init_kwh_lst(kl);
}


void
free_rb_ref(rb_ref *rr) 
{
  assert(rr);

  /* don't free RB image */
  free_pic_def(&rr->wpic);
  free_pic_def(&rr->shdw1);
  free_pic_def(&rr->shdw2);
  free_kwh_lst(&rr->kl);
  free_reg_def(&rr->xreg);

  FREEMEM(rr->pic_state);
}


void 
free_reg_def(reg_def *rd) 
{
  int c;

  assert(rd);

  if (!rd->cnt) {                                    /* consistency check */
    assert(!rd->yrl);
  }
  else {
    assert(rd->cnt>0);
    assert(rd->yrl);
  }

  for (c = 0; c < rd->cnt; c++) {
    free_y_reg_lst(&rd->yrl[c]); 
  }
  FREEMEM(rd->yrl);

  init_reg_def(rd);
}


void 
free_y_reg_def(y_reg_def *yrd) 
{
  assert(yrd);

  FREEMEM(yrd->pic_state);

  free_cand_lst(&yrd->vlst);

  init_y_reg_def(yrd); 
}


void 
free_y_reg_lst(y_reg_lst *yrl) 
{
  int c;

  assert(yrl);

  if (!yrl->cnt) {                                    /* consistency check */
    assert(!yrl->yrd);
  }
  else {
    assert(yrl->cnt>0);
    assert(yrl->yrd);
  }

  for (c = 0; c < yrl->cnt; c++) {
    free_y_reg_def(&yrl->yrd[c]); 
  }
  FREEMEM(yrl->yrd);

  init_y_reg_lst(yrl);
}


void
free_cand_lst_ary(cand_lst_ary *cla) 
{
  int c;

  assert(cla);

  if (!cla->cnt) {                                    /* consistency check */
    assert(!cla->clst);
  }
  else {
    assert(cla->cnt>0);
    assert(cla->clst);
  }

  for (c = 0; c < cla->cnt; c++) {
    free_cand_lst(&cla->clst[c]); 
  }
  FREEMEM(cla->clst);

  init_cand_lst_ary(cla);
}


void
free_hash_def(hash_def *hdef) 
{
  int i;

  assert(hdef);

  FREEMEM(hdef->cid);
  FREEMEM(hdef->pid);

  if (hdef->cnt) {
    assert(hdef->kr);                             /* must be consistent */
  }

  for (i = 0; i < hdef->cnt; i++) {
    free_cand_def(&hdef->kr[i]); 
  }
  FREEMEM(hdef->kr);

  init_hash_def(hdef);
}


void
free_hash_lst(hash_lst *cdef) 
{
  int i;

  assert(cdef);

  if (cdef->cnt) {
    assert(cdef->nid);                             /* must be consistent */
  }

  for (i = 0; i < cdef->cnt; i++) {
    free_hash_def(&cdef->nid[i]); 
  }
  FREEMEM(cdef->nid);

  init_hash_lst(cdef);
}


void 
free_gen_lst(gen_lst *gl)
{
  int i;

  assert(gl);

  if (gl->cnt) {
    assert(gl->gd);                             /* must be consistent */
  }

  for (i = 0; i < gl->cnt; i++) {
    free_gen_def(&gl->gd[i]); 
  }
  FREEMEM(gl->gd);

  init_gen_lst(gl);
}


void 
free_gen_def(gen_def *gd)
{
  int i;

  assert(gd);

  FREEMEM(gd->c);
  FREEMEM(gd->n);

  if (gd->cnt) {
    assert(gd->hl);                             /* must be consistent */
  }

  for (i = 0; i < gd->cnt; i++) {
    free_cand_def(&gd->hl[i]); 
  }
  FREEMEM(gd->hl);

  init_gen_def(gd);
}


void
free_fmatrix(int ny, float **m)
{
  int i;

  assert(m);

  for (i = 0; i < ny+1; i++) {
    FREEMEM(m[i]);                                      /* deallocate rows */
  }

  FREEMEM(m);                                /* deallocate pointer to rows */
}


int
set_int_lst(int_lst il, int to, int nrignore) 
{
  int cnt = il.cnt, set = 0;

  if (!il.nr) {                                       /* consistency check */
    assert(!cnt);
    return set;
  }
  else {
    assert(cnt>0);
  }

  while (cnt--) {
    if (il.nr[cnt] != nrignore) {
      il.nr[cnt] = to;
      set++;
    }
  }

  return set;
}


int
max_int_lst(int_lst il) 
{
  int indx;

  indx = max_index_int_lst(il);

  return il.nr[indx];
}


int
max_index_int_lst(int_lst il) 
{
  int i, max, indx = -1;

  assert(il.cnt > 0);

  max = il.nr[0];
  indx = 0;
  for (i = 1; i < il.cnt; i++) {
    if (max < il.nr[i]) {
      max = il.nr[i];
      indx = i;
    }
  }

  return indx;                                    /* first maximum in list */
}


double
min_double_left(double_lst dl) 
{
  int indx;

  indx = min_index_double_left(dl.dbl, dl.cnt);

  return dl.dbl[indx];
}


int
min_index_double_left(double *dbl, int cnt) 
{
  int   i, indx;
  float min;

  assert(dbl);
  assert(cnt > 0);

  min = dbl[0];
  indx = 0;
  for (i = 0; i < cnt; i++) {
    if (min > (float)dbl[i]) {         /* avoiding rounding error problems */
      min = dbl[i];
      indx = i;
    }
  }

  return indx;                                    /* first maximum in list */
}


double
max_double_left(double_lst dl) 
{
  int indx;

  indx = max_index_double_left(dl.dbl, dl.cnt);

  return dl.dbl[indx];
}


int
max_index_double_left(double *dbl, int cnt) 
{
  int   i, indx;
  float max;

  assert(dbl);
  assert(cnt > 0);

  max = dbl[0];
  indx = 0;
  for (i = 0; i < cnt; i++) {
    if (max < (float)dbl[i]) {         /* avoiding rounding error problems */
      max = dbl[i];
      indx = i;
    }
  }

  return indx;                                    /* first maximum in list */
}


double
max_double_right(double_lst dl) 
{
  int indx;

  indx = max_index_double_right(dl.dbl, dl.cnt);

  return dl.dbl[indx];
}


int
max_index_double_right(double *dbl, int cnt) 
{
  int   i, indx;
  float max;

  assert(dbl);
  assert(cnt > 0);

  max = dbl[0];
  indx = 0;
  for (i = 0; i < cnt; i++) {
    if (max <= (float)dbl[i]) {        /* avoiding rounding error problems */
      max = dbl[i];
      indx = i;
    }
  }

  return indx;                                     /* last maximum in list */
}


int_lst
max_index_double_all(double *dbl, int cnt) 
{
  int     i, indx;
  double  max;
  int_lst indx_lst;

  assert(dbl);
  assert(cnt > 0);

  init_int_lst(&indx_lst);

  indx = max_index_double_left(dbl, cnt);
  max = dbl[indx];

  indx_lst.cnt = cnt;
  CALLOC(int, indx_lst.cnt, indx_lst.nr);

  for (i = indx = 0; i < cnt; i++) {
    if (fabs(max-dbl[i]) <= FLT_EPSILON) {
      indx_lst.nr[indx++] = i;
    }
  }

  if (indx != indx_lst.cnt) {
    indx_lst.cnt = indx;
    REALLOC(int, indx_lst.cnt, indx_lst.nr);
  }

  return indx_lst;                                   /* all maxima in list */
}


void
mult_double(double *dbl, int cnt, double mult) 
{
  int i;

  assert(dbl);
  assert(cnt > 0);

  for (i = 0; i < cnt; i++) {
    dbl[i] *= mult;
  }
}


void 
add_two_pic_dbl(pic_dbl dbl1, pic_dbl dbl2)
{
  int    len;
  double *p1 = dbl1.mem, *p2 = dbl2.mem;

  assert(dbl1.ar.width==dbl2.ar.width);
  assert(dbl1.ar.height==dbl2.ar.height);

  len = AREA_SIZE(dbl1.ar);

  while (len-- > 0) {
    *p2++ += *p1++;
  }
}


void 
store_pic_dbl(pic_dbl src, pic_dbl dst)
{
  int    len;
  double *p1 = src.mem, *p2 = dst.mem;

  if (!src.mem) {
    return;
  }

  assert(dst.mem);
  assert(src.ar.width==dst.ar.width);
  assert(src.ar.height==dst.ar.height);

  len = AREA_SIZE(src.ar);

  assert(len>0);

  while (len-- > 0) {
    *p2++ = *p1++;
  }
}


void 
add_two_hst_def(hst_def hd1, hst_def *hd2)
{
  int    len;
  double *p1 = hd1.hst.dbl, *p2 = hd2->hst.dbl;

  assert(hd1.hst.cnt==hd2->hst.cnt);

  len = hd1.hst.cnt;

  while (len-- > 0) {
    *p2++ += *p1++;
  }

  hd2->max_bin = max_index_double_left(hd2->hst.dbl, hd2->hst.cnt);
  hd2->maxval = hd2->hst.dbl[hd2->max_bin];            /* adjust histogram */
}



/* Wrapper for add_string.

Revision: 04.06.2003 E.Sack

Input : - character
        - pointer to destination string
Output: - destination string updated as flagged
        - flag (0: nothing changed, 1: character added)
*/
int
add_character(char src, char **dst)
{
  int  added;
  char char_buf[2];

  char_buf[0] = src;
  char_buf[1] = '\0';

  added = add_string(char_buf, dst);

  return added;
}



/* Adds two strings with arbitrary length.

Revision: 04.06.2003 E.Sack

Input : - pointer to source string
        - pointer to destination string
Output: - destination string updated as flagged
        - flag (0: nothing changed, 1: string added)
*/
int
add_string(const char *src, char **dst)
{
  int len;

  if (!src) {
    return 0;                                            /* nothing to add */
  }

  assert(dst);

  len = strlen(src) + 1;                            /* charged trailing \0 */

  if (len == 1) {
    return 0;                                       /* empty source string */
  }

  assert(len <= 1024*1024);                     /* there should be a limit */

  if (*dst) {                         
    len += strlen(*dst);                /* overwrites empty rest of string */
    assert(len <= 1024*1024);                   /* there should be a limit */
    REALLOC(char, len, *dst); 
  }
  else {                              
    CALLOC(char, len, *dst);  
  }

  strcat(*dst, src);                       

  return 1;
}



/* Sorts a list in descending order.

Revision: 20.12.2002 E.Sack

Input : - pointer to list
Output: - list updated
*/
void
descending_sort_ilst(sort_ilst sil) 
{
  int i, j, k, idx, max_val;

  if (sil.il) {                                             /* consistency */
    assert(sil.cnt>0);
  }
  else {
    assert(!sil.cnt);
  }

  if (sil.cnt <= 1) {
    return;                                           /* no sorting needed */
  }

  for (i = 0; i < sil.cnt-1; i++) {
    max_val = sil.il[i].val;
    idx = sil.il[i].idx;
    k = i;

    for (j = i+1; j < sil.cnt; j++) {/* look for a greater val in the list */
      if (max_val < sil.il[j].val) {      /* current value less than next? */
	max_val = sil.il[j].val;                           /* yes - get it */
	idx = sil.il[j].idx;    
	k = j;
      }                                                    /* no - skip it */
    }

    if (k == i) {
      continue;                               /* no greater value detected */
    }

    sil.il[k].val = sil.il[i].val;               /* exchange greater value */
    sil.il[k].idx = sil.il[i].idx;
    sil.il[i].val = max_val;
    sil.il[i].idx = idx;
  }
}



/* Sorts a list in ascending order.

Revision: 20.12.2002 E.Sack

Input : - pointer to list
Output: - list updated
*/
void
ascending_sort_ilst(sort_ilst sil) 
{
  int i, j, k, idx, min_val;

  if (sil.il) {                                             /* consistency */
    assert(sil.cnt>0);
  }
  else {
    assert(!sil.cnt);
  }

  if (sil.cnt <= 1) {
    return;                                           /* no sorting needed */
  }

  for (i = 0; i < sil.cnt-1; i++) {
    min_val = sil.il[i].val;
    idx = sil.il[i].idx;
    k = i;

    for (j = i+1; j < sil.cnt; j++) { /* look for a lesser val in the list */
      if (min_val > sil.il[j].val) {   /* current value greater than next? */
	min_val = sil.il[j].val;                           /* yes - get it */
	idx = sil.il[j].idx;    
	k = j;
      }                                                    /* no - skip it */
    }

    if (k == i) {
      continue;                                /* no lesser value detected */
    }

    sil.il[k].val = sil.il[i].val;                /* exchange lesser value */
    sil.il[k].idx = sil.il[i].idx;
    sil.il[i].val = min_val;
    sil.il[i].idx = idx;
  }
}



/* Sorts a list in ascending order.

Revision: 20.12.2002 E.Sack

Input : - pointer to list
Output: - list updated
*/
void
ascending_sort_dlst(sort_dlst sdl) 
{
  int      i, j, k;
  sort_dbl min_val;

  if (sdl.dl) {                                             /* consistency */
    assert(sdl.cnt>0);
  }
  else {
    assert(!sdl.cnt);
  }

  if (sdl.cnt <= 1) {
    return;                                           /* no sorting needed */
  }

  for (i = 0; i < sdl.cnt-1; i++) {
    min_val = sdl.dl[i];
    k = i;

    for (j = i+1; j < sdl.cnt; j++) { /* look for a lesser val in the list */
      if (min_val.val > sdl.dl[j].val) {/* current value greater than next? */
	min_val = sdl.dl[j];                                /* yes - get it */
	k = j;
      }                                                    /* no - skip it */
    }

    if (k == i) {
      continue;                                /* no lesser value detected */
    }

    sdl.dl[k] = sdl.dl[i];                        /* exchange lesser value */
    sdl.dl[i] = min_val;
  }
}



/* Concatenates strings.

Revision: 19.05.2004 E.Sack

Input : - pointer to prefix string
        - pointer to root string
        - pointer to suffix string
        - pointer to pointer to string
Output: - string updated
Postcondition: - string to free
*/
void
cat_string(const char *prefix, const char *root, const char *suffix, 
	   char **string_out)
{
  assert(string_out);

  add_string(prefix, string_out);
  add_string(root, string_out);
  add_string(suffix, string_out);
}



/* Converts an integer to a string.

Revision: 29.07.2005 E.Sack

Input : - pointer to prefix string
        - integer
        - pointer to suffix string
        - pointer to pointer to string
Output: - string updated
Postcondition: - string to free
*/
void
int_to_string(char *prefix, int in, char *suffix, gchar **string_out)
{
  char num[256];

  assert(string_out);

  g_snprintf(num, sizeof(num), "%d", in);
  add_string(prefix, string_out);
  add_string(num, string_out);
  add_string(suffix, string_out);
}



/* Converts an double to a string.

Revision: 29.07.2005 E.Sack

Input : - pointer to prefix string
        - double
        - pointer to suffix string
        - pointer to pointer to string
Output: - string updated
Postcondition: - string to free
*/
void
double_to_string(char *prefix, double in, char *suffix, char **string_out)
{
  char num[256];

  assert(string_out);

  g_snprintf(num, sizeof(num), "%f", in);

  add_string(prefix, string_out);
  add_string(num, string_out);
  add_string(suffix, string_out);
}



/* Converts a string to limited L phase value.

Revision: 26.12.2003 E.Sack

Input : - pointer to string
        - pointer to integer
Output: - integer updated
        - flag (0: integer corrected/not in string, 1: no correction)
*/
int
string_to_l_phase(const char *in, int *out)
{
  int d = -1, flag = 0;

  assert(out);

  if (in) {
    d = sscanf(in, "%d", out);
  }

  if (d <= 0) {
    *out = XPLAB_L_PHASE_DEF;
  }
  else if (*out < XPLAB_L_PHASE_MIN) {
    *out = XPLAB_L_PHASE_MIN;
  }
  else if (*out > XPLAB_L_PHASE_MAX) {
    *out = XPLAB_L_PHASE_MAX;
  }
  else {
    flag = 1;                                     /* successful conversion */
  }

  return flag;
}



/* Converts a string to limited H phase value.

Revision: 26.12.2003 E.Sack

Input : - pointer to string
        - pointer to integer
Output: - integer updated
        - flag (0: integer corrected/not in string, 1: no correction)
*/
int
string_to_h_phase(const char *in, int *out)
{
  int d = -1, flag = 0;

  assert(out);

  if (in) {
    d = sscanf(in, "%d", out);
  }

  if (d <= 0) {
    *out = XPLAB_H_PHASE_DEF;
  }
  else if (*out < XPLAB_H_PHASE_MIN) {
    *out = XPLAB_H_PHASE_MIN;
  }
  else if (*out > XPLAB_H_PHASE_MAX) {
    *out = XPLAB_H_PHASE_MAX;
  }
  else {
    flag = 1;                                     /* successful conversion */
  }

  return flag;
}



/* Converts a string to a limited dpi value.

Revision: 29.07.2005 E.Sack

Input : - pointer to string
        - pointer to integer
Output: - integer updated
        - flag (0: integer corrected/not in string, 1: no correction)
*/
int
string_to_dpi(const char *in, int *out)
{
  int d = -1, flag = 0, err;

  assert(out);

  err = check_name(in);

  if (!err) {
    d = sscanf(in, "%d", out);
  }

  if (d <= 0) {
    *out = DPI_DEF;
  }
  else if (*out < DPI_MIN) {
    *out = DPI_MIN;
  }
  else if (*out > DPI_MAX) {
    *out = DPI_MAX;
  }
  else {
    flag = 1;                                     /* successful conversion */
  }

  return flag;
}



/* Converts enumeration to a string.

Revision: 26.12.2003 E.Sack

Input : - enumeration
        - pointer to string
Output: - string updated as flagged
        - flag (0: no conversion, 1: conversion)
*/
int
inline_method_to_string(INLINE_METHOD enum_in, char **string_out)
{
  assert(string_out);

  switch (enum_in) {
  case OVERLAPPING:  add_string(OVERLAPPING_STRING, string_out); break;
  case LINE_SPACING: add_string(LINE_SPACING_STRING, string_out); break;
  case METHOD_FIRST: case METHOD_LAST:
    add_string(METHOD_AUTO_STRING, string_out);
    return 0;                                /* conversion doesn't succeed */
    break;
  default: assert(0); break;                               /* unknown type */
  }

  return 1;                                         /* conversion succeeds */
}



/* Converts string to a enumeration.

Revision: 23.12.2003 E.Sack

Input : - pointer to string
        - pointer to enumeration
Output: - string updated as flagged
        - flag (0: no enumeration, 1: enumeration)
*/
int
string_to_inline_method(const char *string_in, INLINE_METHOD *enum_out)
{
  assert(enum_out);

  if (!string_in) {
    return 0;
  }

  if (!strlen(string_in)) {
    return 0;
  }

  if (!strcmp(string_in, OVERLAPPING_STRING)) {
    *enum_out = OVERLAPPING;
    return 1;
  }

  if (!strcmp(string_in, LINE_SPACING_STRING)) {
    *enum_out = LINE_SPACING;
    return 1;
  }

  return 0;                                           /* conversion failes */
}



/* Converts enumeration to a string.

Revision: 02.03.2003 E.Sack

Input : - enumeration
        - pointer to string
Output: - string updated as flagged
        - flag (0: no conversion, 1: conversion)
*/
int
aligmnt_to_string(ALIGMNT enum_in, char **string_out)
{
  assert(string_out);

  switch (enum_in) {
  case UPPERLINE:  add_string("UPPERLINE", string_out); break;
  case MIDDLELINE: add_string("MIDDLELINE", string_out); break;
  case BASELINE:   add_string("BASELINE", string_out); break;
  case UNDERLINE:  add_string("UNDERLINE", string_out); break;
  case UDEFLINE: case ALIGNMENT_MAX:
    return 0;                                /* conversion doesn't succeed */
  default: assert(0); break;                               /* unknown type */
  }

  return 1;                                         /* conversion succeeds */
}



/* Converts string to a enumeration.

Revision: 21.06.2003 E.Sack

Input : - pointer to string
        - pointer to enumeration
Output: - string updated as flagged
        - flag (0: no enumeration, 1: enumeration)
*/
int
string_to_aligmnt(const char *string_in, ALIGMNT *enum_out)
{
  assert(enum_out);

  if (!string_in) {
    return 0;
  }

  if (!strlen(string_in)) {
    return 0;
  }

  if (!strcmp(string_in, "BASELINE")) {
    *enum_out = BASELINE;
    return 1;
  }

  if (!strcmp(string_in, "UPPERLINE")) {
    *enum_out = UPPERLINE;
    return 1;
  }

  if (!strcmp(string_in, "UNDERLINE")) {
    *enum_out = UNDERLINE;
    return 1;
  }

  if (!strcmp(string_in, "MIDDLELINE")) {
    *enum_out = MIDDLELINE;
    return 1;
  }

  return 0;                                           /* conversion failes */
}



/* Converts enumeration to integer.

Revision: 02.03.2003 E.Sack

Input : - enumeration
        - pointer to integer
Output: - integer updated as flagged
        - flag (0: no conversion, 1: conversion)
*/
int
aligmnt_to_int(ALIGMNT enum_in, int *int_out)
{
  switch (enum_in) {
  case UPPERLINE:  *int_out = 4; break;
  case MIDDLELINE: *int_out = 3; break;
  case BASELINE:   *int_out = 2; break;
  case UNDERLINE:  *int_out = 1; break;
  case UDEFLINE: case ALIGNMENT_MAX:
    return 0;                                 /* conversion don't succeeds */
  default: assert(0); break;                               /* unknown type */
  }

  return 1;                                         /* conversion succeeds */
}



/* Converts integer to enumeration.

Revision: 02.03.2003 E.Sack

Input : - integer
        - pointer to enumeration
Output: - enumeration updated as flagged
        - flag (0: no conversion, 1: conversion)
*/
int
int_to_aligmnt(int int_in, ALIGMNT *enum_out)
{
  switch (int_in) {
  case 4: *enum_out = UPPERLINE; break;
  case 3: *enum_out = MIDDLELINE; break;
  case 2: *enum_out = BASELINE; break;
  case 1: *enum_out = UNDERLINE; break;
  default:
    return 0;                                 /* conversion don't succeeds */
  }

  return 1;                                         /* conversion succeeds */
}



/* Puts a key to list.

Revision: 16.12.2004 E.Sack

Input : - pointer to key list
        - pointer to key
        - pointer to comment string
        - pointer to unknown
Output: - key list updated
        - key updated
        - flag (0: no error, 1: key is not zero)
*/
int
put_key(key_lst *klst, int *key, char *key_comment, int write_access, 
	void *ptr)
{
  static int skey = 0;
  int        idx;

  assert(klst);

  if (!klst->kar) {                                        /* may be empty */
    assert(!klst->cnt);                              /* must be consistent */
  }
  else {
    assert(klst->cnt>0);                             /* must be consistent */
  }

  assert(write_access==0 || write_access==1);

  if (!key || !klst) {
    return 0;                                             /* no key wanted */
  }

  if (*key) {
    return 1;                                           /* nought expected */
  }

  *key = ++skey;                                           /* generate key */

  idx = klst->cnt++;

  if (idx) {
    REALLOC(key_def, klst->cnt, klst->kar);
  }
  else {
    CALLOC(key_def, klst->cnt, klst->kar);
  }

  init_key_def(&klst->kar[idx]);

  klst->kar[idx].key = *key;
  klst->kar[idx].wac = write_access;
  klst->kar[idx].aux = ptr;
  add_string(key_comment, &klst->kar[idx].cmt);

  return 0;
}



/* Puts a sector to a sector list.

Revision: 15.01.2006 E.Sack

Input : - pointer to sector list
        - sector
Output: - sector list updated
*/
void
put_sector_lst(sector_lst *slst, sector_def sdef)
{
  int idx;

  assert(slst);

  if (!slst->sdl) {                                        /* may be empty */
    assert(!slst->cnt);                              /* must be consistent */
  }
  else {
    assert(slst->cnt>0);                             /* must be consistent */
  }

  idx = slst->cnt++;

  if (idx) {
    REALLOC(sector_def, slst->cnt, slst->sdl);
  }
  else {
    CALLOC(sector_def, slst->cnt, slst->sdl);
  }

  init_sector_def(&slst->sdl[idx]);

  copy_sector_def(&sdef, &slst->sdl[idx]);
}



/* Copies a string to a text list.

Revision: 17.01.2006 E.Sack

Input : - pointer to text list
        - text
Output: - list updated
*/
void
copy_to_text_def(text_def *tdef, const gchar *txt)
{
  int idx;

  assert(tdef);

  if (!tdef->tl) {                                         /* may be empty */
    assert(!tdef->cnt);                              /* must be consistent */
  }
  else {
    assert(tdef->cnt>0);                             /* must be consistent */
  }

  idx = tdef->cnt++;

  if (idx) {
    REALLOC(gchar *, tdef->cnt, tdef->tl);
  }
  else {
    CALLOC(gchar *, tdef->cnt, tdef->tl);
  }

  tdef->tl[idx] = NULL;                               /* init zero pointer */

  add_string(txt, &tdef->tl[idx]);
}



/* Requests a key list write access index.

Revision: 22.11.2003 E.Sack

Input : - key list
Output: - index (-1 no write access, >=0: read and write access)
*/
int
request_key_lst(key_lst klst)
{
  int kaccess, idx;

  for (idx = 0; idx < klst.cnt; idx++) {
    kaccess = klst.kar[idx].wac;

    assert(kaccess==XPLAB_READ_ACCESS ||              /* consistency check */ 
	   kaccess==XPLAB_WRITE_ACCESS);

    if (kaccess == XPLAB_WRITE_ACCESS) {
      return idx;
    }
  }

  return -1;
}



void
delete_key_def(key_lst *kl, int idx)
{
  int last;

  assert(kl);
  assert(idx>=0);
  assert(idx<kl->cnt);

  if (!kl->kar) {                                     /* consistency check */
    assert(!kl->cnt);
    return;
  }
  else {
    assert(kl->cnt>0);
  }

  last = kl->cnt-1;

  if (idx == last) {
    free_key_def(&kl->kar[idx]);

    if (!last) {
      FREEMEM(kl->kar);
    }
  }
  else {
    move_key_def(&kl->kar[last], &kl->kar[idx]);
  }

  kl->cnt = last;
}



/* Erases a key from list

Revision: 21.05.2004 E.Sack

Input : - pointer to key list
        - pointer to key
        - pointer to key comment string pointer
Output: - key list updated as flagged
        - key updated as flagged
        - comment string pointer updated as flagged
        - flag (0: key found, 1: key not found)
*/
int
erase_key(key_lst *klst, int *key, char **cmt)
{
  int i, found = -1;

  assert(key);

  if (!klst) {
    return 1;                                               /* no key list */
  }

  for (i = 0; i < klst->cnt; i++) {
    if (klst->kar[i].key == *key) {
      found = i;
      break;
    }
  }

  if (found == -1) {
    return 1;                                             /* key not found */
  }

  if (cmt != NULL && klst->kar[found].cmt != NULL) {
    add_string(klst->kar[found].cmt, cmt);
  }

  delete_key_def(klst, found);

  *key = 0;            /* used to supervise consistency in processing keys */

  return 0;
}



/* Updates lists: 

   - if the source list contains data not contained in the destination
     list, the data will be stored in the destination list.

   - if the destination list contains data not contained in the source
     list, the data will be erased in the destination list.

Usage: This function doesn't work like a copy, because the data items
       are pairs. One of them is for the comparison and one, which
       contains an arbitrary one (working data). This is useful for
       controlling dynamic data flows.

Revision: 10.04.2004 E.Sack

Input : - source list
        - destination list
        - flag (0: no shift, 1: shift source datum to aux)
Output: - source list updated as flagged
        - destination list updated 
*/
void
update_int2_lst(int2_lst src, int2_lst *dst, int src_shift_to_aux)
{
  int  i, j, idx, found, dst_item;
  int2 src_item;

  assert(dst);
  assert(src_shift_to_aux == 0 || src_shift_to_aux == 1);

  if (src_shift_to_aux) {                               /* equalize source */
    for (i = 0; i < src.cnt; i++) {
      src.n[i].aux = src.n[i].nr;
    }
  }

  for (i = 0; i < dst->cnt; i++) {/* first part erases old destination data */
    found = 0;
    dst_item = dst->n[i].aux;

    for (j = 0; j < src.cnt; j++) {
      if (src.n[j].aux == dst_item) {
	found = 1;
	break;
      }
    }

    if (!found) {
      delete_int2(dst, i);                                    /* attention */
      i--;
    }
  }

  for (i = 0; i < src.cnt; i++) {      /* second part adds new source data */
    found = 0;
    src_item = src.n[i];

    for (j = 0; j < dst->cnt; j++) {
      if (dst->n[j].aux == src_item.aux) {
	found = 1;
	break;
      }
    }

    if (found) {
      continue;
    }

    idx = dst->cnt++;

    if (idx) {
      REALLOC(int2, dst->cnt, dst->n);
    }
    else {
      CALLOC(int2, dst->cnt, dst->n);
    }

    dst->n[idx] = src_item;
  }
}


void
delete_int2(int2_lst *il, int idx)
{
  int last;

  assert(il);
  assert(idx>=0);
  assert(idx<il->cnt);

  if (!il->n) {                                       /* consistency check */
    assert(!il->cnt);
    return;
  }
  else {
    assert(il->cnt>0);
  }

  last = il->cnt-1;

  if (idx == last) {
    if (!last) {
      free_int2_lst(il);
      return;
    }
  }
  else {
    il->n[idx] = il->n[last];
  }

  il->cnt = last;
  REALLOC(int2, il->cnt, il->n);
}



/* Gives back a default setting.

Revision: 15.01.2006 E.Sack

Input : - -
Output: - default setting
*/
pattern_color
default_pattern_color(void)
{
  pattern_color pc;
  col_def  xplab_color_foreground = XPLAB_COLOR_FOREGROUND;
  col_def  xplab_color_trained    = XPLAB_COLOR_TRAINED;
  col_def  xplab_color_selected   = XPLAB_COLOR_SELECTED;
  col_def  xplab_color_selectbl   = XPLAB_COLOR_SELECTBL;
  col_def  xplab_sign_bgd_in      = XPLAB_SIGN_BGD_IN;
  col_def  xplab_sign_bgd_out     = XPLAB_SIGN_BGD_OUT;
  col_def  xplab_color_ignored    = XPLAB_COLOR_IGNORED;
  col_def  xplab_color_reserved   = XPLAB_COLOR_RESERVED;

  init_pattern_color(&pc);

  pc.fgd = xplab_color_foreground;
  pc.tnd = xplab_color_trained;
  pc.res = xplab_color_reserved;
  pc.sel = xplab_color_selected;
  pc.sbl = xplab_color_selectbl;
  pc.ign = xplab_color_ignored;
  pc.bgd_in  = xplab_sign_bgd_in;
  pc.bgd_out = xplab_sign_bgd_out;

  return pc;
}



/* Gives back a default setting.

Revision: 03.04.2004 E.Sack

Input : - pointer to string
Output: - default setting
*/
char *
default_config_basename(char *fname)
{
  int  err;
  char *cfg_file = NULL, defname[] = CFGNAME_DEFAULT;

  err = check_name(fname);

  if (err) {
    FREEMEM(fname);
    fname = defname;
  }

  err = check_name(fname);
  assert(!err);

  add_string(fname, &cfg_file);

  return cfg_file;
}



/* Gives back a default setting.

Revision: 29.10.2003 E.Sack

Input : - pointer to string
Output: - default setting
*/
char *
default_report_filename(char *fname)
{
  int  err;
  char *rpt_file = NULL, defname[] = REPORT_FNAME_DEFAULT;

  err = check_name(fname);

  if (err) {
    FREEMEM(fname);
    fname = defname;
  }

  err = check_name(fname);
  assert(!err);

  cat_string(fname, FILE_REPORT, "", &rpt_file);

  return rpt_file;
}



/* Gives back a default setting.

Revision: 08.10.2004 E.Sack

Input : - pointer to string
Output: - default setting
*/
char *
default_auto_filename(char *fname)
{
  int  err;
  char *auto_file = NULL, defname[] = XPLAB_MAIN_LOGRUN_DEFAULT;

  err = check_name(fname);

  if (err) {
    FREEMEM(fname);
    fname = defname;
  }

  err = check_name(fname);
  assert(!err);

  add_string(fname, &auto_file);

  return auto_file;
}



/* Gives back a default setting.

Revision: 29.10.2003 E.Sack

Input : - pointer to string
Output: - default setting
*/
char *
default_replay_filename(char *fname)
{
  int  err;
  char *rpl_file = NULL, defname[] = REPLAY_FNAME_DEFAULT;

  err = check_name(fname);

  if (err) {
    FREEMEM(fname);
    fname = defname;
  }

  err = check_name(fname);
  assert(!err);

  cat_string(fname, FILE_REPLAY, "", &rpl_file);

  return rpl_file;
}



/* Gives back a default setting.

Revision: 17.01.2006 E.Sack

Input : - pointer to string
Output: - pointer to default string
*/
char *
default_session_filename(char *fname)
{
  int  err;
  char *ptr, *ses_file = NULL, *buf = NULL;

  add_string(fname, &buf);
  fname = NULL;                                             /* protect input */

  err = check_name(buf);

  if (err) {
    FREEMEM(buf);
    add_string(SESSION_FNAME_DEFAULT, &buf);
  }

  err = check_name(buf);
  assert(!err);

  ptr = strrchr(buf, '.');                /* investigate file name extension */

  if (ptr) {                                    /* avoid multiple extensions */
    *ptr = 0;                                        /* C string termination */
  }

  cat_string(buf, FILE_TRAINING_SESSION, "", &ses_file);

  FREEMEM(buf);

  return ses_file;
}



/* Test string.

Revision: 25.10.2003 E.Sack

Input : - pointer to string
Output: - flag (0: no error, 1: error)
*/
int
check_name(const char *name) 
{
  int err = 1;

  if (name) {                                            /* pointer exists */
    if (strlen(name)) {                                   /* length exists */
      err = 0;
    }
  }

  return err;
}



/* Gives back a default setting.

Revision: 10.10.2003 E.Sack

Input : - -
Output: - default setting
*/
char *
default_subst_rb(void)
{
  char *subst_rb = NULL;

  add_string(SUBSTRB_DEFAULT, &subst_rb);

  return subst_rb;
}



/* Gives back a default setting.

Revision: 10.10.2003 E.Sack

Input : - -
Output: - default setting
*/
char *
default_subst_nb(void)
{
  char *subst_nb = NULL;

  add_string(SUBSTNB_DEFAULT, &subst_nb);

  return subst_nb;
}



/* Gives back a default setting.

Revision: 17.10.2003 E.Sack

Input : - -
Output: - default setting
*/
char *
default_tps_id(void)
{
  char *tps_id = NULL;

  add_string(TPSID_DEFAULT, &tps_id);

  return tps_id;
}



/* Searches the NB index of an auxiliary item.

Revision: 02.10.2004 E.Sack

Input : - NB list
Output: - NB index (-1: no index found)
*/
int
search_aux_nb_def(nb_lst nbl, int item)
{
  int idx;

  if (!nbl.nb) {                                      /* consistency check */
    assert(!nbl.cnt);
  }
  else {
    assert(nbl.cnt>0);
  }

  for (idx = 0; idx < nbl.cnt; idx++) {
    if (nbl.nb[idx].aux == item) {
      return idx;
    }
  }

  return -1;
}



/* Converts 2-D integer values to double.

Revision: 23.10.2004 E.Sack

Input : - image with integers
Output: - image with doubles
*/
pic_dbl
pic_def_to_pic_dbl(pic_def ipic)
{
  int     cnt = AREA_SIZE(ipic.ar);
  PICT    *iptr = ipic.mem;
  double  *dbl_ptr;
  pic_dbl dpic;

  init_pic_dbl(&dpic);

  dpic.ar = ipic.ar;

  CALLOC(double, cnt, dpic.mem);

  dbl_ptr = dpic.mem;

  while (cnt--) {
    *dbl_ptr++ = (double)(*iptr++);
  }

  return dpic;
}
