/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
******************************************************************************/

/* XPLAB is a simple pattern recognition program using pattern matching
   with template pattern stored in a database. There are currently a CLI
   version and a GUI version available. There is a tutorial and a more
   theoretical documentation available on www.pattern-lab.de. */

#ifndef MODULTEST
#define xplab__main main
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_preferences_fe.h"
#include "xplab_preferences_be.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_maintenance_fe.h"
#include "xplab_maintenance_be.h"
#include "xplab_recall_fe.h"
#include "xplab_cluster.h"
#include "xplab_dbio.h"
#include "xplab_image.h"
#include "xplab_tprint.h"
#include "xplab_training_fe.h"
#include "xplab_selection.h"
#include "xplab_replay.h"
#include "xplab_batch.h"
#include "xplab_toolkit.h"
#include "xplab_statistic_fe.h"
#include "xplab_dchannel.h"

/* global variables *********************************************************/
#ifdef NDEBUG
char   P_assert              = 0;
#else
char   P_assert              = 1;
#endif

char   P_debug               = 0;   /* for debugging use, stores allocation */
char   P_print_pointer       = 0;   /* for debugging use, prints allocation */
char   P_print_debug         = 0;   /* for debugging use                    */
void   **P_alloc             = 0;    /* allocation debugging, pointer array */
int    P_alloc_size          = 0;    /* allocation debugging, array size    */
int    P_malloc_free         = 0;    /* allocation debugging, alloc. number */

typedef struct {
  GtkWidget   *top_main, *about_dialog, *log_dump, *log_clear, *match_entry;
  GtkTooltips *tips;
  char        *text_store; /*stores all which cannot delivered to main log*/
} main_def;

static main_def md;
GtkWidget *main_log = NULL;                         /* the main text view */

/* local definitions *******************************************************/
#define XPLAB_MATCHING_TITLE_ITEM    "Matching"
#define XPLAB_MAIN_LOG_WINDOW_WIDTH  511
#define XPLAB_MAIN_LOG_WINDOW_HEIGHT 342

/* local prototypes ********************************************************/
int xplab__main(int argc, char *argv[]);
static void init_main_def(main_def *md);
static void free_main_def(main_def *md);
static void print_info(void);
static void print_license(void);
static void print_terminology(void);
static void main_log_window(void);
static char *xplab_main_log_header(char *suffix);
static void about_text(char *suffix, char **buf);
static void license_text(char **buf);
static void lib_text(char **buf);
static void xplab_rc_parse(void);

/* Functions assembling widgets */
static GtkTooltips *xplab_tooltips_new(void);
static void main_menu_bar(GtkWidget *vbox);
static void main_menu_entry_file(GtkWidget *menu_bar);
static void main_menu_entry_match(GtkWidget *menu_bar);
static void main_help_menu_item_about(GtkWidget *help_menu, GtkWidget **top);
static void main_help_menu_item_term(GtkWidget *help_menu);
static void main_file_menu_item_exit(GtkWidget *file_menu);
static void main_file_menu_item_save(GtkWidget *file_menu);

/* Handlers/Callbacks */
static gboolean cb_main_top_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN);
static void cb_main_exit(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_help_about_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_help_term(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_main_dump(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_main_clear(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_main_check_file_menu(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_main_default(GTK_CONVENTION_MENUITEM_RTS_VOID);

/* Functions **************************************************************/

static void
init_main_def(main_def *md)
{
  assert(md);

  md->top_main = NULL;                                     /* zero pointer */
  md->about_dialog = NULL;                                 /* zero pointer */
  md->text_store = NULL;                                   /* zero pointer */
}

static void
free_main_def(main_def *md)
{
  assert(md);
                           /* don't free GTK stuff */

  FREEMEM(md->text_store);

  init_main_def(md);
}



/* Prints short descriptions.

Revision: 29.07.2005 E.Sack

Input : - -
Output: - log updated
*/
static void
print_terminology(void)
{
  int  len;
  char *termy = NULL, txt[1024];
#define EX_FORMAT "%-10s: %-18s\n"

  len = sizeof(txt);

  g_snprintf(txt, len, "\nShort description of terminology/acronyms:\n");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "ACF", "Auto Correlation Function");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "CCF", "Cross Correlation Function");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "CNB", "Compound NB (consists of several NNBs)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "CV", "Correlation Value");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "DB", "DataBase");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "H phase", 
	   "upper period part of a regular rectangular signal");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "L phase", 
	   "lower period part of a regular rectangular signal");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "MRB","Merged RB (bounding box of several RBs)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "NB", "Named Box (bounding box)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "NNB", "Neighbored NB (part/fragment of a CNB)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "PNM",
	   "portable anymap image file format (PBM, PGM, PPM)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "RB", "Raw Box (bounding box)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "RB chain", "RBs not exceeding a certain distance to their next neighbor");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "RBS", "RB sector");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "TF", "Test Function (binary)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "TPS", "Trained Pattern Set (consists of NBs)");
  add_string(txt, &termy);

  g_snprintf(txt, len, EX_FORMAT, "TPS ID", "TPS IDentification string");
  add_string(txt, &termy);

  add_string("\n", &termy);

  xplab_text_insert(main_log, termy);

  FREEMEM(termy);
#undef EX_FORMAT
}



/* Prints help text of XPLAB.

Revision: 30.07.2004 E.Sack

Input : - -
Output: - help infos
*/
static void
print_info(void)
{
  size_t i;

  static const char* khelp1[] = { /* to avoid long constant string effects */
    "usage: xplab [-h] [-v] [-Y] [-c config-file] [-D db] [ image ]\n",
    " xplab performs Pattern Matching using a trained database.\n",
    "\n"
    };

  static const char* khelp2[] = { /* to avoid long constant string effects */
    "-h     Print this information.\n",
    "-v     Print the version information, some information\n",
    "       about gtk+ it is compiled against and license\n",
    "       information.\n",
    "-Y     Start non-interactive batch job without GUI usage.\n",
    "-c config-file\n",
    "       Use the alternate config-file instead of xplab.cfg.\n",
    "-D db  The database file.\n",
    "image  The image file. The only accepted format is the PNM\n",
    "       format (PBM for black-and-white images, PGM for\n",
    "       gray-scale images, and PPM for color images).\n"
    };

  for (i = 0;  i < sizeof khelp1/sizeof khelp1[0]; i++) {
    fprintf(stdout, "%s", khelp1[i]);
  }

  for (i = 0;  i < sizeof khelp2/sizeof khelp2[0]; i++) {
    fprintf(stdout, " %s", khelp2[i]);
  }
}



/* Prints license of XPLAB.

Revision: 30.07.2004 E.Sack

Input : - -
Output: - license
*/
static void
print_license(void)
{
  size_t i;

  static const char* xplab_license_text[] = {
    "XPLAB comes with ABSOLUTELY NO WARRANTY.\n",
    "This is free software, and you are welcome\n",
    "to redistribute it under certain conditions.\n",
    "For details read the license file which\n",
    "comes with this distribution.\n"
  };

  char suff[] = "  ", *buf = NULL;

  about_text(suff, &buf);

  fprintf(stdout, "%s %s\n%s\n", XPLAB_VERSION, XPLAB_COPYRIGHT_STRING, buf);

  for (i = 0; i<sizeof xplab_license_text/sizeof xplab_license_text[0];i++) {
    fprintf(stdout, "  %s", xplab_license_text[i]);
  }

  fprintf(stdout, "\n");

  FREEMEM(buf);
}



/* Generates main log header.

Revision: 08.10.2005 E.Sack

Input : - pointer to suffix string
Output: - pointer to header string
Postcondition: - header string must be freed
*/
static char *
xplab_main_log_header(char *suffix)
{
  char   *buf = NULL, *p;
  time_t timep = time((time_t *)NULL);

  cat_string(XPLAB_MAIN_LOG_HEADER, "  ", 
	     xctrl.cfg.timestamp ? ctime(&timep):"", &buf);

  p = strchr(buf, '\n');

  if (p != NULL) {
    *p = 0;
  }

  add_string(suffix, &buf);

  return buf;
}



/* Main routine of XPLAB. 

Revision: 29.07.2005 E.Sack

*/
int 
xplab__main(int argc, char *argv[])
{
  int     verbose;
  char    *buf;
  cli_def clid;

  if (0/*macro*/) {                                                 /* debug purposes */
  gtk_debug_flags |= GTK_DEBUG_UPDATES | 
                     GTK_DEBUG_GEOMETRY | 
                     GTK_DEBUG_TEXT | 
                     GTK_DEBUG_TREE | 
                     GTK_DEBUG_MISC | 
                     GTK_DEBUG_PLUGSOCKET |
                     GTK_DEBUG_KEYBINDINGS |
                     GTK_DEBUG_MULTIHEAD |
                     GTK_DEBUG_ICONTHEME |
                     GTK_DEBUG_MODULES;
  }

  init_prog_ctrl(&xctrl);                                /* init global data */
  init_main_def(&md);                                     /* init local data */

  buf = xplab_main_log_header("\n\n"),
  xplab_text_insert(main_log, buf);
  FREEMEM(buf);

  xplab_dc_init();
  xplab_init_selection();
  xplab_init_replay();
  xplab_init_recall();
  xplab_init_clustering();
  xplab_init_maintenance();
  xplab_init_training();
  xplab_init_toolkit();
  xplab_init_statistic();

  clid = xplab_init_preferences(argc, argv);            /* opens report file */

  if (clid.help) {                                          /* printout help */
    print_info();
  }
  else if (clid.license) {                               /* printout license */
    print_license();
  }
  else if (xctrl.set.dobatch == TRUE) {                   /* start batch job */
    xplab_batch(argc, argv, &xctrl);                               /* no GUI */
  }
  else {                                                        /* start GUI */
    gtk_init(&argc, &argv);                                      /* GTK init */

    xplab_rc_parse();                             /* parse Gtk resource file */

    main_log_window();

    gtk_main();
  }

  xplab_dc_exit();

  verbose = xctrl.cfg.verbose >= VERBOSE_LOT;              /* debug purposes */
  xplab_free_statistic();
  xplab_free_toolkit();
  xplab_free_maintenance();
  xplab_free_clustering();
  xplab_free_recall();
  xplab_free_replay();
  xplab_free_selection();
  xplab_free_training();
  xplab_free_preferences();                            /* closes report file */
  free_main_def(&md);

  if (P_debug && verbose) {
    int ali;

    fprintf(stderr, "DEBUG: allocation table size %d\n", P_alloc_size);

    for (ali = 0; ali < P_alloc_size; ali++) {
      fprintf(stderr, "DEBUG: Palloc[%d] = %p\n", ali, P_alloc[ali]);
    }
  }

  if (P_print_debug && verbose) {
    fprintf(stderr, "DEBUG: print routines used\n");
  }

  if (P_assert && verbose) {
    fprintf(stderr, "DEBUG: assertions used\n");
  }

  assert(!P_malloc_free);              /* internal test of memory allocation */

  exit(EXIT_SUCCESS);
}



/* Parses Gtk resource file for XPLAB.

Revision: 14.01.2005 E.Sack

Input : - -
Output: - Gtk resources updated
Usage: - as long as there is no file name defined in the preferences file
         of XPLAB, no Gtk resource file will be read.
*/
static void
xplab_rc_parse(void)
{
  gint err;

  err = check_name(xctrl.set.gtk_rc);

  if (!err) {
    gtk_rc_parse(xctrl.set.gtk_rc);
  }
}



/* Creates main window for XPLAB.

Revision: 29.07.2005 E.Sack

Input : - -
Output: - top level window
*/
static void
main_log_window(void)
{
  GtkWidget *vbox;

  md.tips = xplab_tooltips_new();       /* create the general tooltips group */
  xplab_tooltips(xctrl.cfg.tips);

  /* set the main top level widget                                           */
  md.top_main = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(md.top_main), XPLAB_VERSION);
  gtk_window_set_default_size(GTK_WINDOW(md.top_main), 
			      XPLAB_MAIN_LOG_WINDOW_WIDTH, 
			      XPLAB_MAIN_LOG_WINDOW_HEIGHT);
  gtk_widget_set_name(md.top_main, XPLAB_RC_MAIN_LOG);         /* for Gtk RC */

  /* put into top level widget a vertical box                                */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(md.top_main), vbox);

  /* put into vertical box a menubar                                         */
  main_menu_bar(vbox);

  /* put into vertical box a scrolled text window */
  main_log = scrolled_text(vbox);

  gtk_widget_show_all(md.top_main);

  xplab_text_insert(main_log, "");

  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_clustering_top_delete), NULL);
  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_replay_top_delete), NULL);
  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_recall_top_delete), NULL);
  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_training_delete), NULL);
  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_maintenance_top_delete), NULL);
  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_statistic_delete_all), NULL);
  g_signal_connect(G_OBJECT(md.top_main), "delete-event", G_CALLBACK(cb_main_top_delete), NULL);

  xplab_set_window_icon(md.top_main);                       /* window icon */
}



/* Builds a menu bar.

Revision: 29.07.2005 E.Sack

Input : - pointer to container
Output: - container updated
Bug: There is an internal GTK2.x delay MENU_SHELL_TIMEOUT which prevents 
     fast press-releases. So items not containing a submenu but a callback
     needs a double-click or an extended release if the menu-bar isn't 
     selected already. A complicated solution could be using the signal
     "button_release_event" bound to the menu-bar.
*/
static void
main_menu_bar(GtkWidget *vbox)
{
  GtkWidget *menu_bar;

  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

  main_menu_entry_file(menu_bar);                             /* menu file */
  xplab_training_menu(menu_bar, &md.top_main);            /* menu training */
  main_menu_entry_match(menu_bar);                        /* menu matching */
  xplab_maintenance_menu(menu_bar, &md.top_main);      /* menu maintenance */
  xplab_statistic_menu(menu_bar, &md.top_main);          /* menu statistic */
  xplab_toolkit_menu(menu_bar);                            /* menu toolkit */
  xplab_menu_help(menu_bar, &md.top_main);                    /* menu help */
}



/* Builds an entry for a menu bar.

Revision: 29.07.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
main_menu_entry_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *file_menu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  file_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), file_menu);

  xplab_preferences_menu(file_menu, &md.top_main);     /* item preferences */
  xplab_selection_menu(file_menu, &md.top_main);         /* item selection */
  xplab_preferences_menu_unload(file_menu);                 /* item unload */
  xplab_preferences_menu_tpsid(file_menu);                  /* item TPS ID */
  file_menu_item_log(file_menu,                                /* item log */
		     GTK_SIGNAL_FUNC(cb_main_dump), &md.log_dump,
		     GTK_SIGNAL_FUNC(cb_main_clear), &md.log_clear,
		     GTK_SIGNAL_FUNC(cb_main_default));

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep);

  main_file_menu_item_save(file_menu);                        /* item save */
  main_file_menu_item_exit(file_menu);                        /* item exit */

  g_signal_connect(G_OBJECT(file_entry), "activate", 
		   G_CALLBACK(cb_main_check_file_menu), NULL);
}



/* Builds an entry for a menu bar.

Revision: 29.07.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
void
xplab_menu_help(GtkWidget *menu_bar, GtkWidget **top)
{
  GtkWidget *help_entry, *help_menu;

  help_entry = gtk_menu_item_new_with_label(MENU_BUTTON_HELP);
  gtk_menu_item_set_right_justified(GTK_MENU_ITEM(help_entry), TRUE);

  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), help_entry);

  help_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_entry), help_menu);

  main_help_menu_item_about(help_menu, top);                 /* item about */
  main_help_menu_item_term(help_menu);                 /* item terminology */
}



/* Controls tools tip group.

Revision: 04.05.2004 E.Sack

Input : - tooltip group
        - flag
Output: - tooltip group updated
*/
void
xplab_tooltips(gboolean enable)
{
  if (GTK_IS_TOOLTIPS(md.tips) == FALSE) {
    return;
  }

  if (enable == TRUE) {
    gtk_tooltips_enable(GTK_TOOLTIPS(md.tips));
  }
  else {
    gtk_tooltips_disable(GTK_TOOLTIPS(md.tips));
  }
}



/* Returns tool tip group.

Revision: 29.07.2005 E.Sack

Input : - -
Output: - tooltip group
*/
static GtkTooltips*
xplab_tooltips_new(void)
{
  GtkTooltips *tips;
  GdkColor    tips_fg = XPLAB_TOOLTIPS_FGND, tips_bg = XPLAB_TOOLTIPS_BGND;
  GdkColormap *colormap;
  GtkStyle    *current_style;

  tips = gtk_tooltips_new();

  colormap = gdk_colormap_get_system();

  gdk_colormap_alloc_color(colormap, &tips_fg, FALSE, TRUE);
  gdk_colormap_alloc_color(colormap, &tips_bg, FALSE, TRUE);

  gtk_tooltips_force_window(tips);                         /* settle style */

  current_style = gtk_style_copy(gtk_widget_get_style(tips->tip_window));
  current_style->bg[GTK_STATE_NORMAL] = tips_bg;
  current_style->fg[GTK_STATE_NORMAL] = tips_fg;
  gtk_widget_set_style(tips->tip_window, current_style);  

  return tips;
}



/* Sets up tool tip group.

Revision: 04.06.2004 E.Sack

Input : - tooltip group
        - flag
Output: - tooltip group updated
*/
void
xplab_tooltips_set_tip(GtkWidget *widget, const gchar *tip_text, 
		       const gchar *tip_private)
{
  if (GTK_WIDGET_NO_WINDOW(widget) == TRUE) {
    return;
  }

  assert(GTK_IS_TOOLTIPS(md.tips) == TRUE);

  gtk_tooltips_set_tip(GTK_TOOLTIPS(md.tips), widget, tip_text, tip_private);
}



/* Builds an entry for a menu bar.

Revision: 29.07.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
main_menu_entry_match(GtkWidget *menu_bar)
{
  GtkWidget *match_menu;

  md.match_entry = gtk_menu_item_new_with_label(XPLAB_MATCHING_TITLE_ITEM);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), md.match_entry);

  match_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(md.match_entry), match_menu);

  xplab_recall_menu(match_menu, &md.top_main);              /* item recall */
  xplab_replay_menu(match_menu, &md.top_main);              /* item replay */
  xplab_clustering_menu(match_menu, &md.top_main);      /* item clustering */
}



/* Adds an item to a menu.

Revision: 29.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
main_file_menu_item_exit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_exit;

  file_menu_exit = gtk_menu_item_new_with_label(MENU_BUTTON_EXIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_exit);

  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_clustering_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_replay_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_recall_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_training_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_maintenance_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_statistic_quit_all), NULL);
  g_signal_connect(G_OBJECT(file_menu_exit), "activate", G_CALLBACK(cb_main_exit), NULL);
}



/* Adds an item to a menu.

Revision: 29.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
main_help_menu_item_about(GtkWidget *help_menu, GtkWidget **top)
{
  GtkWidget *help_menu_about;

  help_menu_about = gtk_menu_item_new_with_label(MENU_BUTTON_ABOUT XPLAB_MENU_EXT);

  gtk_menu_shell_append(GTK_MENU_SHELL(help_menu), help_menu_about);

  g_signal_connect(G_OBJECT(help_menu_about), "activate", 
		   G_CALLBACK(cb_help_about_dialog), (gpointer)top);
}



/* Adds an item to a menu.

Revision: 29.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
main_help_menu_item_term(GtkWidget *help_menu)
{
  GtkWidget *help_menu_term;

  help_menu_term = gtk_menu_item_new_with_label(MENU_BUTTON_TERMINOLOGY);
  gtk_menu_shell_append(GTK_MENU_SHELL(help_menu), help_menu_term);

  g_signal_connect(G_OBJECT(help_menu_term), "activate", 
		   G_CALLBACK(cb_help_term), NULL);
}



/* Adds an item to a menu.

Revision: 29.07.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
main_file_menu_item_save(GtkWidget *file_menu)
{
  GtkWidget *file_menu_save;

  file_menu_save = gtk_menu_item_new_with_label(MENU_BUTTON_SAVE_EXIT);

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_save);

  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_clustering_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_replay_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_recall_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_training_save_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_maintenance_save_quit), NULL);
  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_statistic_quit_all), NULL);
  g_signal_connect(G_OBJECT(file_menu_save), "activate", G_CALLBACK(cb_main_exit), NULL);
}



/* Deletes top window.

Revision: 19.12.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_main_top_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_main_exit(NULL, NULL);

  return FALSE;
}



/* Quits GTK main loop.

Revision: 08.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_main_exit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  gtk_main_quit();
}



/* Dumps main log.

Revision: 30.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_main_dump(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  dump_gtk_text(main_log, XPLAB_MAIN_LOGNAME_DEFAULT);
}



/* Checks log menu entry.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_main_check_file_menu(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int           len;
  GtkTextBuffer *buffer;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  xplab_preferences_menu_tpsid(NULL);             /* build new item TPS ID */

  if (GTK_IS_TEXT_VIEW(main_log) == TRUE) {              /* check main log */
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(main_log));/* fetch default buffer */
    len = gtk_text_buffer_get_char_count(buffer);

    gtk_widget_set_sensitive(md.log_clear, len ? TRUE : FALSE);
    gtk_widget_set_sensitive(md.log_dump, len ? TRUE : FALSE);
  }
}



/* Clears main log.

Revision: 02.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_main_clear(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  clear_gtk_text(main_log);
}



/* Sets default main log header.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_main_default(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char *buf = NULL;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_TEXT_VIEW(main_log) == TRUE) {
    clear_gtk_text(main_log);

    buf = xplab_main_log_header("\n\n");

    xplab_text_insert(main_log, buf);
  }

  FREEMEM(buf);
}



/* Dumps terminology text into log.

Revision: 27.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_help_term(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  raised = xplab_window_raise(&md.top_main, 1);                /* main log */

  if (raised == TRUE) {
    gtk_main_iteration_do(FALSE);

    print_terminology();                               /* dump to main log */
  }
}



/* Shows text about XPLAB.

Revision: 29.07.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_help_about_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean    raised;
  char        *buf = NULL;
  const gchar *authors[] = {XPLAB_COPYRIGHT_AUTHOR "   " XPLAB_EMAIL, NULL};

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  raised = xplab_window_raise(&md.about_dialog, 1);

  if (raised == TRUE) {
    gtk_widget_show_all(md.about_dialog);
    return;
  }

  md.about_dialog = gtk_about_dialog_new();

  gtk_about_dialog_set_name(GTK_ABOUT_DIALOG(md.about_dialog), "");

  gtk_about_dialog_set_logo(GTK_ABOUT_DIALOG(md.about_dialog),
			    xplab_logo());

  gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(md.about_dialog),
			       XPLAB_VERSION);

  gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG(md.about_dialog),
				 XPLAB_COPYRIGHT_STRING);

  gtk_about_dialog_set_website(GTK_ABOUT_DIALOG(md.about_dialog),
                               XPLAB_HOMEPAGE);

  gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG(md.about_dialog),
				   authors);

  lib_text(&buf);

  gtk_about_dialog_set_comments(GTK_ABOUT_DIALOG(md.about_dialog), buf);
  FREEMEM(buf);

  license_text(&buf);

  gtk_about_dialog_set_license(GTK_ABOUT_DIALOG(md.about_dialog), buf);
  FREEMEM(buf);

  gtk_window_set_title(GTK_WINDOW(md.about_dialog), "about");

  gtk_widget_show_all(md.about_dialog);


  xplab_set_window_icon(md.about_dialog);                   /* window icon */
}



/* Assembles the license text string.

Revision: 29.07.2005 E.Sack

Input : - pointer to input string
Output: - input string updated
*/
static void
license_text(char **io_buf)
{
  gint i;

  const gchar *license[] = {
    "This program is free software; you can redistribute it and/or",
    "modify it under the terms of the GNU General Public License",
    "as published by the Free Software Foundation; either version 2",
    "of the License, or (at your option) any later version.",
    "",
    "This program is distributed in the hope that it will be useful,",
    "but WITHOUT ANY WARRANTY; without even the implied warranty of",
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
    "GNU General Public License for more details.",
    "",
    "You should have received a copy of the GNU General Public License",
    "along with this program; if not, write to the Free Software",
    "Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.",
    NULL};

  if (!io_buf) {
    return;
  }

  for (i = 0; license[i]; i++) {
    cat_string(license[i], "\n", "", io_buf);
  }
}



/* Assembles the library version text string.

Revision: 29.07.2005 E.Sack

Input : - pointer to input string
Output: - input string updated
*/
static void
lib_text(char **io_buf)
{
  if (!io_buf) {
    return;
  }

  int_to_string(XPLAB_GRAPHIC_LIB, gtk_major_version, "", io_buf);
  int_to_string(".", gtk_minor_version, ".", io_buf);
  int_to_string("", gtk_micro_version, "\n", io_buf);
}



/* Adds text about XPLAB.

Revision: 29.07.2005 E.Sack

Input : - pointer to string
Output: - string updated
*/
static void
about_text(char *suffix, char **buf)
{
  if (!buf) {
    return;
  }

  cat_string(suffix, "Library: ", "", buf);
  lib_text(buf);

  cat_string(suffix, "Homepage: ", XPLAB_HOMEPAGE "\n", buf);
  cat_string(suffix, "Email: ", XPLAB_EMAIL "\n", buf);
}



/* Wrapper using internal store for undefined GTK.

Revision: 29.07.2005 E.Sack

Input : - log view
Output: - storage updated
Usage:  - function should be used whenever a gtk_text_insert() has to 
          be called.
Postcondition: - the text is UTF8 converted.
*/
void
xplab_text_insert(GtkWidget *text_view, const char *chars)
{
  gint          d;
  gchar         *sum;
  gboolean      valid_utf;
  GtkTextIter   iter;
  GtkTextBuffer *buffer;

  if (!chars) {
    return;                                                /* empty string */
  }

  /* First of all, try to write unsaved text into the main log not done so
     far after program start.                                              */
  if (main_log) {
    if (GTK_IS_TEXT_VIEW(main_log) == TRUE && md.text_store) {
      dump_main_log_text(main_log, md.text_store, XPLAB_MAIN_LOGRUN_DEFAULT);

      buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(main_log));

      gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer), &iter);

      valid_utf = g_utf8_validate(md.text_store, -1, NULL);

      sum = (valid_utf == TRUE) ? 
	    g_strdup(md.text_store) : xplab_to_utf(md.text_store);

      gtk_text_buffer_insert(GTK_TEXT_BUFFER(buffer), &iter, 
			     sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
      g_free(sum);

      FREEMEM(md.text_store);        /* empty general auxiliary text store */
    }
  }

  /* now write the current input to the log */
  if (text_view) {
    if (GTK_IS_TEXT_VIEW(text_view) == TRUE) {
      GtkTextMark* mark;

      dump_main_log_text(text_view, chars, XPLAB_MAIN_LOGRUN_DEFAULT);

      buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));

      gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer), &iter);

      mark = gtk_text_buffer_create_mark(GTK_TEXT_BUFFER(buffer),
					 "MARK_END", &iter, FALSE);

      valid_utf = g_utf8_validate(chars, -1, NULL);

      sum = (valid_utf == TRUE) ? 
	g_strdup(chars) : xplab_to_utf(chars);

      gtk_text_buffer_insert(GTK_TEXT_BUFFER(buffer), &iter, 
			     sum?sum:XPLAB_UTF_CONVERSION_ERROR, -1);
      g_free(sum);

      /* follow text in display */
      gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(text_view), mark, 0.0, FALSE, 
				   0.0, 0.0);
      gtk_text_buffer_delete_mark_by_name(GTK_TEXT_BUFFER(buffer),
					  "MARK_END");
      return;                                          /* GUI or batch log */
    }
  }
  else if (xctrl.set.dobatch == FALSE) {
    add_string(chars, &md.text_store);/* fill general auxiliary text store */
  }

  if (xctrl.set.dobatch == TRUE) {
    if (md.text_store) {             /* empty general auxiliary text store */
      d = write_to_batch_log(xctrl.cfg.batch.log_stream, md.text_store);

      if (d) {
	FREEMEM(md.text_store);
      }
    }

    d = write_to_batch_log(xctrl.cfg.batch.log_stream, chars);

    if (!d) {
      add_string(chars, &md.text_store);/* fill general auxiliary text store */
    }
  }
}
